import{j as o,Y as td,O as rg,k as dg,u as ad,Z as mg,c as hg,l as gg,g as pg,S as yg,T as vg,n as bg,m as nd,o as Sg,p as Tg,$ as ud,a0 as id,a1 as cd,a2 as xg}from"./ui-vendor-CeCm8EER.js";import{d as Ag,h as Dg,r as E,u as sd,H as Ng,i as Eg,j as kr}from"./react-vendor-DEwriMA6.js";import{w as ll,c as we,a6 as od,u as ql,v as Bt,a7 as fd,a8 as rd,I as us,B as Cn,D as Mg,i as zg,j as Cg,k as Og,l as jg,a9 as Rg,aa as Ug,ab as _g,ac as Hg,ad as Ll,ae as dd,af as ss,ag as is,ah as Lg,ai as qg,aj as Bg,ak as Gg,al as Yg,am as wg,an as md,ao as Xg,ap as Qg,aq as hd,ar as Vg,as as gd,C as Kg,z as Zg,G as kg,d as En,at as Jg,au as Fg,av as $g}from"./feature-graph-B5dlndAs.js";import{S as Jr,a as Fr,b as $r,c as Wr,d as ot,R as Wg}from"./feature-retrieval-ClAhVsls.js";import{D as Pg}from"./feature-documents-C4vtaK7t.js";import{i as cs}from"./utils-vendor-BysuhMZA.js";import"./graph-vendor-B-X5JegA.js";import"./mermaid-vendor-uLomm4dI.js";import"./markdown-vendor-BcK8zFK7.js";(function(){const b=document.createElement("link").relList;if(b&&b.supports&&b.supports("modulepreload"))return;for(const N of document.querySelectorAll('link[rel="modulepreload"]'))d(N);new MutationObserver(N=>{for(const j of N)if(j.type==="childList")for(const H of j.addedNodes)H.tagName==="LINK"&&H.rel==="modulepreload"&&d(H)}).observe(document,{childList:!0,subtree:!0});function x(N){const j={};return N.integrity&&(j.integrity=N.integrity),N.referrerPolicy&&(j.referrerPolicy=N.referrerPolicy),N.crossOrigin==="use-credentials"?j.credentials="include":N.crossOrigin==="anonymous"?j.credentials="omit":j.credentials="same-origin",j}function d(N){if(N.ep)return;N.ep=!0;const j=x(N);fetch(N.href,j)}})();var ts={exports:{}},Mn={},as={exports:{}},ns={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Pr;function Ig(){return Pr||(Pr=1,function(g){function b(A,L){var U=A.length;A.push(L);e:for(;0<U;){var te=U-1>>>1,oe=A[te];if(0<N(oe,L))A[te]=L,A[U]=oe,U=te;else break e}}function x(A){return A.length===0?null:A[0]}function d(A){if(A.length===0)return null;var L=A[0],U=A.pop();if(U!==L){A[0]=U;e:for(var te=0,oe=A.length,wt=oe>>>1;te<wt;){var gt=2*(te+1)-1,vl=A[gt],V=gt+1,Xe=A[V];if(0>N(vl,U))V<oe&&0>N(Xe,vl)?(A[te]=Xe,A[V]=U,te=V):(A[te]=vl,A[gt]=U,te=gt);else if(V<oe&&0>N(Xe,U))A[te]=Xe,A[V]=U,te=V;else break e}}return L}function N(A,L){var U=A.sortIndex-L.sortIndex;return U!==0?U:A.id-L.id}if(g.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var j=performance;g.unstable_now=function(){return j.now()}}else{var H=Date,$=H.now();g.unstable_now=function(){return H.now()-$}}var Y=[],W=[],he=1,ge=null,X=3,pe=!1,le=!1,C=!1,tl=typeof setTimeout=="function"?setTimeout:null,Gt=typeof clearTimeout=="function"?clearTimeout:null,al=typeof setImmediate<"u"?setImmediate:null;function ft(A){for(var L=x(W);L!==null;){if(L.callback===null)d(W);else if(L.startTime<=A)d(W),L.sortIndex=L.expirationTime,b(Y,L);else break;L=x(W)}}function Ea(A){if(C=!1,ft(A),!le)if(x(Y)!==null)le=!0,mt();else{var L=x(W);L!==null&&ht(Ea,L.startTime-A)}}var rt=!1,nl=-1,On=5,Yt=-1;function R(){return!(g.unstable_now()-Yt<On)}function k(){if(rt){var A=g.unstable_now();Yt=A;var L=!0;try{e:{le=!1,C&&(C=!1,Gt(nl),nl=-1),pe=!0;var U=X;try{l:{for(ft(A),ge=x(Y);ge!==null&&!(ge.expirationTime>A&&R());){var te=ge.callback;if(typeof te=="function"){ge.callback=null,X=ge.priorityLevel;var oe=te(ge.expirationTime<=A);if(A=g.unstable_now(),typeof oe=="function"){ge.callback=oe,ft(A),L=!0;break l}ge===x(Y)&&d(Y),ft(A)}else d(Y);ge=x(Y)}if(ge!==null)L=!0;else{var wt=x(W);wt!==null&&ht(Ea,wt.startTime-A),L=!1}}break e}finally{ge=null,X=U,pe=!1}L=void 0}}finally{L?yl():rt=!1}}}var yl;if(typeof al=="function")yl=function(){al(k)};else if(typeof MessageChannel<"u"){var Ma=new MessageChannel,dt=Ma.port2;Ma.port1.onmessage=k,yl=function(){dt.postMessage(null)}}else yl=function(){tl(k,0)};function mt(){rt||(rt=!0,yl())}function ht(A,L){nl=tl(function(){A(g.unstable_now())},L)}g.unstable_IdlePriority=5,g.unstable_ImmediatePriority=1,g.unstable_LowPriority=4,g.unstable_NormalPriority=3,g.unstable_Profiling=null,g.unstable_UserBlockingPriority=2,g.unstable_cancelCallback=function(A){A.callback=null},g.unstable_continueExecution=function(){le||pe||(le=!0,mt())},g.unstable_forceFrameRate=function(A){0>A||125<A?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):On=0<A?Math.floor(1e3/A):5},g.unstable_getCurrentPriorityLevel=function(){return X},g.unstable_getFirstCallbackNode=function(){return x(Y)},g.unstable_next=function(A){switch(X){case 1:case 2:case 3:var L=3;break;default:L=X}var U=X;X=L;try{return A()}finally{X=U}},g.unstable_pauseExecution=function(){},g.unstable_requestPaint=function(){},g.unstable_runWithPriority=function(A,L){switch(A){case 1:case 2:case 3:case 4:case 5:break;default:A=3}var U=X;X=A;try{return L()}finally{X=U}},g.unstable_scheduleCallback=function(A,L,U){var te=g.unstable_now();switch(typeof U=="object"&&U!==null?(U=U.delay,U=typeof U=="number"&&0<U?te+U:te):U=te,A){case 1:var oe=-1;break;case 2:oe=250;break;case 5:oe=1073741823;break;case 4:oe=1e4;break;default:oe=5e3}return oe=U+oe,A={id:he++,callback:L,priorityLevel:A,startTime:U,expirationTime:oe,sortIndex:-1},U>te?(A.sortIndex=U,b(W,A),x(Y)===null&&A===x(W)&&(C?(Gt(nl),nl=-1):C=!0,ht(Ea,U-te))):(A.sortIndex=oe,b(Y,A),le||pe||(le=!0,mt())),A},g.unstable_shouldYield=R,g.unstable_wrapCallback=function(A){var L=X;return function(){var U=X;X=L;try{return A.apply(this,arguments)}finally{X=U}}}}(ns)),ns}var Ir;function ep(){return Ir||(Ir=1,as.exports=Ig()),as.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var ed;function lp(){if(ed)return Mn;ed=1;var g=ep(),b=Ag(),x=Dg();function d(e){var l="https://react.dev/errors/"+e;if(1<arguments.length){l+="?args[]="+encodeURIComponent(arguments[1]);for(var t=2;t<arguments.length;t++)l+="&args[]="+encodeURIComponent(arguments[t])}return"Minified React error #"+e+"; visit "+l+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function N(e){return!(!e||e.nodeType!==1&&e.nodeType!==9&&e.nodeType!==11)}var j=Symbol.for("react.element"),H=Symbol.for("react.transitional.element"),$=Symbol.for("react.portal"),Y=Symbol.for("react.fragment"),W=Symbol.for("react.strict_mode"),he=Symbol.for("react.profiler"),ge=Symbol.for("react.provider"),X=Symbol.for("react.consumer"),pe=Symbol.for("react.context"),le=Symbol.for("react.forward_ref"),C=Symbol.for("react.suspense"),tl=Symbol.for("react.suspense_list"),Gt=Symbol.for("react.memo"),al=Symbol.for("react.lazy"),ft=Symbol.for("react.offscreen"),Ea=Symbol.for("react.memo_cache_sentinel"),rt=Symbol.iterator;function nl(e){return e===null||typeof e!="object"?null:(e=rt&&e[rt]||e["@@iterator"],typeof e=="function"?e:null)}var On=Symbol.for("react.client.reference");function Yt(e){if(e==null)return null;if(typeof e=="function")return e.$$typeof===On?null:e.displayName||e.name||null;if(typeof e=="string")return e;switch(e){case Y:return"Fragment";case $:return"Portal";case he:return"Profiler";case W:return"StrictMode";case C:return"Suspense";case tl:return"SuspenseList"}if(typeof e=="object")switch(e.$$typeof){case pe:return(e.displayName||"Context")+".Provider";case X:return(e._context.displayName||"Context")+".Consumer";case le:var l=e.render;return e=e.displayName,e||(e=l.displayName||l.name||"",e=e!==""?"ForwardRef("+e+")":"ForwardRef"),e;case Gt:return l=e.displayName||null,l!==null?l:Yt(e.type)||"Memo";case al:l=e._payload,e=e._init;try{return Yt(e(l))}catch{}}return null}var R=b.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,k=Object.assign,yl,Ma;function dt(e){if(yl===void 0)try{throw Error()}catch(t){var l=t.stack.trim().match(/\n( *(at )?)/);yl=l&&l[1]||"",Ma=-1<t.stack.indexOf(`
    at`)?" (<anonymous>)":-1<t.stack.indexOf("@")?"@unknown:0:0":""}return`
`+yl+e+Ma}var mt=!1;function ht(e,l){if(!e||mt)return"";mt=!0;var t=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var a={DetermineComponentFrameRoot:function(){try{if(l){var T=function(){throw Error()};if(Object.defineProperty(T.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(T,[])}catch(y){var p=y}Reflect.construct(e,[],T)}else{try{T.call()}catch(y){p=y}e.call(T.prototype)}}else{try{throw Error()}catch(y){p=y}(T=e())&&typeof T.catch=="function"&&T.catch(function(){})}}catch(y){if(y&&p&&typeof y.stack=="string")return[y.stack,p.stack]}return[null,null]}};a.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var n=Object.getOwnPropertyDescriptor(a.DetermineComponentFrameRoot,"name");n&&n.configurable&&Object.defineProperty(a.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var u=a.DetermineComponentFrameRoot(),i=u[0],c=u[1];if(i&&c){var s=i.split(`
`),r=c.split(`
`);for(n=a=0;a<s.length&&!s[a].includes("DetermineComponentFrameRoot");)a++;for(;n<r.length&&!r[n].includes("DetermineComponentFrameRoot");)n++;if(a===s.length||n===r.length)for(a=s.length-1,n=r.length-1;1<=a&&0<=n&&s[a]!==r[n];)n--;for(;1<=a&&0<=n;a--,n--)if(s[a]!==r[n]){if(a!==1||n!==1)do if(a--,n--,0>n||s[a]!==r[n]){var v=`
`+s[a].replace(" at new "," at ");return e.displayName&&v.includes("<anonymous>")&&(v=v.replace("<anonymous>",e.displayName)),v}while(1<=a&&0<=n);break}}}finally{mt=!1,Error.prepareStackTrace=t}return(t=e?e.displayName||e.name:"")?dt(t):""}function A(e){switch(e.tag){case 26:case 27:case 5:return dt(e.type);case 16:return dt("Lazy");case 13:return dt("Suspense");case 19:return dt("SuspenseList");case 0:case 15:return e=ht(e.type,!1),e;case 11:return e=ht(e.type.render,!1),e;case 1:return e=ht(e.type,!0),e;default:return""}}function L(e){try{var l="";do l+=A(e),e=e.return;while(e);return l}catch(t){return`
Error generating stack: `+t.message+`
`+t.stack}}function U(e){var l=e,t=e;if(e.alternate)for(;l.return;)l=l.return;else{e=l;do l=e,l.flags&4098&&(t=l.return),e=l.return;while(e)}return l.tag===3?t:null}function te(e){if(e.tag===13){var l=e.memoizedState;if(l===null&&(e=e.alternate,e!==null&&(l=e.memoizedState)),l!==null)return l.dehydrated}return null}function oe(e){if(U(e)!==e)throw Error(d(188))}function wt(e){var l=e.alternate;if(!l){if(l=U(e),l===null)throw Error(d(188));return l!==e?null:e}for(var t=e,a=l;;){var n=t.return;if(n===null)break;var u=n.alternate;if(u===null){if(a=n.return,a!==null){t=a;continue}break}if(n.child===u.child){for(u=n.child;u;){if(u===t)return oe(n),e;if(u===a)return oe(n),l;u=u.sibling}throw Error(d(188))}if(t.return!==a.return)t=n,a=u;else{for(var i=!1,c=n.child;c;){if(c===t){i=!0,t=n,a=u;break}if(c===a){i=!0,a=n,t=u;break}c=c.sibling}if(!i){for(c=u.child;c;){if(c===t){i=!0,t=u,a=n;break}if(c===a){i=!0,a=u,t=n;break}c=c.sibling}if(!i)throw Error(d(189))}}if(t.alternate!==a)throw Error(d(190))}if(t.tag!==3)throw Error(d(188));return t.stateNode.current===t?e:l}function gt(e){var l=e.tag;if(l===5||l===26||l===27||l===6)return e;for(e=e.child;e!==null;){if(l=gt(e),l!==null)return l;e=e.sibling}return null}var vl=Array.isArray,V=x.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,Xe={pending:!1,data:null,method:null,action:null},Zu=[],Xt=-1;function ol(e){return{current:e}}function be(e){0>Xt||(e.current=Zu[Xt],Zu[Xt]=null,Xt--)}function ae(e,l){Xt++,Zu[Xt]=e.current,e.current=l}var fl=ol(null),za=ol(null),Gl=ol(null),jn=ol(null);function Rn(e,l){switch(ae(Gl,l),ae(za,e),ae(fl,null),e=l.nodeType,e){case 9:case 11:l=(l=l.documentElement)&&(l=l.namespaceURI)?xr(l):0;break;default:if(e=e===8?l.parentNode:l,l=e.tagName,e=e.namespaceURI)e=xr(e),l=Ar(e,l);else switch(l){case"svg":l=1;break;case"math":l=2;break;default:l=0}}be(fl),ae(fl,l)}function Qt(){be(fl),be(za),be(Gl)}function ku(e){e.memoizedState!==null&&ae(jn,e);var l=fl.current,t=Ar(l,e.type);l!==t&&(ae(za,e),ae(fl,t))}function Un(e){za.current===e&&(be(fl),be(za)),jn.current===e&&(be(jn),Tn._currentValue=Xe)}var Ju=Object.prototype.hasOwnProperty,Fu=g.unstable_scheduleCallback,$u=g.unstable_cancelCallback,Qd=g.unstable_shouldYield,Vd=g.unstable_requestPaint,rl=g.unstable_now,Kd=g.unstable_getCurrentPriorityLevel,os=g.unstable_ImmediatePriority,fs=g.unstable_UserBlockingPriority,_n=g.unstable_NormalPriority,Zd=g.unstable_LowPriority,rs=g.unstable_IdlePriority,kd=g.log,Jd=g.unstable_setDisableYieldValue,Ca=null,_e=null;function Fd(e){if(_e&&typeof _e.onCommitFiberRoot=="function")try{_e.onCommitFiberRoot(Ca,e,void 0,(e.current.flags&128)===128)}catch{}}function Yl(e){if(typeof kd=="function"&&Jd(e),_e&&typeof _e.setStrictMode=="function")try{_e.setStrictMode(Ca,e)}catch{}}var He=Math.clz32?Math.clz32:Pd,$d=Math.log,Wd=Math.LN2;function Pd(e){return e>>>=0,e===0?32:31-($d(e)/Wd|0)|0}var Hn=128,Ln=4194304;function pt(e){var l=e&42;if(l!==0)return l;switch(e&-e){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return e&4194176;case 4194304:case 8388608:case 16777216:case 33554432:return e&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return e}}function qn(e,l){var t=e.pendingLanes;if(t===0)return 0;var a=0,n=e.suspendedLanes,u=e.pingedLanes,i=e.warmLanes;e=e.finishedLanes!==0;var c=t&134217727;return c!==0?(t=c&~n,t!==0?a=pt(t):(u&=c,u!==0?a=pt(u):e||(i=c&~i,i!==0&&(a=pt(i))))):(c=t&~n,c!==0?a=pt(c):u!==0?a=pt(u):e||(i=t&~i,i!==0&&(a=pt(i)))),a===0?0:l!==0&&l!==a&&!(l&n)&&(n=a&-a,i=l&-l,n>=i||n===32&&(i&4194176)!==0)?l:a}function Oa(e,l){return(e.pendingLanes&~(e.suspendedLanes&~e.pingedLanes)&l)===0}function Id(e,l){switch(e){case 1:case 2:case 4:case 8:return l+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return l+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function ds(){var e=Hn;return Hn<<=1,!(Hn&4194176)&&(Hn=128),e}function ms(){var e=Ln;return Ln<<=1,!(Ln&62914560)&&(Ln=4194304),e}function Wu(e){for(var l=[],t=0;31>t;t++)l.push(e);return l}function ja(e,l){e.pendingLanes|=l,l!==268435456&&(e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0)}function em(e,l,t,a,n,u){var i=e.pendingLanes;e.pendingLanes=t,e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0,e.expiredLanes&=t,e.entangledLanes&=t,e.errorRecoveryDisabledLanes&=t,e.shellSuspendCounter=0;var c=e.entanglements,s=e.expirationTimes,r=e.hiddenUpdates;for(t=i&~t;0<t;){var v=31-He(t),T=1<<v;c[v]=0,s[v]=-1;var p=r[v];if(p!==null)for(r[v]=null,v=0;v<p.length;v++){var y=p[v];y!==null&&(y.lane&=-536870913)}t&=~T}a!==0&&hs(e,a,0),u!==0&&n===0&&e.tag!==0&&(e.suspendedLanes|=u&~(i&~l))}function hs(e,l,t){e.pendingLanes|=l,e.suspendedLanes&=~l;var a=31-He(l);e.entangledLanes|=l,e.entanglements[a]=e.entanglements[a]|1073741824|t&4194218}function gs(e,l){var t=e.entangledLanes|=l;for(e=e.entanglements;t;){var a=31-He(t),n=1<<a;n&l|e[a]&l&&(e[a]|=l),t&=~n}}function ps(e){return e&=-e,2<e?8<e?e&134217727?32:268435456:8:2}function ys(){var e=V.p;return e!==0?e:(e=window.event,e===void 0?32:wr(e.type))}function lm(e,l){var t=V.p;try{return V.p=e,l()}finally{V.p=t}}var wl=Math.random().toString(36).slice(2),Me="__reactFiber$"+wl,je="__reactProps$"+wl,Vt="__reactContainer$"+wl,Pu="__reactEvents$"+wl,tm="__reactListeners$"+wl,am="__reactHandles$"+wl,vs="__reactResources$"+wl,Ra="__reactMarker$"+wl;function Iu(e){delete e[Me],delete e[je],delete e[Pu],delete e[tm],delete e[am]}function yt(e){var l=e[Me];if(l)return l;for(var t=e.parentNode;t;){if(l=t[Vt]||t[Me]){if(t=l.alternate,l.child!==null||t!==null&&t.child!==null)for(e=Er(e);e!==null;){if(t=e[Me])return t;e=Er(e)}return l}e=t,t=e.parentNode}return null}function Kt(e){if(e=e[Me]||e[Vt]){var l=e.tag;if(l===5||l===6||l===13||l===26||l===27||l===3)return e}return null}function Ua(e){var l=e.tag;if(l===5||l===26||l===27||l===6)return e.stateNode;throw Error(d(33))}function Zt(e){var l=e[vs];return l||(l=e[vs]={hoistableStyles:new Map,hoistableScripts:new Map}),l}function Se(e){e[Ra]=!0}var bs=new Set,Ss={};function vt(e,l){kt(e,l),kt(e+"Capture",l)}function kt(e,l){for(Ss[e]=l,e=0;e<l.length;e++)bs.add(l[e])}var bl=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),nm=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),Ts={},xs={};function um(e){return Ju.call(xs,e)?!0:Ju.call(Ts,e)?!1:nm.test(e)?xs[e]=!0:(Ts[e]=!0,!1)}function Bn(e,l,t){if(um(l))if(t===null)e.removeAttribute(l);else{switch(typeof t){case"undefined":case"function":case"symbol":e.removeAttribute(l);return;case"boolean":var a=l.toLowerCase().slice(0,5);if(a!=="data-"&&a!=="aria-"){e.removeAttribute(l);return}}e.setAttribute(l,""+t)}}function Gn(e,l,t){if(t===null)e.removeAttribute(l);else{switch(typeof t){case"undefined":case"function":case"symbol":case"boolean":e.removeAttribute(l);return}e.setAttribute(l,""+t)}}function Sl(e,l,t,a){if(a===null)e.removeAttribute(t);else{switch(typeof a){case"undefined":case"function":case"symbol":case"boolean":e.removeAttribute(t);return}e.setAttributeNS(l,t,""+a)}}function Qe(e){switch(typeof e){case"bigint":case"boolean":case"number":case"string":case"undefined":return e;case"object":return e;default:return""}}function As(e){var l=e.type;return(e=e.nodeName)&&e.toLowerCase()==="input"&&(l==="checkbox"||l==="radio")}function im(e){var l=As(e)?"checked":"value",t=Object.getOwnPropertyDescriptor(e.constructor.prototype,l),a=""+e[l];if(!e.hasOwnProperty(l)&&typeof t<"u"&&typeof t.get=="function"&&typeof t.set=="function"){var n=t.get,u=t.set;return Object.defineProperty(e,l,{configurable:!0,get:function(){return n.call(this)},set:function(i){a=""+i,u.call(this,i)}}),Object.defineProperty(e,l,{enumerable:t.enumerable}),{getValue:function(){return a},setValue:function(i){a=""+i},stopTracking:function(){e._valueTracker=null,delete e[l]}}}}function Yn(e){e._valueTracker||(e._valueTracker=im(e))}function Ds(e){if(!e)return!1;var l=e._valueTracker;if(!l)return!0;var t=l.getValue(),a="";return e&&(a=As(e)?e.checked?"true":"false":e.value),e=a,e!==t?(l.setValue(e),!0):!1}function wn(e){if(e=e||(typeof document<"u"?document:void 0),typeof e>"u")return null;try{return e.activeElement||e.body}catch{return e.body}}var cm=/[\n"\\]/g;function Ve(e){return e.replace(cm,function(l){return"\\"+l.charCodeAt(0).toString(16)+" "})}function ei(e,l,t,a,n,u,i,c){e.name="",i!=null&&typeof i!="function"&&typeof i!="symbol"&&typeof i!="boolean"?e.type=i:e.removeAttribute("type"),l!=null?i==="number"?(l===0&&e.value===""||e.value!=l)&&(e.value=""+Qe(l)):e.value!==""+Qe(l)&&(e.value=""+Qe(l)):i!=="submit"&&i!=="reset"||e.removeAttribute("value"),l!=null?li(e,i,Qe(l)):t!=null?li(e,i,Qe(t)):a!=null&&e.removeAttribute("value"),n==null&&u!=null&&(e.defaultChecked=!!u),n!=null&&(e.checked=n&&typeof n!="function"&&typeof n!="symbol"),c!=null&&typeof c!="function"&&typeof c!="symbol"&&typeof c!="boolean"?e.name=""+Qe(c):e.removeAttribute("name")}function Ns(e,l,t,a,n,u,i,c){if(u!=null&&typeof u!="function"&&typeof u!="symbol"&&typeof u!="boolean"&&(e.type=u),l!=null||t!=null){if(!(u!=="submit"&&u!=="reset"||l!=null))return;t=t!=null?""+Qe(t):"",l=l!=null?""+Qe(l):t,c||l===e.value||(e.value=l),e.defaultValue=l}a=a??n,a=typeof a!="function"&&typeof a!="symbol"&&!!a,e.checked=c?e.checked:!!a,e.defaultChecked=!!a,i!=null&&typeof i!="function"&&typeof i!="symbol"&&typeof i!="boolean"&&(e.name=i)}function li(e,l,t){l==="number"&&wn(e.ownerDocument)===e||e.defaultValue===""+t||(e.defaultValue=""+t)}function Jt(e,l,t,a){if(e=e.options,l){l={};for(var n=0;n<t.length;n++)l["$"+t[n]]=!0;for(t=0;t<e.length;t++)n=l.hasOwnProperty("$"+e[t].value),e[t].selected!==n&&(e[t].selected=n),n&&a&&(e[t].defaultSelected=!0)}else{for(t=""+Qe(t),l=null,n=0;n<e.length;n++){if(e[n].value===t){e[n].selected=!0,a&&(e[n].defaultSelected=!0);return}l!==null||e[n].disabled||(l=e[n])}l!==null&&(l.selected=!0)}}function Es(e,l,t){if(l!=null&&(l=""+Qe(l),l!==e.value&&(e.value=l),t==null)){e.defaultValue!==l&&(e.defaultValue=l);return}e.defaultValue=t!=null?""+Qe(t):""}function Ms(e,l,t,a){if(l==null){if(a!=null){if(t!=null)throw Error(d(92));if(vl(a)){if(1<a.length)throw Error(d(93));a=a[0]}t=a}t==null&&(t=""),l=t}t=Qe(l),e.defaultValue=t,a=e.textContent,a===t&&a!==""&&a!==null&&(e.value=a)}function Ft(e,l){if(l){var t=e.firstChild;if(t&&t===e.lastChild&&t.nodeType===3){t.nodeValue=l;return}}e.textContent=l}var sm=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function zs(e,l,t){var a=l.indexOf("--")===0;t==null||typeof t=="boolean"||t===""?a?e.setProperty(l,""):l==="float"?e.cssFloat="":e[l]="":a?e.setProperty(l,t):typeof t!="number"||t===0||sm.has(l)?l==="float"?e.cssFloat=t:e[l]=(""+t).trim():e[l]=t+"px"}function Cs(e,l,t){if(l!=null&&typeof l!="object")throw Error(d(62));if(e=e.style,t!=null){for(var a in t)!t.hasOwnProperty(a)||l!=null&&l.hasOwnProperty(a)||(a.indexOf("--")===0?e.setProperty(a,""):a==="float"?e.cssFloat="":e[a]="");for(var n in l)a=l[n],l.hasOwnProperty(n)&&t[n]!==a&&zs(e,n,a)}else for(var u in l)l.hasOwnProperty(u)&&zs(e,u,l[u])}function ti(e){if(e.indexOf("-")===-1)return!1;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var om=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),fm=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function Xn(e){return fm.test(""+e)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":e}var ai=null;function ni(e){return e=e.target||e.srcElement||window,e.correspondingUseElement&&(e=e.correspondingUseElement),e.nodeType===3?e.parentNode:e}var $t=null,Wt=null;function Os(e){var l=Kt(e);if(l&&(e=l.stateNode)){var t=e[je]||null;e:switch(e=l.stateNode,l.type){case"input":if(ei(e,t.value,t.defaultValue,t.defaultValue,t.checked,t.defaultChecked,t.type,t.name),l=t.name,t.type==="radio"&&l!=null){for(t=e;t.parentNode;)t=t.parentNode;for(t=t.querySelectorAll('input[name="'+Ve(""+l)+'"][type="radio"]'),l=0;l<t.length;l++){var a=t[l];if(a!==e&&a.form===e.form){var n=a[je]||null;if(!n)throw Error(d(90));ei(a,n.value,n.defaultValue,n.defaultValue,n.checked,n.defaultChecked,n.type,n.name)}}for(l=0;l<t.length;l++)a=t[l],a.form===e.form&&Ds(a)}break e;case"textarea":Es(e,t.value,t.defaultValue);break e;case"select":l=t.value,l!=null&&Jt(e,!!t.multiple,l,!1)}}}var ui=!1;function js(e,l,t){if(ui)return e(l,t);ui=!0;try{var a=e(l);return a}finally{if(ui=!1,($t!==null||Wt!==null)&&(Nu(),$t&&(l=$t,e=Wt,Wt=$t=null,Os(l),e)))for(l=0;l<e.length;l++)Os(e[l])}}function _a(e,l){var t=e.stateNode;if(t===null)return null;var a=t[je]||null;if(a===null)return null;t=a[l];e:switch(l){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(a=!a.disabled)||(e=e.type,a=!(e==="button"||e==="input"||e==="select"||e==="textarea")),e=!a;break e;default:e=!1}if(e)return null;if(t&&typeof t!="function")throw Error(d(231,l,typeof t));return t}var ii=!1;if(bl)try{var Ha={};Object.defineProperty(Ha,"passive",{get:function(){ii=!0}}),window.addEventListener("test",Ha,Ha),window.removeEventListener("test",Ha,Ha)}catch{ii=!1}var Xl=null,ci=null,Qn=null;function Rs(){if(Qn)return Qn;var e,l=ci,t=l.length,a,n="value"in Xl?Xl.value:Xl.textContent,u=n.length;for(e=0;e<t&&l[e]===n[e];e++);var i=t-e;for(a=1;a<=i&&l[t-a]===n[u-a];a++);return Qn=n.slice(e,1<a?1-a:void 0)}function Vn(e){var l=e.keyCode;return"charCode"in e?(e=e.charCode,e===0&&l===13&&(e=13)):e=l,e===10&&(e=13),32<=e||e===13?e:0}function Kn(){return!0}function Us(){return!1}function Re(e){function l(t,a,n,u,i){this._reactName=t,this._targetInst=n,this.type=a,this.nativeEvent=u,this.target=i,this.currentTarget=null;for(var c in e)e.hasOwnProperty(c)&&(t=e[c],this[c]=t?t(u):u[c]);return this.isDefaultPrevented=(u.defaultPrevented!=null?u.defaultPrevented:u.returnValue===!1)?Kn:Us,this.isPropagationStopped=Us,this}return k(l.prototype,{preventDefault:function(){this.defaultPrevented=!0;var t=this.nativeEvent;t&&(t.preventDefault?t.preventDefault():typeof t.returnValue!="unknown"&&(t.returnValue=!1),this.isDefaultPrevented=Kn)},stopPropagation:function(){var t=this.nativeEvent;t&&(t.stopPropagation?t.stopPropagation():typeof t.cancelBubble!="unknown"&&(t.cancelBubble=!0),this.isPropagationStopped=Kn)},persist:function(){},isPersistent:Kn}),l}var bt={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Zn=Re(bt),La=k({},bt,{view:0,detail:0}),rm=Re(La),si,oi,qa,kn=k({},La,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:ri,button:0,buttons:0,relatedTarget:function(e){return e.relatedTarget===void 0?e.fromElement===e.srcElement?e.toElement:e.fromElement:e.relatedTarget},movementX:function(e){return"movementX"in e?e.movementX:(e!==qa&&(qa&&e.type==="mousemove"?(si=e.screenX-qa.screenX,oi=e.screenY-qa.screenY):oi=si=0,qa=e),si)},movementY:function(e){return"movementY"in e?e.movementY:oi}}),_s=Re(kn),dm=k({},kn,{dataTransfer:0}),mm=Re(dm),hm=k({},La,{relatedTarget:0}),fi=Re(hm),gm=k({},bt,{animationName:0,elapsedTime:0,pseudoElement:0}),pm=Re(gm),ym=k({},bt,{clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}}),vm=Re(ym),bm=k({},bt,{data:0}),Hs=Re(bm),Sm={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Tm={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},xm={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function Am(e){var l=this.nativeEvent;return l.getModifierState?l.getModifierState(e):(e=xm[e])?!!l[e]:!1}function ri(){return Am}var Dm=k({},La,{key:function(e){if(e.key){var l=Sm[e.key]||e.key;if(l!=="Unidentified")return l}return e.type==="keypress"?(e=Vn(e),e===13?"Enter":String.fromCharCode(e)):e.type==="keydown"||e.type==="keyup"?Tm[e.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:ri,charCode:function(e){return e.type==="keypress"?Vn(e):0},keyCode:function(e){return e.type==="keydown"||e.type==="keyup"?e.keyCode:0},which:function(e){return e.type==="keypress"?Vn(e):e.type==="keydown"||e.type==="keyup"?e.keyCode:0}}),Nm=Re(Dm),Em=k({},kn,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),Ls=Re(Em),Mm=k({},La,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:ri}),zm=Re(Mm),Cm=k({},bt,{propertyName:0,elapsedTime:0,pseudoElement:0}),Om=Re(Cm),jm=k({},kn,{deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:0,deltaMode:0}),Rm=Re(jm),Um=k({},bt,{newState:0,oldState:0}),_m=Re(Um),Hm=[9,13,27,32],di=bl&&"CompositionEvent"in window,Ba=null;bl&&"documentMode"in document&&(Ba=document.documentMode);var Lm=bl&&"TextEvent"in window&&!Ba,qs=bl&&(!di||Ba&&8<Ba&&11>=Ba),Bs=" ",Gs=!1;function Ys(e,l){switch(e){case"keyup":return Hm.indexOf(l.keyCode)!==-1;case"keydown":return l.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function ws(e){return e=e.detail,typeof e=="object"&&"data"in e?e.data:null}var Pt=!1;function qm(e,l){switch(e){case"compositionend":return ws(l);case"keypress":return l.which!==32?null:(Gs=!0,Bs);case"textInput":return e=l.data,e===Bs&&Gs?null:e;default:return null}}function Bm(e,l){if(Pt)return e==="compositionend"||!di&&Ys(e,l)?(e=Rs(),Qn=ci=Xl=null,Pt=!1,e):null;switch(e){case"paste":return null;case"keypress":if(!(l.ctrlKey||l.altKey||l.metaKey)||l.ctrlKey&&l.altKey){if(l.char&&1<l.char.length)return l.char;if(l.which)return String.fromCharCode(l.which)}return null;case"compositionend":return qs&&l.locale!=="ko"?null:l.data;default:return null}}var Gm={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Xs(e){var l=e&&e.nodeName&&e.nodeName.toLowerCase();return l==="input"?!!Gm[e.type]:l==="textarea"}function Qs(e,l,t,a){$t?Wt?Wt.push(a):Wt=[a]:$t=a,l=Ou(l,"onChange"),0<l.length&&(t=new Zn("onChange","change",null,t,a),e.push({event:t,listeners:l}))}var Ga=null,Ya=null;function Ym(e){yr(e,0)}function Jn(e){var l=Ua(e);if(Ds(l))return e}function Vs(e,l){if(e==="change")return l}var Ks=!1;if(bl){var mi;if(bl){var hi="oninput"in document;if(!hi){var Zs=document.createElement("div");Zs.setAttribute("oninput","return;"),hi=typeof Zs.oninput=="function"}mi=hi}else mi=!1;Ks=mi&&(!document.documentMode||9<document.documentMode)}function ks(){Ga&&(Ga.detachEvent("onpropertychange",Js),Ya=Ga=null)}function Js(e){if(e.propertyName==="value"&&Jn(Ya)){var l=[];Qs(l,Ya,e,ni(e)),js(Ym,l)}}function wm(e,l,t){e==="focusin"?(ks(),Ga=l,Ya=t,Ga.attachEvent("onpropertychange",Js)):e==="focusout"&&ks()}function Xm(e){if(e==="selectionchange"||e==="keyup"||e==="keydown")return Jn(Ya)}function Qm(e,l){if(e==="click")return Jn(l)}function Vm(e,l){if(e==="input"||e==="change")return Jn(l)}function Km(e,l){return e===l&&(e!==0||1/e===1/l)||e!==e&&l!==l}var Le=typeof Object.is=="function"?Object.is:Km;function wa(e,l){if(Le(e,l))return!0;if(typeof e!="object"||e===null||typeof l!="object"||l===null)return!1;var t=Object.keys(e),a=Object.keys(l);if(t.length!==a.length)return!1;for(a=0;a<t.length;a++){var n=t[a];if(!Ju.call(l,n)||!Le(e[n],l[n]))return!1}return!0}function Fs(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function $s(e,l){var t=Fs(e);e=0;for(var a;t;){if(t.nodeType===3){if(a=e+t.textContent.length,e<=l&&a>=l)return{node:t,offset:l-e};e=a}e:{for(;t;){if(t.nextSibling){t=t.nextSibling;break e}t=t.parentNode}t=void 0}t=Fs(t)}}function Ws(e,l){return e&&l?e===l?!0:e&&e.nodeType===3?!1:l&&l.nodeType===3?Ws(e,l.parentNode):"contains"in e?e.contains(l):e.compareDocumentPosition?!!(e.compareDocumentPosition(l)&16):!1:!1}function Ps(e){e=e!=null&&e.ownerDocument!=null&&e.ownerDocument.defaultView!=null?e.ownerDocument.defaultView:window;for(var l=wn(e.document);l instanceof e.HTMLIFrameElement;){try{var t=typeof l.contentWindow.location.href=="string"}catch{t=!1}if(t)e=l.contentWindow;else break;l=wn(e.document)}return l}function gi(e){var l=e&&e.nodeName&&e.nodeName.toLowerCase();return l&&(l==="input"&&(e.type==="text"||e.type==="search"||e.type==="tel"||e.type==="url"||e.type==="password")||l==="textarea"||e.contentEditable==="true")}function Zm(e,l){var t=Ps(l);l=e.focusedElem;var a=e.selectionRange;if(t!==l&&l&&l.ownerDocument&&Ws(l.ownerDocument.documentElement,l)){if(a!==null&&gi(l)){if(e=a.start,t=a.end,t===void 0&&(t=e),"selectionStart"in l)l.selectionStart=e,l.selectionEnd=Math.min(t,l.value.length);else if(t=(e=l.ownerDocument||document)&&e.defaultView||window,t.getSelection){t=t.getSelection();var n=l.textContent.length,u=Math.min(a.start,n);a=a.end===void 0?u:Math.min(a.end,n),!t.extend&&u>a&&(n=a,a=u,u=n),n=$s(l,u);var i=$s(l,a);n&&i&&(t.rangeCount!==1||t.anchorNode!==n.node||t.anchorOffset!==n.offset||t.focusNode!==i.node||t.focusOffset!==i.offset)&&(e=e.createRange(),e.setStart(n.node,n.offset),t.removeAllRanges(),u>a?(t.addRange(e),t.extend(i.node,i.offset)):(e.setEnd(i.node,i.offset),t.addRange(e)))}}for(e=[],t=l;t=t.parentNode;)t.nodeType===1&&e.push({element:t,left:t.scrollLeft,top:t.scrollTop});for(typeof l.focus=="function"&&l.focus(),l=0;l<e.length;l++)t=e[l],t.element.scrollLeft=t.left,t.element.scrollTop=t.top}}var km=bl&&"documentMode"in document&&11>=document.documentMode,It=null,pi=null,Xa=null,yi=!1;function Is(e,l,t){var a=t.window===t?t.document:t.nodeType===9?t:t.ownerDocument;yi||It==null||It!==wn(a)||(a=It,"selectionStart"in a&&gi(a)?a={start:a.selectionStart,end:a.selectionEnd}:(a=(a.ownerDocument&&a.ownerDocument.defaultView||window).getSelection(),a={anchorNode:a.anchorNode,anchorOffset:a.anchorOffset,focusNode:a.focusNode,focusOffset:a.focusOffset}),Xa&&wa(Xa,a)||(Xa=a,a=Ou(pi,"onSelect"),0<a.length&&(l=new Zn("onSelect","select",null,l,t),e.push({event:l,listeners:a}),l.target=It)))}function St(e,l){var t={};return t[e.toLowerCase()]=l.toLowerCase(),t["Webkit"+e]="webkit"+l,t["Moz"+e]="moz"+l,t}var ea={animationend:St("Animation","AnimationEnd"),animationiteration:St("Animation","AnimationIteration"),animationstart:St("Animation","AnimationStart"),transitionrun:St("Transition","TransitionRun"),transitionstart:St("Transition","TransitionStart"),transitioncancel:St("Transition","TransitionCancel"),transitionend:St("Transition","TransitionEnd")},vi={},eo={};bl&&(eo=document.createElement("div").style,"AnimationEvent"in window||(delete ea.animationend.animation,delete ea.animationiteration.animation,delete ea.animationstart.animation),"TransitionEvent"in window||delete ea.transitionend.transition);function Tt(e){if(vi[e])return vi[e];if(!ea[e])return e;var l=ea[e],t;for(t in l)if(l.hasOwnProperty(t)&&t in eo)return vi[e]=l[t];return e}var lo=Tt("animationend"),to=Tt("animationiteration"),ao=Tt("animationstart"),Jm=Tt("transitionrun"),Fm=Tt("transitionstart"),$m=Tt("transitioncancel"),no=Tt("transitionend"),uo=new Map,io="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function ul(e,l){uo.set(e,l),vt(l,[e])}var Ke=[],la=0,bi=0;function Fn(){for(var e=la,l=bi=la=0;l<e;){var t=Ke[l];Ke[l++]=null;var a=Ke[l];Ke[l++]=null;var n=Ke[l];Ke[l++]=null;var u=Ke[l];if(Ke[l++]=null,a!==null&&n!==null){var i=a.pending;i===null?n.next=n:(n.next=i.next,i.next=n),a.pending=n}u!==0&&co(t,n,u)}}function $n(e,l,t,a){Ke[la++]=e,Ke[la++]=l,Ke[la++]=t,Ke[la++]=a,bi|=a,e.lanes|=a,e=e.alternate,e!==null&&(e.lanes|=a)}function Si(e,l,t,a){return $n(e,l,t,a),Wn(e)}function Ql(e,l){return $n(e,null,null,l),Wn(e)}function co(e,l,t){e.lanes|=t;var a=e.alternate;a!==null&&(a.lanes|=t);for(var n=!1,u=e.return;u!==null;)u.childLanes|=t,a=u.alternate,a!==null&&(a.childLanes|=t),u.tag===22&&(e=u.stateNode,e===null||e._visibility&1||(n=!0)),e=u,u=u.return;n&&l!==null&&e.tag===3&&(u=e.stateNode,n=31-He(t),u=u.hiddenUpdates,e=u[n],e===null?u[n]=[l]:e.push(l),l.lane=t|536870912)}function Wn(e){if(50<hn)throw hn=0,Ec=null,Error(d(185));for(var l=e.return;l!==null;)e=l,l=e.return;return e.tag===3?e.stateNode:null}var ta={},so=new WeakMap;function Ze(e,l){if(typeof e=="object"&&e!==null){var t=so.get(e);return t!==void 0?t:(l={value:e,source:l,stack:L(l)},so.set(e,l),l)}return{value:e,source:l,stack:L(l)}}var aa=[],na=0,Pn=null,In=0,ke=[],Je=0,xt=null,Tl=1,xl="";function At(e,l){aa[na++]=In,aa[na++]=Pn,Pn=e,In=l}function oo(e,l,t){ke[Je++]=Tl,ke[Je++]=xl,ke[Je++]=xt,xt=e;var a=Tl;e=xl;var n=32-He(a)-1;a&=~(1<<n),t+=1;var u=32-He(l)+n;if(30<u){var i=n-n%5;u=(a&(1<<i)-1).toString(32),a>>=i,n-=i,Tl=1<<32-He(l)+n|t<<n|a,xl=u+e}else Tl=1<<u|t<<n|a,xl=e}function Ti(e){e.return!==null&&(At(e,1),oo(e,1,0))}function xi(e){for(;e===Pn;)Pn=aa[--na],aa[na]=null,In=aa[--na],aa[na]=null;for(;e===xt;)xt=ke[--Je],ke[Je]=null,xl=ke[--Je],ke[Je]=null,Tl=ke[--Je],ke[Je]=null}var Ce=null,De=null,K=!1,il=null,dl=!1,Ai=Error(d(519));function Dt(e){var l=Error(d(418,""));throw Ka(Ze(l,e)),Ai}function fo(e){var l=e.stateNode,t=e.type,a=e.memoizedProps;switch(l[Me]=e,l[je]=a,t){case"dialog":w("cancel",l),w("close",l);break;case"iframe":case"object":case"embed":w("load",l);break;case"video":case"audio":for(t=0;t<pn.length;t++)w(pn[t],l);break;case"source":w("error",l);break;case"img":case"image":case"link":w("error",l),w("load",l);break;case"details":w("toggle",l);break;case"input":w("invalid",l),Ns(l,a.value,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name,!0),Yn(l);break;case"select":w("invalid",l);break;case"textarea":w("invalid",l),Ms(l,a.value,a.defaultValue,a.children),Yn(l)}t=a.children,typeof t!="string"&&typeof t!="number"&&typeof t!="bigint"||l.textContent===""+t||a.suppressHydrationWarning===!0||Tr(l.textContent,t)?(a.popover!=null&&(w("beforetoggle",l),w("toggle",l)),a.onScroll!=null&&w("scroll",l),a.onScrollEnd!=null&&w("scrollend",l),a.onClick!=null&&(l.onclick=ju),l=!0):l=!1,l||Dt(e)}function ro(e){for(Ce=e.return;Ce;)switch(Ce.tag){case 3:case 27:dl=!0;return;case 5:case 13:dl=!1;return;default:Ce=Ce.return}}function Qa(e){if(e!==Ce)return!1;if(!K)return ro(e),K=!0,!1;var l=!1,t;if((t=e.tag!==3&&e.tag!==27)&&((t=e.tag===5)&&(t=e.type,t=!(t!=="form"&&t!=="button")||Qc(e.type,e.memoizedProps)),t=!t),t&&(l=!0),l&&De&&Dt(e),ro(e),e.tag===13){if(e=e.memoizedState,e=e!==null?e.dehydrated:null,!e)throw Error(d(317));e:{for(e=e.nextSibling,l=0;e;){if(e.nodeType===8)if(t=e.data,t==="/$"){if(l===0){De=sl(e.nextSibling);break e}l--}else t!=="$"&&t!=="$!"&&t!=="$?"||l++;e=e.nextSibling}De=null}}else De=Ce?sl(e.stateNode.nextSibling):null;return!0}function Va(){De=Ce=null,K=!1}function Ka(e){il===null?il=[e]:il.push(e)}var Za=Error(d(460)),mo=Error(d(474)),Di={then:function(){}};function ho(e){return e=e.status,e==="fulfilled"||e==="rejected"}function eu(){}function go(e,l,t){switch(t=e[t],t===void 0?e.push(l):t!==l&&(l.then(eu,eu),l=t),l.status){case"fulfilled":return l.value;case"rejected":throw e=l.reason,e===Za?Error(d(483)):e;default:if(typeof l.status=="string")l.then(eu,eu);else{if(e=I,e!==null&&100<e.shellSuspendCounter)throw Error(d(482));e=l,e.status="pending",e.then(function(a){if(l.status==="pending"){var n=l;n.status="fulfilled",n.value=a}},function(a){if(l.status==="pending"){var n=l;n.status="rejected",n.reason=a}})}switch(l.status){case"fulfilled":return l.value;case"rejected":throw e=l.reason,e===Za?Error(d(483)):e}throw ka=l,Za}}var ka=null;function po(){if(ka===null)throw Error(d(459));var e=ka;return ka=null,e}var ua=null,Ja=0;function lu(e){var l=Ja;return Ja+=1,ua===null&&(ua=[]),go(ua,e,l)}function Fa(e,l){l=l.props.ref,e.ref=l!==void 0?l:null}function tu(e,l){throw l.$$typeof===j?Error(d(525)):(e=Object.prototype.toString.call(l),Error(d(31,e==="[object Object]"?"object with keys {"+Object.keys(l).join(", ")+"}":e)))}function yo(e){var l=e._init;return l(e._payload)}function vo(e){function l(m,f){if(e){var h=m.deletions;h===null?(m.deletions=[f],m.flags|=16):h.push(f)}}function t(m,f){if(!e)return null;for(;f!==null;)l(m,f),f=f.sibling;return null}function a(m){for(var f=new Map;m!==null;)m.key!==null?f.set(m.key,m):f.set(m.index,m),m=m.sibling;return f}function n(m,f){return m=lt(m,f),m.index=0,m.sibling=null,m}function u(m,f,h){return m.index=h,e?(h=m.alternate,h!==null?(h=h.index,h<f?(m.flags|=33554434,f):h):(m.flags|=33554434,f)):(m.flags|=1048576,f)}function i(m){return e&&m.alternate===null&&(m.flags|=33554434),m}function c(m,f,h,S){return f===null||f.tag!==6?(f=vc(h,m.mode,S),f.return=m,f):(f=n(f,h),f.return=m,f)}function s(m,f,h,S){var D=h.type;return D===Y?v(m,f,h.props.children,S,h.key):f!==null&&(f.elementType===D||typeof D=="object"&&D!==null&&D.$$typeof===al&&yo(D)===f.type)?(f=n(f,h.props),Fa(f,h),f.return=m,f):(f=Su(h.type,h.key,h.props,null,m.mode,S),Fa(f,h),f.return=m,f)}function r(m,f,h,S){return f===null||f.tag!==4||f.stateNode.containerInfo!==h.containerInfo||f.stateNode.implementation!==h.implementation?(f=bc(h,m.mode,S),f.return=m,f):(f=n(f,h.children||[]),f.return=m,f)}function v(m,f,h,S,D){return f===null||f.tag!==7?(f=_t(h,m.mode,S,D),f.return=m,f):(f=n(f,h),f.return=m,f)}function T(m,f,h){if(typeof f=="string"&&f!==""||typeof f=="number"||typeof f=="bigint")return f=vc(""+f,m.mode,h),f.return=m,f;if(typeof f=="object"&&f!==null){switch(f.$$typeof){case H:return h=Su(f.type,f.key,f.props,null,m.mode,h),Fa(h,f),h.return=m,h;case $:return f=bc(f,m.mode,h),f.return=m,f;case al:var S=f._init;return f=S(f._payload),T(m,f,h)}if(vl(f)||nl(f))return f=_t(f,m.mode,h,null),f.return=m,f;if(typeof f.then=="function")return T(m,lu(f),h);if(f.$$typeof===pe)return T(m,yu(m,f),h);tu(m,f)}return null}function p(m,f,h,S){var D=f!==null?f.key:null;if(typeof h=="string"&&h!==""||typeof h=="number"||typeof h=="bigint")return D!==null?null:c(m,f,""+h,S);if(typeof h=="object"&&h!==null){switch(h.$$typeof){case H:return h.key===D?s(m,f,h,S):null;case $:return h.key===D?r(m,f,h,S):null;case al:return D=h._init,h=D(h._payload),p(m,f,h,S)}if(vl(h)||nl(h))return D!==null?null:v(m,f,h,S,null);if(typeof h.then=="function")return p(m,f,lu(h),S);if(h.$$typeof===pe)return p(m,f,yu(m,h),S);tu(m,h)}return null}function y(m,f,h,S,D){if(typeof S=="string"&&S!==""||typeof S=="number"||typeof S=="bigint")return m=m.get(h)||null,c(f,m,""+S,D);if(typeof S=="object"&&S!==null){switch(S.$$typeof){case H:return m=m.get(S.key===null?h:S.key)||null,s(f,m,S,D);case $:return m=m.get(S.key===null?h:S.key)||null,r(f,m,S,D);case al:var B=S._init;return S=B(S._payload),y(m,f,h,S,D)}if(vl(S)||nl(S))return m=m.get(h)||null,v(f,m,S,D,null);if(typeof S.then=="function")return y(m,f,h,lu(S),D);if(S.$$typeof===pe)return y(m,f,h,yu(f,S),D);tu(f,S)}return null}function M(m,f,h,S){for(var D=null,B=null,z=f,O=f=0,Ae=null;z!==null&&O<h.length;O++){z.index>O?(Ae=z,z=null):Ae=z.sibling;var Z=p(m,z,h[O],S);if(Z===null){z===null&&(z=Ae);break}e&&z&&Z.alternate===null&&l(m,z),f=u(Z,f,O),B===null?D=Z:B.sibling=Z,B=Z,z=Ae}if(O===h.length)return t(m,z),K&&At(m,O),D;if(z===null){for(;O<h.length;O++)z=T(m,h[O],S),z!==null&&(f=u(z,f,O),B===null?D=z:B.sibling=z,B=z);return K&&At(m,O),D}for(z=a(z);O<h.length;O++)Ae=y(z,m,O,h[O],S),Ae!==null&&(e&&Ae.alternate!==null&&z.delete(Ae.key===null?O:Ae.key),f=u(Ae,f,O),B===null?D=Ae:B.sibling=Ae,B=Ae);return e&&z.forEach(function(st){return l(m,st)}),K&&At(m,O),D}function _(m,f,h,S){if(h==null)throw Error(d(151));for(var D=null,B=null,z=f,O=f=0,Ae=null,Z=h.next();z!==null&&!Z.done;O++,Z=h.next()){z.index>O?(Ae=z,z=null):Ae=z.sibling;var st=p(m,z,Z.value,S);if(st===null){z===null&&(z=Ae);break}e&&z&&st.alternate===null&&l(m,z),f=u(st,f,O),B===null?D=st:B.sibling=st,B=st,z=Ae}if(Z.done)return t(m,z),K&&At(m,O),D;if(z===null){for(;!Z.done;O++,Z=h.next())Z=T(m,Z.value,S),Z!==null&&(f=u(Z,f,O),B===null?D=Z:B.sibling=Z,B=Z);return K&&At(m,O),D}for(z=a(z);!Z.done;O++,Z=h.next())Z=y(z,m,O,Z.value,S),Z!==null&&(e&&Z.alternate!==null&&z.delete(Z.key===null?O:Z.key),f=u(Z,f,O),B===null?D=Z:B.sibling=Z,B=Z);return e&&z.forEach(function(fg){return l(m,fg)}),K&&At(m,O),D}function se(m,f,h,S){if(typeof h=="object"&&h!==null&&h.type===Y&&h.key===null&&(h=h.props.children),typeof h=="object"&&h!==null){switch(h.$$typeof){case H:e:{for(var D=h.key;f!==null;){if(f.key===D){if(D=h.type,D===Y){if(f.tag===7){t(m,f.sibling),S=n(f,h.props.children),S.return=m,m=S;break e}}else if(f.elementType===D||typeof D=="object"&&D!==null&&D.$$typeof===al&&yo(D)===f.type){t(m,f.sibling),S=n(f,h.props),Fa(S,h),S.return=m,m=S;break e}t(m,f);break}else l(m,f);f=f.sibling}h.type===Y?(S=_t(h.props.children,m.mode,S,h.key),S.return=m,m=S):(S=Su(h.type,h.key,h.props,null,m.mode,S),Fa(S,h),S.return=m,m=S)}return i(m);case $:e:{for(D=h.key;f!==null;){if(f.key===D)if(f.tag===4&&f.stateNode.containerInfo===h.containerInfo&&f.stateNode.implementation===h.implementation){t(m,f.sibling),S=n(f,h.children||[]),S.return=m,m=S;break e}else{t(m,f);break}else l(m,f);f=f.sibling}S=bc(h,m.mode,S),S.return=m,m=S}return i(m);case al:return D=h._init,h=D(h._payload),se(m,f,h,S)}if(vl(h))return M(m,f,h,S);if(nl(h)){if(D=nl(h),typeof D!="function")throw Error(d(150));return h=D.call(h),_(m,f,h,S)}if(typeof h.then=="function")return se(m,f,lu(h),S);if(h.$$typeof===pe)return se(m,f,yu(m,h),S);tu(m,h)}return typeof h=="string"&&h!==""||typeof h=="number"||typeof h=="bigint"?(h=""+h,f!==null&&f.tag===6?(t(m,f.sibling),S=n(f,h),S.return=m,m=S):(t(m,f),S=vc(h,m.mode,S),S.return=m,m=S),i(m)):t(m,f)}return function(m,f,h,S){try{Ja=0;var D=se(m,f,h,S);return ua=null,D}catch(z){if(z===Za)throw z;var B=Pe(29,z,null,m.mode);return B.lanes=S,B.return=m,B}finally{}}}var Nt=vo(!0),bo=vo(!1),ia=ol(null),au=ol(0);function So(e,l){e=Ul,ae(au,e),ae(ia,l),Ul=e|l.baseLanes}function Ni(){ae(au,Ul),ae(ia,ia.current)}function Ei(){Ul=au.current,be(ia),be(au)}var Fe=ol(null),ml=null;function Vl(e){var l=e.alternate;ae(ye,ye.current&1),ae(Fe,e),ml===null&&(l===null||ia.current!==null||l.memoizedState!==null)&&(ml=e)}function To(e){if(e.tag===22){if(ae(ye,ye.current),ae(Fe,e),ml===null){var l=e.alternate;l!==null&&l.memoizedState!==null&&(ml=e)}}else Kl()}function Kl(){ae(ye,ye.current),ae(Fe,Fe.current)}function Al(e){be(Fe),ml===e&&(ml=null),be(ye)}var ye=ol(0);function nu(e){for(var l=e;l!==null;){if(l.tag===13){var t=l.memoizedState;if(t!==null&&(t=t.dehydrated,t===null||t.data==="$?"||t.data==="$!"))return l}else if(l.tag===19&&l.memoizedProps.revealOrder!==void 0){if(l.flags&128)return l}else if(l.child!==null){l.child.return=l,l=l.child;continue}if(l===e)break;for(;l.sibling===null;){if(l.return===null||l.return===e)return null;l=l.return}l.sibling.return=l.return,l=l.sibling}return null}var Wm=typeof AbortController<"u"?AbortController:function(){var e=[],l=this.signal={aborted:!1,addEventListener:function(t,a){e.push(a)}};this.abort=function(){l.aborted=!0,e.forEach(function(t){return t()})}},Pm=g.unstable_scheduleCallback,Im=g.unstable_NormalPriority,ve={$$typeof:pe,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Mi(){return{controller:new Wm,data:new Map,refCount:0}}function $a(e){e.refCount--,e.refCount===0&&Pm(Im,function(){e.controller.abort()})}var Wa=null,zi=0,ca=0,sa=null;function eh(e,l){if(Wa===null){var t=Wa=[];zi=0,ca=_c(),sa={status:"pending",value:void 0,then:function(a){t.push(a)}}}return zi++,l.then(xo,xo),l}function xo(){if(--zi===0&&Wa!==null){sa!==null&&(sa.status="fulfilled");var e=Wa;Wa=null,ca=0,sa=null;for(var l=0;l<e.length;l++)(0,e[l])()}}function lh(e,l){var t=[],a={status:"pending",value:null,reason:null,then:function(n){t.push(n)}};return e.then(function(){a.status="fulfilled",a.value=l;for(var n=0;n<t.length;n++)(0,t[n])(l)},function(n){for(a.status="rejected",a.reason=n,n=0;n<t.length;n++)(0,t[n])(void 0)}),a}var Ao=R.S;R.S=function(e,l){typeof l=="object"&&l!==null&&typeof l.then=="function"&&eh(e,l),Ao!==null&&Ao(e,l)};var Et=ol(null);function Ci(){var e=Et.current;return e!==null?e:I.pooledCache}function uu(e,l){l===null?ae(Et,Et.current):ae(Et,l.pool)}function Do(){var e=Ci();return e===null?null:{parent:ve._currentValue,pool:e}}var Zl=0,q=null,J=null,re=null,iu=!1,oa=!1,Mt=!1,cu=0,Pa=0,fa=null,th=0;function fe(){throw Error(d(321))}function Oi(e,l){if(l===null)return!1;for(var t=0;t<l.length&&t<e.length;t++)if(!Le(e[t],l[t]))return!1;return!0}function ji(e,l,t,a,n,u){return Zl=u,q=l,l.memoizedState=null,l.updateQueue=null,l.lanes=0,R.H=e===null||e.memoizedState===null?zt:kl,Mt=!1,u=t(a,n),Mt=!1,oa&&(u=Eo(l,t,a,n)),No(e),u}function No(e){R.H=hl;var l=J!==null&&J.next!==null;if(Zl=0,re=J=q=null,iu=!1,Pa=0,fa=null,l)throw Error(d(300));e===null||Te||(e=e.dependencies,e!==null&&pu(e)&&(Te=!0))}function Eo(e,l,t,a){q=e;var n=0;do{if(oa&&(fa=null),Pa=0,oa=!1,25<=n)throw Error(d(301));if(n+=1,re=J=null,e.updateQueue!=null){var u=e.updateQueue;u.lastEffect=null,u.events=null,u.stores=null,u.memoCache!=null&&(u.memoCache.index=0)}R.H=Ct,u=l(t,a)}while(oa);return u}function ah(){var e=R.H,l=e.useState()[0];return l=typeof l.then=="function"?Ia(l):l,e=e.useState()[0],(J!==null?J.memoizedState:null)!==e&&(q.flags|=1024),l}function Ri(){var e=cu!==0;return cu=0,e}function Ui(e,l,t){l.updateQueue=e.updateQueue,l.flags&=-2053,e.lanes&=~t}function _i(e){if(iu){for(e=e.memoizedState;e!==null;){var l=e.queue;l!==null&&(l.pending=null),e=e.next}iu=!1}Zl=0,re=J=q=null,oa=!1,Pa=cu=0,fa=null}function Ue(){var e={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return re===null?q.memoizedState=re=e:re=re.next=e,re}function de(){if(J===null){var e=q.alternate;e=e!==null?e.memoizedState:null}else e=J.next;var l=re===null?q.memoizedState:re.next;if(l!==null)re=l,J=e;else{if(e===null)throw q.alternate===null?Error(d(467)):Error(d(310));J=e,e={memoizedState:J.memoizedState,baseState:J.baseState,baseQueue:J.baseQueue,queue:J.queue,next:null},re===null?q.memoizedState=re=e:re=re.next=e}return re}var su;su=function(){return{lastEffect:null,events:null,stores:null,memoCache:null}};function Ia(e){var l=Pa;return Pa+=1,fa===null&&(fa=[]),e=go(fa,e,l),l=q,(re===null?l.memoizedState:re.next)===null&&(l=l.alternate,R.H=l===null||l.memoizedState===null?zt:kl),e}function ou(e){if(e!==null&&typeof e=="object"){if(typeof e.then=="function")return Ia(e);if(e.$$typeof===pe)return ze(e)}throw Error(d(438,String(e)))}function Hi(e){var l=null,t=q.updateQueue;if(t!==null&&(l=t.memoCache),l==null){var a=q.alternate;a!==null&&(a=a.updateQueue,a!==null&&(a=a.memoCache,a!=null&&(l={data:a.data.map(function(n){return n.slice()}),index:0})))}if(l==null&&(l={data:[],index:0}),t===null&&(t=su(),q.updateQueue=t),t.memoCache=l,t=l.data[l.index],t===void 0)for(t=l.data[l.index]=Array(e),a=0;a<e;a++)t[a]=Ea;return l.index++,t}function Dl(e,l){return typeof l=="function"?l(e):l}function fu(e){var l=de();return Li(l,J,e)}function Li(e,l,t){var a=e.queue;if(a===null)throw Error(d(311));a.lastRenderedReducer=t;var n=e.baseQueue,u=a.pending;if(u!==null){if(n!==null){var i=n.next;n.next=u.next,u.next=i}l.baseQueue=n=u,a.pending=null}if(u=e.baseState,n===null)e.memoizedState=u;else{l=n.next;var c=i=null,s=null,r=l,v=!1;do{var T=r.lane&-536870913;if(T!==r.lane?(Q&T)===T:(Zl&T)===T){var p=r.revertLane;if(p===0)s!==null&&(s=s.next={lane:0,revertLane:0,action:r.action,hasEagerState:r.hasEagerState,eagerState:r.eagerState,next:null}),T===ca&&(v=!0);else if((Zl&p)===p){r=r.next,p===ca&&(v=!0);continue}else T={lane:0,revertLane:r.revertLane,action:r.action,hasEagerState:r.hasEagerState,eagerState:r.eagerState,next:null},s===null?(c=s=T,i=u):s=s.next=T,q.lanes|=p,tt|=p;T=r.action,Mt&&t(u,T),u=r.hasEagerState?r.eagerState:t(u,T)}else p={lane:T,revertLane:r.revertLane,action:r.action,hasEagerState:r.hasEagerState,eagerState:r.eagerState,next:null},s===null?(c=s=p,i=u):s=s.next=p,q.lanes|=T,tt|=T;r=r.next}while(r!==null&&r!==l);if(s===null?i=u:s.next=c,!Le(u,e.memoizedState)&&(Te=!0,v&&(t=sa,t!==null)))throw t;e.memoizedState=u,e.baseState=i,e.baseQueue=s,a.lastRenderedState=u}return n===null&&(a.lanes=0),[e.memoizedState,a.dispatch]}function qi(e){var l=de(),t=l.queue;if(t===null)throw Error(d(311));t.lastRenderedReducer=e;var a=t.dispatch,n=t.pending,u=l.memoizedState;if(n!==null){t.pending=null;var i=n=n.next;do u=e(u,i.action),i=i.next;while(i!==n);Le(u,l.memoizedState)||(Te=!0),l.memoizedState=u,l.baseQueue===null&&(l.baseState=u),t.lastRenderedState=u}return[u,a]}function Mo(e,l,t){var a=q,n=de(),u=K;if(u){if(t===void 0)throw Error(d(407));t=t()}else t=l();var i=!Le((J||n).memoizedState,t);if(i&&(n.memoizedState=t,Te=!0),n=n.queue,Yi(Oo.bind(null,a,n,e),[e]),n.getSnapshot!==l||i||re!==null&&re.memoizedState.tag&1){if(a.flags|=2048,ra(9,Co.bind(null,a,n,t,l),{destroy:void 0},null),I===null)throw Error(d(349));u||Zl&60||zo(a,l,t)}return t}function zo(e,l,t){e.flags|=16384,e={getSnapshot:l,value:t},l=q.updateQueue,l===null?(l=su(),q.updateQueue=l,l.stores=[e]):(t=l.stores,t===null?l.stores=[e]:t.push(e))}function Co(e,l,t,a){l.value=t,l.getSnapshot=a,jo(l)&&Ro(e)}function Oo(e,l,t){return t(function(){jo(l)&&Ro(e)})}function jo(e){var l=e.getSnapshot;e=e.value;try{var t=l();return!Le(e,t)}catch{return!0}}function Ro(e){var l=Ql(e,2);l!==null&&Oe(l,e,2)}function Bi(e){var l=Ue();if(typeof e=="function"){var t=e;if(e=t(),Mt){Yl(!0);try{t()}finally{Yl(!1)}}}return l.memoizedState=l.baseState=e,l.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Dl,lastRenderedState:e},l}function Uo(e,l,t,a){return e.baseState=t,Li(e,J,typeof a=="function"?a:Dl)}function nh(e,l,t,a,n){if(mu(e))throw Error(d(485));if(e=l.action,e!==null){var u={payload:n,action:e,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(i){u.listeners.push(i)}};R.T!==null?t(!0):u.isTransition=!1,a(u),t=l.pending,t===null?(u.next=l.pending=u,_o(l,u)):(u.next=t.next,l.pending=t.next=u)}}function _o(e,l){var t=l.action,a=l.payload,n=e.state;if(l.isTransition){var u=R.T,i={};R.T=i;try{var c=t(n,a),s=R.S;s!==null&&s(i,c),Ho(e,l,c)}catch(r){Gi(e,l,r)}finally{R.T=u}}else try{u=t(n,a),Ho(e,l,u)}catch(r){Gi(e,l,r)}}function Ho(e,l,t){t!==null&&typeof t=="object"&&typeof t.then=="function"?t.then(function(a){Lo(e,l,a)},function(a){return Gi(e,l,a)}):Lo(e,l,t)}function Lo(e,l,t){l.status="fulfilled",l.value=t,qo(l),e.state=t,l=e.pending,l!==null&&(t=l.next,t===l?e.pending=null:(t=t.next,l.next=t,_o(e,t)))}function Gi(e,l,t){var a=e.pending;if(e.pending=null,a!==null){a=a.next;do l.status="rejected",l.reason=t,qo(l),l=l.next;while(l!==a)}e.action=null}function qo(e){e=e.listeners;for(var l=0;l<e.length;l++)(0,e[l])()}function Bo(e,l){return l}function Go(e,l){if(K){var t=I.formState;if(t!==null){e:{var a=q;if(K){if(De){l:{for(var n=De,u=dl;n.nodeType!==8;){if(!u){n=null;break l}if(n=sl(n.nextSibling),n===null){n=null;break l}}u=n.data,n=u==="F!"||u==="F"?n:null}if(n){De=sl(n.nextSibling),a=n.data==="F!";break e}}Dt(a)}a=!1}a&&(l=t[0])}}return t=Ue(),t.memoizedState=t.baseState=l,a={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Bo,lastRenderedState:l},t.queue=a,t=af.bind(null,q,a),a.dispatch=t,a=Bi(!1),u=Ki.bind(null,q,!1,a.queue),a=Ue(),n={state:l,dispatch:null,action:e,pending:null},a.queue=n,t=nh.bind(null,q,n,u,t),n.dispatch=t,a.memoizedState=e,[l,t,!1]}function Yo(e){var l=de();return wo(l,J,e)}function wo(e,l,t){l=Li(e,l,Bo)[0],e=fu(Dl)[0],l=typeof l=="object"&&l!==null&&typeof l.then=="function"?Ia(l):l;var a=de(),n=a.queue,u=n.dispatch;return t!==a.memoizedState&&(q.flags|=2048,ra(9,uh.bind(null,n,t),{destroy:void 0},null)),[l,u,e]}function uh(e,l){e.action=l}function Xo(e){var l=de(),t=J;if(t!==null)return wo(l,t,e);de(),l=l.memoizedState,t=de();var a=t.queue.dispatch;return t.memoizedState=e,[l,a,!1]}function ra(e,l,t,a){return e={tag:e,create:l,inst:t,deps:a,next:null},l=q.updateQueue,l===null&&(l=su(),q.updateQueue=l),t=l.lastEffect,t===null?l.lastEffect=e.next=e:(a=t.next,t.next=e,e.next=a,l.lastEffect=e),e}function Qo(){return de().memoizedState}function ru(e,l,t,a){var n=Ue();q.flags|=e,n.memoizedState=ra(1|l,t,{destroy:void 0},a===void 0?null:a)}function du(e,l,t,a){var n=de();a=a===void 0?null:a;var u=n.memoizedState.inst;J!==null&&a!==null&&Oi(a,J.memoizedState.deps)?n.memoizedState=ra(l,t,u,a):(q.flags|=e,n.memoizedState=ra(1|l,t,u,a))}function Vo(e,l){ru(8390656,8,e,l)}function Yi(e,l){du(2048,8,e,l)}function Ko(e,l){return du(4,2,e,l)}function Zo(e,l){return du(4,4,e,l)}function ko(e,l){if(typeof l=="function"){e=e();var t=l(e);return function(){typeof t=="function"?t():l(null)}}if(l!=null)return e=e(),l.current=e,function(){l.current=null}}function Jo(e,l,t){t=t!=null?t.concat([e]):null,du(4,4,ko.bind(null,l,e),t)}function wi(){}function Fo(e,l){var t=de();l=l===void 0?null:l;var a=t.memoizedState;return l!==null&&Oi(l,a[1])?a[0]:(t.memoizedState=[e,l],e)}function $o(e,l){var t=de();l=l===void 0?null:l;var a=t.memoizedState;if(l!==null&&Oi(l,a[1]))return a[0];if(a=e(),Mt){Yl(!0);try{e()}finally{Yl(!1)}}return t.memoizedState=[a,l],a}function Xi(e,l,t){return t===void 0||Zl&1073741824?e.memoizedState=l:(e.memoizedState=t,e=If(),q.lanes|=e,tt|=e,t)}function Wo(e,l,t,a){return Le(t,l)?t:ia.current!==null?(e=Xi(e,t,a),Le(e,l)||(Te=!0),e):Zl&42?(e=If(),q.lanes|=e,tt|=e,l):(Te=!0,e.memoizedState=t)}function Po(e,l,t,a,n){var u=V.p;V.p=u!==0&&8>u?u:8;var i=R.T,c={};R.T=c,Ki(e,!1,l,t);try{var s=n(),r=R.S;if(r!==null&&r(c,s),s!==null&&typeof s=="object"&&typeof s.then=="function"){var v=lh(s,a);en(e,l,v,Ye(e))}else en(e,l,a,Ye(e))}catch(T){en(e,l,{then:function(){},status:"rejected",reason:T},Ye())}finally{V.p=u,R.T=i}}function ih(){}function Qi(e,l,t,a){if(e.tag!==5)throw Error(d(476));var n=Io(e).queue;Po(e,n,l,Xe,t===null?ih:function(){return ef(e),t(a)})}function Io(e){var l=e.memoizedState;if(l!==null)return l;l={memoizedState:Xe,baseState:Xe,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Dl,lastRenderedState:Xe},next:null};var t={};return l.next={memoizedState:t,baseState:t,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Dl,lastRenderedState:t},next:null},e.memoizedState=l,e=e.alternate,e!==null&&(e.memoizedState=l),l}function ef(e){var l=Io(e).next.queue;en(e,l,{},Ye())}function Vi(){return ze(Tn)}function lf(){return de().memoizedState}function tf(){return de().memoizedState}function ch(e){for(var l=e.return;l!==null;){switch(l.tag){case 24:case 3:var t=Ye();e=$l(t);var a=Wl(l,e,t);a!==null&&(Oe(a,l,t),an(a,l,t)),l={cache:Mi()},e.payload=l;return}l=l.return}}function sh(e,l,t){var a=Ye();t={lane:a,revertLane:0,action:t,hasEagerState:!1,eagerState:null,next:null},mu(e)?nf(l,t):(t=Si(e,l,t,a),t!==null&&(Oe(t,e,a),uf(t,l,a)))}function af(e,l,t){var a=Ye();en(e,l,t,a)}function en(e,l,t,a){var n={lane:a,revertLane:0,action:t,hasEagerState:!1,eagerState:null,next:null};if(mu(e))nf(l,n);else{var u=e.alternate;if(e.lanes===0&&(u===null||u.lanes===0)&&(u=l.lastRenderedReducer,u!==null))try{var i=l.lastRenderedState,c=u(i,t);if(n.hasEagerState=!0,n.eagerState=c,Le(c,i))return $n(e,l,n,0),I===null&&Fn(),!1}catch{}finally{}if(t=Si(e,l,n,a),t!==null)return Oe(t,e,a),uf(t,l,a),!0}return!1}function Ki(e,l,t,a){if(a={lane:2,revertLane:_c(),action:a,hasEagerState:!1,eagerState:null,next:null},mu(e)){if(l)throw Error(d(479))}else l=Si(e,t,a,2),l!==null&&Oe(l,e,2)}function mu(e){var l=e.alternate;return e===q||l!==null&&l===q}function nf(e,l){oa=iu=!0;var t=e.pending;t===null?l.next=l:(l.next=t.next,t.next=l),e.pending=l}function uf(e,l,t){if(t&4194176){var a=l.lanes;a&=e.pendingLanes,t|=a,l.lanes=t,gs(e,t)}}var hl={readContext:ze,use:ou,useCallback:fe,useContext:fe,useEffect:fe,useImperativeHandle:fe,useLayoutEffect:fe,useInsertionEffect:fe,useMemo:fe,useReducer:fe,useRef:fe,useState:fe,useDebugValue:fe,useDeferredValue:fe,useTransition:fe,useSyncExternalStore:fe,useId:fe};hl.useCacheRefresh=fe,hl.useMemoCache=fe,hl.useHostTransitionStatus=fe,hl.useFormState=fe,hl.useActionState=fe,hl.useOptimistic=fe;var zt={readContext:ze,use:ou,useCallback:function(e,l){return Ue().memoizedState=[e,l===void 0?null:l],e},useContext:ze,useEffect:Vo,useImperativeHandle:function(e,l,t){t=t!=null?t.concat([e]):null,ru(4194308,4,ko.bind(null,l,e),t)},useLayoutEffect:function(e,l){return ru(4194308,4,e,l)},useInsertionEffect:function(e,l){ru(4,2,e,l)},useMemo:function(e,l){var t=Ue();l=l===void 0?null:l;var a=e();if(Mt){Yl(!0);try{e()}finally{Yl(!1)}}return t.memoizedState=[a,l],a},useReducer:function(e,l,t){var a=Ue();if(t!==void 0){var n=t(l);if(Mt){Yl(!0);try{t(l)}finally{Yl(!1)}}}else n=l;return a.memoizedState=a.baseState=n,e={pending:null,lanes:0,dispatch:null,lastRenderedReducer:e,lastRenderedState:n},a.queue=e,e=e.dispatch=sh.bind(null,q,e),[a.memoizedState,e]},useRef:function(e){var l=Ue();return e={current:e},l.memoizedState=e},useState:function(e){e=Bi(e);var l=e.queue,t=af.bind(null,q,l);return l.dispatch=t,[e.memoizedState,t]},useDebugValue:wi,useDeferredValue:function(e,l){var t=Ue();return Xi(t,e,l)},useTransition:function(){var e=Bi(!1);return e=Po.bind(null,q,e.queue,!0,!1),Ue().memoizedState=e,[!1,e]},useSyncExternalStore:function(e,l,t){var a=q,n=Ue();if(K){if(t===void 0)throw Error(d(407));t=t()}else{if(t=l(),I===null)throw Error(d(349));Q&60||zo(a,l,t)}n.memoizedState=t;var u={value:t,getSnapshot:l};return n.queue=u,Vo(Oo.bind(null,a,u,e),[e]),a.flags|=2048,ra(9,Co.bind(null,a,u,t,l),{destroy:void 0},null),t},useId:function(){var e=Ue(),l=I.identifierPrefix;if(K){var t=xl,a=Tl;t=(a&~(1<<32-He(a)-1)).toString(32)+t,l=":"+l+"R"+t,t=cu++,0<t&&(l+="H"+t.toString(32)),l+=":"}else t=th++,l=":"+l+"r"+t.toString(32)+":";return e.memoizedState=l},useCacheRefresh:function(){return Ue().memoizedState=ch.bind(null,q)}};zt.useMemoCache=Hi,zt.useHostTransitionStatus=Vi,zt.useFormState=Go,zt.useActionState=Go,zt.useOptimistic=function(e){var l=Ue();l.memoizedState=l.baseState=e;var t={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return l.queue=t,l=Ki.bind(null,q,!0,t),t.dispatch=l,[e,l]};var kl={readContext:ze,use:ou,useCallback:Fo,useContext:ze,useEffect:Yi,useImperativeHandle:Jo,useInsertionEffect:Ko,useLayoutEffect:Zo,useMemo:$o,useReducer:fu,useRef:Qo,useState:function(){return fu(Dl)},useDebugValue:wi,useDeferredValue:function(e,l){var t=de();return Wo(t,J.memoizedState,e,l)},useTransition:function(){var e=fu(Dl)[0],l=de().memoizedState;return[typeof e=="boolean"?e:Ia(e),l]},useSyncExternalStore:Mo,useId:lf};kl.useCacheRefresh=tf,kl.useMemoCache=Hi,kl.useHostTransitionStatus=Vi,kl.useFormState=Yo,kl.useActionState=Yo,kl.useOptimistic=function(e,l){var t=de();return Uo(t,J,e,l)};var Ct={readContext:ze,use:ou,useCallback:Fo,useContext:ze,useEffect:Yi,useImperativeHandle:Jo,useInsertionEffect:Ko,useLayoutEffect:Zo,useMemo:$o,useReducer:qi,useRef:Qo,useState:function(){return qi(Dl)},useDebugValue:wi,useDeferredValue:function(e,l){var t=de();return J===null?Xi(t,e,l):Wo(t,J.memoizedState,e,l)},useTransition:function(){var e=qi(Dl)[0],l=de().memoizedState;return[typeof e=="boolean"?e:Ia(e),l]},useSyncExternalStore:Mo,useId:lf};Ct.useCacheRefresh=tf,Ct.useMemoCache=Hi,Ct.useHostTransitionStatus=Vi,Ct.useFormState=Xo,Ct.useActionState=Xo,Ct.useOptimistic=function(e,l){var t=de();return J!==null?Uo(t,J,e,l):(t.baseState=e,[e,t.queue.dispatch])};function Zi(e,l,t,a){l=e.memoizedState,t=t(a,l),t=t==null?l:k({},l,t),e.memoizedState=t,e.lanes===0&&(e.updateQueue.baseState=t)}var ki={isMounted:function(e){return(e=e._reactInternals)?U(e)===e:!1},enqueueSetState:function(e,l,t){e=e._reactInternals;var a=Ye(),n=$l(a);n.payload=l,t!=null&&(n.callback=t),l=Wl(e,n,a),l!==null&&(Oe(l,e,a),an(l,e,a))},enqueueReplaceState:function(e,l,t){e=e._reactInternals;var a=Ye(),n=$l(a);n.tag=1,n.payload=l,t!=null&&(n.callback=t),l=Wl(e,n,a),l!==null&&(Oe(l,e,a),an(l,e,a))},enqueueForceUpdate:function(e,l){e=e._reactInternals;var t=Ye(),a=$l(t);a.tag=2,l!=null&&(a.callback=l),l=Wl(e,a,t),l!==null&&(Oe(l,e,t),an(l,e,t))}};function cf(e,l,t,a,n,u,i){return e=e.stateNode,typeof e.shouldComponentUpdate=="function"?e.shouldComponentUpdate(a,u,i):l.prototype&&l.prototype.isPureReactComponent?!wa(t,a)||!wa(n,u):!0}function sf(e,l,t,a){e=l.state,typeof l.componentWillReceiveProps=="function"&&l.componentWillReceiveProps(t,a),typeof l.UNSAFE_componentWillReceiveProps=="function"&&l.UNSAFE_componentWillReceiveProps(t,a),l.state!==e&&ki.enqueueReplaceState(l,l.state,null)}function Ot(e,l){var t=l;if("ref"in l){t={};for(var a in l)a!=="ref"&&(t[a]=l[a])}if(e=e.defaultProps){t===l&&(t=k({},t));for(var n in e)t[n]===void 0&&(t[n]=e[n])}return t}var hu=typeof reportError=="function"?reportError:function(e){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var l=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof e=="object"&&e!==null&&typeof e.message=="string"?String(e.message):String(e),error:e});if(!window.dispatchEvent(l))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",e);return}console.error(e)};function of(e){hu(e)}function ff(e){console.error(e)}function rf(e){hu(e)}function gu(e,l){try{var t=e.onUncaughtError;t(l.value,{componentStack:l.stack})}catch(a){setTimeout(function(){throw a})}}function df(e,l,t){try{var a=e.onCaughtError;a(t.value,{componentStack:t.stack,errorBoundary:l.tag===1?l.stateNode:null})}catch(n){setTimeout(function(){throw n})}}function Ji(e,l,t){return t=$l(t),t.tag=3,t.payload={element:null},t.callback=function(){gu(e,l)},t}function mf(e){return e=$l(e),e.tag=3,e}function hf(e,l,t,a){var n=t.type.getDerivedStateFromError;if(typeof n=="function"){var u=a.value;e.payload=function(){return n(u)},e.callback=function(){df(l,t,a)}}var i=t.stateNode;i!==null&&typeof i.componentDidCatch=="function"&&(e.callback=function(){df(l,t,a),typeof n!="function"&&(at===null?at=new Set([this]):at.add(this));var c=a.stack;this.componentDidCatch(a.value,{componentStack:c!==null?c:""})})}function oh(e,l,t,a,n){if(t.flags|=32768,a!==null&&typeof a=="object"&&typeof a.then=="function"){if(l=t.alternate,l!==null&&tn(l,t,n,!0),t=Fe.current,t!==null){switch(t.tag){case 13:return ml===null?Cc():t.alternate===null&&ce===0&&(ce=3),t.flags&=-257,t.flags|=65536,t.lanes=n,a===Di?t.flags|=16384:(l=t.updateQueue,l===null?t.updateQueue=new Set([a]):l.add(a),jc(e,a,n)),!1;case 22:return t.flags|=65536,a===Di?t.flags|=16384:(l=t.updateQueue,l===null?(l={transitions:null,markerInstances:null,retryQueue:new Set([a])},t.updateQueue=l):(t=l.retryQueue,t===null?l.retryQueue=new Set([a]):t.add(a)),jc(e,a,n)),!1}throw Error(d(435,t.tag))}return jc(e,a,n),Cc(),!1}if(K)return l=Fe.current,l!==null?(!(l.flags&65536)&&(l.flags|=256),l.flags|=65536,l.lanes=n,a!==Ai&&(e=Error(d(422),{cause:a}),Ka(Ze(e,t)))):(a!==Ai&&(l=Error(d(423),{cause:a}),Ka(Ze(l,t))),e=e.current.alternate,e.flags|=65536,n&=-n,e.lanes|=n,a=Ze(a,t),n=Ji(e.stateNode,a,n),oc(e,n),ce!==4&&(ce=2)),!1;var u=Error(d(520),{cause:a});if(u=Ze(u,t),dn===null?dn=[u]:dn.push(u),ce!==4&&(ce=2),l===null)return!0;a=Ze(a,t),t=l;do{switch(t.tag){case 3:return t.flags|=65536,e=n&-n,t.lanes|=e,e=Ji(t.stateNode,a,e),oc(t,e),!1;case 1:if(l=t.type,u=t.stateNode,(t.flags&128)===0&&(typeof l.getDerivedStateFromError=="function"||u!==null&&typeof u.componentDidCatch=="function"&&(at===null||!at.has(u))))return t.flags|=65536,n&=-n,t.lanes|=n,n=mf(n),hf(n,e,t,a),oc(t,n),!1}t=t.return}while(t!==null);return!1}var gf=Error(d(461)),Te=!1;function Ne(e,l,t,a){l.child=e===null?bo(l,null,t,a):Nt(l,e.child,t,a)}function pf(e,l,t,a,n){t=t.render;var u=l.ref;if("ref"in a){var i={};for(var c in a)c!=="ref"&&(i[c]=a[c])}else i=a;return Rt(l),a=ji(e,l,t,i,u,n),c=Ri(),e!==null&&!Te?(Ui(e,l,n),Nl(e,l,n)):(K&&c&&Ti(l),l.flags|=1,Ne(e,l,a,n),l.child)}function yf(e,l,t,a,n){if(e===null){var u=t.type;return typeof u=="function"&&!yc(u)&&u.defaultProps===void 0&&t.compare===null?(l.tag=15,l.type=u,vf(e,l,u,a,n)):(e=Su(t.type,null,a,l,l.mode,n),e.ref=l.ref,e.return=l,l.child=e)}if(u=e.child,!ac(e,n)){var i=u.memoizedProps;if(t=t.compare,t=t!==null?t:wa,t(i,a)&&e.ref===l.ref)return Nl(e,l,n)}return l.flags|=1,e=lt(u,a),e.ref=l.ref,e.return=l,l.child=e}function vf(e,l,t,a,n){if(e!==null){var u=e.memoizedProps;if(wa(u,a)&&e.ref===l.ref)if(Te=!1,l.pendingProps=a=u,ac(e,n))e.flags&131072&&(Te=!0);else return l.lanes=e.lanes,Nl(e,l,n)}return Fi(e,l,t,a,n)}function bf(e,l,t){var a=l.pendingProps,n=a.children,u=(l.stateNode._pendingVisibility&2)!==0,i=e!==null?e.memoizedState:null;if(ln(e,l),a.mode==="hidden"||u){if(l.flags&128){if(a=i!==null?i.baseLanes|t:t,e!==null){for(n=l.child=e.child,u=0;n!==null;)u=u|n.lanes|n.childLanes,n=n.sibling;l.childLanes=u&~a}else l.childLanes=0,l.child=null;return Sf(e,l,a,t)}if(t&536870912)l.memoizedState={baseLanes:0,cachePool:null},e!==null&&uu(l,i!==null?i.cachePool:null),i!==null?So(l,i):Ni(),To(l);else return l.lanes=l.childLanes=536870912,Sf(e,l,i!==null?i.baseLanes|t:t,t)}else i!==null?(uu(l,i.cachePool),So(l,i),Kl(),l.memoizedState=null):(e!==null&&uu(l,null),Ni(),Kl());return Ne(e,l,n,t),l.child}function Sf(e,l,t,a){var n=Ci();return n=n===null?null:{parent:ve._currentValue,pool:n},l.memoizedState={baseLanes:t,cachePool:n},e!==null&&uu(l,null),Ni(),To(l),e!==null&&tn(e,l,a,!0),null}function ln(e,l){var t=l.ref;if(t===null)e!==null&&e.ref!==null&&(l.flags|=2097664);else{if(typeof t!="function"&&typeof t!="object")throw Error(d(284));(e===null||e.ref!==t)&&(l.flags|=2097664)}}function Fi(e,l,t,a,n){return Rt(l),t=ji(e,l,t,a,void 0,n),a=Ri(),e!==null&&!Te?(Ui(e,l,n),Nl(e,l,n)):(K&&a&&Ti(l),l.flags|=1,Ne(e,l,t,n),l.child)}function Tf(e,l,t,a,n,u){return Rt(l),l.updateQueue=null,t=Eo(l,a,t,n),No(e),a=Ri(),e!==null&&!Te?(Ui(e,l,u),Nl(e,l,u)):(K&&a&&Ti(l),l.flags|=1,Ne(e,l,t,u),l.child)}function xf(e,l,t,a,n){if(Rt(l),l.stateNode===null){var u=ta,i=t.contextType;typeof i=="object"&&i!==null&&(u=ze(i)),u=new t(a,u),l.memoizedState=u.state!==null&&u.state!==void 0?u.state:null,u.updater=ki,l.stateNode=u,u._reactInternals=l,u=l.stateNode,u.props=a,u.state=l.memoizedState,u.refs={},cc(l),i=t.contextType,u.context=typeof i=="object"&&i!==null?ze(i):ta,u.state=l.memoizedState,i=t.getDerivedStateFromProps,typeof i=="function"&&(Zi(l,t,i,a),u.state=l.memoizedState),typeof t.getDerivedStateFromProps=="function"||typeof u.getSnapshotBeforeUpdate=="function"||typeof u.UNSAFE_componentWillMount!="function"&&typeof u.componentWillMount!="function"||(i=u.state,typeof u.componentWillMount=="function"&&u.componentWillMount(),typeof u.UNSAFE_componentWillMount=="function"&&u.UNSAFE_componentWillMount(),i!==u.state&&ki.enqueueReplaceState(u,u.state,null),un(l,a,u,n),nn(),u.state=l.memoizedState),typeof u.componentDidMount=="function"&&(l.flags|=4194308),a=!0}else if(e===null){u=l.stateNode;var c=l.memoizedProps,s=Ot(t,c);u.props=s;var r=u.context,v=t.contextType;i=ta,typeof v=="object"&&v!==null&&(i=ze(v));var T=t.getDerivedStateFromProps;v=typeof T=="function"||typeof u.getSnapshotBeforeUpdate=="function",c=l.pendingProps!==c,v||typeof u.UNSAFE_componentWillReceiveProps!="function"&&typeof u.componentWillReceiveProps!="function"||(c||r!==i)&&sf(l,u,a,i),Fl=!1;var p=l.memoizedState;u.state=p,un(l,a,u,n),nn(),r=l.memoizedState,c||p!==r||Fl?(typeof T=="function"&&(Zi(l,t,T,a),r=l.memoizedState),(s=Fl||cf(l,t,s,a,p,r,i))?(v||typeof u.UNSAFE_componentWillMount!="function"&&typeof u.componentWillMount!="function"||(typeof u.componentWillMount=="function"&&u.componentWillMount(),typeof u.UNSAFE_componentWillMount=="function"&&u.UNSAFE_componentWillMount()),typeof u.componentDidMount=="function"&&(l.flags|=4194308)):(typeof u.componentDidMount=="function"&&(l.flags|=4194308),l.memoizedProps=a,l.memoizedState=r),u.props=a,u.state=r,u.context=i,a=s):(typeof u.componentDidMount=="function"&&(l.flags|=4194308),a=!1)}else{u=l.stateNode,sc(e,l),i=l.memoizedProps,v=Ot(t,i),u.props=v,T=l.pendingProps,p=u.context,r=t.contextType,s=ta,typeof r=="object"&&r!==null&&(s=ze(r)),c=t.getDerivedStateFromProps,(r=typeof c=="function"||typeof u.getSnapshotBeforeUpdate=="function")||typeof u.UNSAFE_componentWillReceiveProps!="function"&&typeof u.componentWillReceiveProps!="function"||(i!==T||p!==s)&&sf(l,u,a,s),Fl=!1,p=l.memoizedState,u.state=p,un(l,a,u,n),nn();var y=l.memoizedState;i!==T||p!==y||Fl||e!==null&&e.dependencies!==null&&pu(e.dependencies)?(typeof c=="function"&&(Zi(l,t,c,a),y=l.memoizedState),(v=Fl||cf(l,t,v,a,p,y,s)||e!==null&&e.dependencies!==null&&pu(e.dependencies))?(r||typeof u.UNSAFE_componentWillUpdate!="function"&&typeof u.componentWillUpdate!="function"||(typeof u.componentWillUpdate=="function"&&u.componentWillUpdate(a,y,s),typeof u.UNSAFE_componentWillUpdate=="function"&&u.UNSAFE_componentWillUpdate(a,y,s)),typeof u.componentDidUpdate=="function"&&(l.flags|=4),typeof u.getSnapshotBeforeUpdate=="function"&&(l.flags|=1024)):(typeof u.componentDidUpdate!="function"||i===e.memoizedProps&&p===e.memoizedState||(l.flags|=4),typeof u.getSnapshotBeforeUpdate!="function"||i===e.memoizedProps&&p===e.memoizedState||(l.flags|=1024),l.memoizedProps=a,l.memoizedState=y),u.props=a,u.state=y,u.context=s,a=v):(typeof u.componentDidUpdate!="function"||i===e.memoizedProps&&p===e.memoizedState||(l.flags|=4),typeof u.getSnapshotBeforeUpdate!="function"||i===e.memoizedProps&&p===e.memoizedState||(l.flags|=1024),a=!1)}return u=a,ln(e,l),a=(l.flags&128)!==0,u||a?(u=l.stateNode,t=a&&typeof t.getDerivedStateFromError!="function"?null:u.render(),l.flags|=1,e!==null&&a?(l.child=Nt(l,e.child,null,n),l.child=Nt(l,null,t,n)):Ne(e,l,t,n),l.memoizedState=u.state,e=l.child):e=Nl(e,l,n),e}function Af(e,l,t,a){return Va(),l.flags|=256,Ne(e,l,t,a),l.child}var $i={dehydrated:null,treeContext:null,retryLane:0};function Wi(e){return{baseLanes:e,cachePool:Do()}}function Pi(e,l,t){return e=e!==null?e.childLanes&~t:0,l&&(e|=Ie),e}function Df(e,l,t){var a=l.pendingProps,n=!1,u=(l.flags&128)!==0,i;if((i=u)||(i=e!==null&&e.memoizedState===null?!1:(ye.current&2)!==0),i&&(n=!0,l.flags&=-129),i=(l.flags&32)!==0,l.flags&=-33,e===null){if(K){if(n?Vl(l):Kl(),K){var c=De,s;if(s=c){e:{for(s=c,c=dl;s.nodeType!==8;){if(!c){c=null;break e}if(s=sl(s.nextSibling),s===null){c=null;break e}}c=s}c!==null?(l.memoizedState={dehydrated:c,treeContext:xt!==null?{id:Tl,overflow:xl}:null,retryLane:536870912},s=Pe(18,null,null,0),s.stateNode=c,s.return=l,l.child=s,Ce=l,De=null,s=!0):s=!1}s||Dt(l)}if(c=l.memoizedState,c!==null&&(c=c.dehydrated,c!==null))return c.data==="$!"?l.lanes=16:l.lanes=536870912,null;Al(l)}return c=a.children,a=a.fallback,n?(Kl(),n=l.mode,c=ec({mode:"hidden",children:c},n),a=_t(a,n,t,null),c.return=l,a.return=l,c.sibling=a,l.child=c,n=l.child,n.memoizedState=Wi(t),n.childLanes=Pi(e,i,t),l.memoizedState=$i,a):(Vl(l),Ii(l,c))}if(s=e.memoizedState,s!==null&&(c=s.dehydrated,c!==null)){if(u)l.flags&256?(Vl(l),l.flags&=-257,l=lc(e,l,t)):l.memoizedState!==null?(Kl(),l.child=e.child,l.flags|=128,l=null):(Kl(),n=a.fallback,c=l.mode,a=ec({mode:"visible",children:a.children},c),n=_t(n,c,t,null),n.flags|=2,a.return=l,n.return=l,a.sibling=n,l.child=a,Nt(l,e.child,null,t),a=l.child,a.memoizedState=Wi(t),a.childLanes=Pi(e,i,t),l.memoizedState=$i,l=n);else if(Vl(l),c.data==="$!"){if(i=c.nextSibling&&c.nextSibling.dataset,i)var r=i.dgst;i=r,a=Error(d(419)),a.stack="",a.digest=i,Ka({value:a,source:null,stack:null}),l=lc(e,l,t)}else if(Te||tn(e,l,t,!1),i=(t&e.childLanes)!==0,Te||i){if(i=I,i!==null){if(a=t&-t,a&42)a=1;else switch(a){case 2:a=1;break;case 8:a=4;break;case 32:a=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:a=64;break;case 268435456:a=134217728;break;default:a=0}if(a=a&(i.suspendedLanes|t)?0:a,a!==0&&a!==s.retryLane)throw s.retryLane=a,Ql(e,a),Oe(i,e,a),gf}c.data==="$?"||Cc(),l=lc(e,l,t)}else c.data==="$?"?(l.flags|=128,l.child=e.child,l=Dh.bind(null,e),c._reactRetry=l,l=null):(e=s.treeContext,De=sl(c.nextSibling),Ce=l,K=!0,il=null,dl=!1,e!==null&&(ke[Je++]=Tl,ke[Je++]=xl,ke[Je++]=xt,Tl=e.id,xl=e.overflow,xt=l),l=Ii(l,a.children),l.flags|=4096);return l}return n?(Kl(),n=a.fallback,c=l.mode,s=e.child,r=s.sibling,a=lt(s,{mode:"hidden",children:a.children}),a.subtreeFlags=s.subtreeFlags&31457280,r!==null?n=lt(r,n):(n=_t(n,c,t,null),n.flags|=2),n.return=l,a.return=l,a.sibling=n,l.child=a,a=n,n=l.child,c=e.child.memoizedState,c===null?c=Wi(t):(s=c.cachePool,s!==null?(r=ve._currentValue,s=s.parent!==r?{parent:r,pool:r}:s):s=Do(),c={baseLanes:c.baseLanes|t,cachePool:s}),n.memoizedState=c,n.childLanes=Pi(e,i,t),l.memoizedState=$i,a):(Vl(l),t=e.child,e=t.sibling,t=lt(t,{mode:"visible",children:a.children}),t.return=l,t.sibling=null,e!==null&&(i=l.deletions,i===null?(l.deletions=[e],l.flags|=16):i.push(e)),l.child=t,l.memoizedState=null,t)}function Ii(e,l){return l=ec({mode:"visible",children:l},e.mode),l.return=e,e.child=l}function ec(e,l){return $f(e,l,0,null)}function lc(e,l,t){return Nt(l,e.child,null,t),e=Ii(l,l.pendingProps.children),e.flags|=2,l.memoizedState=null,e}function Nf(e,l,t){e.lanes|=l;var a=e.alternate;a!==null&&(a.lanes|=l),uc(e.return,l,t)}function tc(e,l,t,a,n){var u=e.memoizedState;u===null?e.memoizedState={isBackwards:l,rendering:null,renderingStartTime:0,last:a,tail:t,tailMode:n}:(u.isBackwards=l,u.rendering=null,u.renderingStartTime=0,u.last=a,u.tail=t,u.tailMode=n)}function Ef(e,l,t){var a=l.pendingProps,n=a.revealOrder,u=a.tail;if(Ne(e,l,a.children,t),a=ye.current,a&2)a=a&1|2,l.flags|=128;else{if(e!==null&&e.flags&128)e:for(e=l.child;e!==null;){if(e.tag===13)e.memoizedState!==null&&Nf(e,t,l);else if(e.tag===19)Nf(e,t,l);else if(e.child!==null){e.child.return=e,e=e.child;continue}if(e===l)break e;for(;e.sibling===null;){if(e.return===null||e.return===l)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}a&=1}switch(ae(ye,a),n){case"forwards":for(t=l.child,n=null;t!==null;)e=t.alternate,e!==null&&nu(e)===null&&(n=t),t=t.sibling;t=n,t===null?(n=l.child,l.child=null):(n=t.sibling,t.sibling=null),tc(l,!1,n,t,u);break;case"backwards":for(t=null,n=l.child,l.child=null;n!==null;){if(e=n.alternate,e!==null&&nu(e)===null){l.child=n;break}e=n.sibling,n.sibling=t,t=n,n=e}tc(l,!0,t,null,u);break;case"together":tc(l,!1,null,null,void 0);break;default:l.memoizedState=null}return l.child}function Nl(e,l,t){if(e!==null&&(l.dependencies=e.dependencies),tt|=l.lanes,!(t&l.childLanes))if(e!==null){if(tn(e,l,t,!1),(t&l.childLanes)===0)return null}else return null;if(e!==null&&l.child!==e.child)throw Error(d(153));if(l.child!==null){for(e=l.child,t=lt(e,e.pendingProps),l.child=t,t.return=l;e.sibling!==null;)e=e.sibling,t=t.sibling=lt(e,e.pendingProps),t.return=l;t.sibling=null}return l.child}function ac(e,l){return e.lanes&l?!0:(e=e.dependencies,!!(e!==null&&pu(e)))}function fh(e,l,t){switch(l.tag){case 3:Rn(l,l.stateNode.containerInfo),Jl(l,ve,e.memoizedState.cache),Va();break;case 27:case 5:ku(l);break;case 4:Rn(l,l.stateNode.containerInfo);break;case 10:Jl(l,l.type,l.memoizedProps.value);break;case 13:var a=l.memoizedState;if(a!==null)return a.dehydrated!==null?(Vl(l),l.flags|=128,null):t&l.child.childLanes?Df(e,l,t):(Vl(l),e=Nl(e,l,t),e!==null?e.sibling:null);Vl(l);break;case 19:var n=(e.flags&128)!==0;if(a=(t&l.childLanes)!==0,a||(tn(e,l,t,!1),a=(t&l.childLanes)!==0),n){if(a)return Ef(e,l,t);l.flags|=128}if(n=l.memoizedState,n!==null&&(n.rendering=null,n.tail=null,n.lastEffect=null),ae(ye,ye.current),a)break;return null;case 22:case 23:return l.lanes=0,bf(e,l,t);case 24:Jl(l,ve,e.memoizedState.cache)}return Nl(e,l,t)}function Mf(e,l,t){if(e!==null)if(e.memoizedProps!==l.pendingProps)Te=!0;else{if(!ac(e,t)&&!(l.flags&128))return Te=!1,fh(e,l,t);Te=!!(e.flags&131072)}else Te=!1,K&&l.flags&1048576&&oo(l,In,l.index);switch(l.lanes=0,l.tag){case 16:e:{e=l.pendingProps;var a=l.elementType,n=a._init;if(a=n(a._payload),l.type=a,typeof a=="function")yc(a)?(e=Ot(a,e),l.tag=1,l=xf(null,l,a,e,t)):(l.tag=0,l=Fi(null,l,a,e,t));else{if(a!=null){if(n=a.$$typeof,n===le){l.tag=11,l=pf(null,l,a,e,t);break e}else if(n===Gt){l.tag=14,l=yf(null,l,a,e,t);break e}}throw l=Yt(a)||a,Error(d(306,l,""))}}return l;case 0:return Fi(e,l,l.type,l.pendingProps,t);case 1:return a=l.type,n=Ot(a,l.pendingProps),xf(e,l,a,n,t);case 3:e:{if(Rn(l,l.stateNode.containerInfo),e===null)throw Error(d(387));var u=l.pendingProps;n=l.memoizedState,a=n.element,sc(e,l),un(l,u,null,t);var i=l.memoizedState;if(u=i.cache,Jl(l,ve,u),u!==n.cache&&ic(l,[ve],t,!0),nn(),u=i.element,n.isDehydrated)if(n={element:u,isDehydrated:!1,cache:i.cache},l.updateQueue.baseState=n,l.memoizedState=n,l.flags&256){l=Af(e,l,u,t);break e}else if(u!==a){a=Ze(Error(d(424)),l),Ka(a),l=Af(e,l,u,t);break e}else for(De=sl(l.stateNode.containerInfo.firstChild),Ce=l,K=!0,il=null,dl=!0,t=bo(l,null,u,t),l.child=t;t;)t.flags=t.flags&-3|4096,t=t.sibling;else{if(Va(),u===a){l=Nl(e,l,t);break e}Ne(e,l,u,t)}l=l.child}return l;case 26:return ln(e,l),e===null?(t=Or(l.type,null,l.pendingProps,null))?l.memoizedState=t:K||(t=l.type,e=l.pendingProps,a=Ru(Gl.current).createElement(t),a[Me]=l,a[je]=e,Ee(a,t,e),Se(a),l.stateNode=a):l.memoizedState=Or(l.type,e.memoizedProps,l.pendingProps,e.memoizedState),null;case 27:return ku(l),e===null&&K&&(a=l.stateNode=Mr(l.type,l.pendingProps,Gl.current),Ce=l,dl=!0,De=sl(a.firstChild)),a=l.pendingProps.children,e!==null||K?Ne(e,l,a,t):l.child=Nt(l,null,a,t),ln(e,l),l.child;case 5:return e===null&&K&&((n=a=De)&&(a=Yh(a,l.type,l.pendingProps,dl),a!==null?(l.stateNode=a,Ce=l,De=sl(a.firstChild),dl=!1,n=!0):n=!1),n||Dt(l)),ku(l),n=l.type,u=l.pendingProps,i=e!==null?e.memoizedProps:null,a=u.children,Qc(n,u)?a=null:i!==null&&Qc(n,i)&&(l.flags|=32),l.memoizedState!==null&&(n=ji(e,l,ah,null,null,t),Tn._currentValue=n),ln(e,l),Ne(e,l,a,t),l.child;case 6:return e===null&&K&&((e=t=De)&&(t=wh(t,l.pendingProps,dl),t!==null?(l.stateNode=t,Ce=l,De=null,e=!0):e=!1),e||Dt(l)),null;case 13:return Df(e,l,t);case 4:return Rn(l,l.stateNode.containerInfo),a=l.pendingProps,e===null?l.child=Nt(l,null,a,t):Ne(e,l,a,t),l.child;case 11:return pf(e,l,l.type,l.pendingProps,t);case 7:return Ne(e,l,l.pendingProps,t),l.child;case 8:return Ne(e,l,l.pendingProps.children,t),l.child;case 12:return Ne(e,l,l.pendingProps.children,t),l.child;case 10:return a=l.pendingProps,Jl(l,l.type,a.value),Ne(e,l,a.children,t),l.child;case 9:return n=l.type._context,a=l.pendingProps.children,Rt(l),n=ze(n),a=a(n),l.flags|=1,Ne(e,l,a,t),l.child;case 14:return yf(e,l,l.type,l.pendingProps,t);case 15:return vf(e,l,l.type,l.pendingProps,t);case 19:return Ef(e,l,t);case 22:return bf(e,l,t);case 24:return Rt(l),a=ze(ve),e===null?(n=Ci(),n===null&&(n=I,u=Mi(),n.pooledCache=u,u.refCount++,u!==null&&(n.pooledCacheLanes|=t),n=u),l.memoizedState={parent:a,cache:n},cc(l),Jl(l,ve,n)):(e.lanes&t&&(sc(e,l),un(l,null,null,t),nn()),n=e.memoizedState,u=l.memoizedState,n.parent!==a?(n={parent:a,cache:a},l.memoizedState=n,l.lanes===0&&(l.memoizedState=l.updateQueue.baseState=n),Jl(l,ve,a)):(a=u.cache,Jl(l,ve,a),a!==n.cache&&ic(l,[ve],t,!0))),Ne(e,l,l.pendingProps.children,t),l.child;case 29:throw l.pendingProps}throw Error(d(156,l.tag))}var nc=ol(null),jt=null,El=null;function Jl(e,l,t){ae(nc,l._currentValue),l._currentValue=t}function Ml(e){e._currentValue=nc.current,be(nc)}function uc(e,l,t){for(;e!==null;){var a=e.alternate;if((e.childLanes&l)!==l?(e.childLanes|=l,a!==null&&(a.childLanes|=l)):a!==null&&(a.childLanes&l)!==l&&(a.childLanes|=l),e===t)break;e=e.return}}function ic(e,l,t,a){var n=e.child;for(n!==null&&(n.return=e);n!==null;){var u=n.dependencies;if(u!==null){var i=n.child;u=u.firstContext;e:for(;u!==null;){var c=u;u=n;for(var s=0;s<l.length;s++)if(c.context===l[s]){u.lanes|=t,c=u.alternate,c!==null&&(c.lanes|=t),uc(u.return,t,e),a||(i=null);break e}u=c.next}}else if(n.tag===18){if(i=n.return,i===null)throw Error(d(341));i.lanes|=t,u=i.alternate,u!==null&&(u.lanes|=t),uc(i,t,e),i=null}else i=n.child;if(i!==null)i.return=n;else for(i=n;i!==null;){if(i===e){i=null;break}if(n=i.sibling,n!==null){n.return=i.return,i=n;break}i=i.return}n=i}}function tn(e,l,t,a){e=null;for(var n=l,u=!1;n!==null;){if(!u){if(n.flags&524288)u=!0;else if(n.flags&262144)break}if(n.tag===10){var i=n.alternate;if(i===null)throw Error(d(387));if(i=i.memoizedProps,i!==null){var c=n.type;Le(n.pendingProps.value,i.value)||(e!==null?e.push(c):e=[c])}}else if(n===jn.current){if(i=n.alternate,i===null)throw Error(d(387));i.memoizedState.memoizedState!==n.memoizedState.memoizedState&&(e!==null?e.push(Tn):e=[Tn])}n=n.return}e!==null&&ic(l,e,t,a),l.flags|=262144}function pu(e){for(e=e.firstContext;e!==null;){if(!Le(e.context._currentValue,e.memoizedValue))return!0;e=e.next}return!1}function Rt(e){jt=e,El=null,e=e.dependencies,e!==null&&(e.firstContext=null)}function ze(e){return zf(jt,e)}function yu(e,l){return jt===null&&Rt(e),zf(e,l)}function zf(e,l){var t=l._currentValue;if(l={context:l,memoizedValue:t,next:null},El===null){if(e===null)throw Error(d(308));El=l,e.dependencies={lanes:0,firstContext:l},e.flags|=524288}else El=El.next=l;return t}var Fl=!1;function cc(e){e.updateQueue={baseState:e.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function sc(e,l){e=e.updateQueue,l.updateQueue===e&&(l.updateQueue={baseState:e.baseState,firstBaseUpdate:e.firstBaseUpdate,lastBaseUpdate:e.lastBaseUpdate,shared:e.shared,callbacks:null})}function $l(e){return{lane:e,tag:0,payload:null,callback:null,next:null}}function Wl(e,l,t){var a=e.updateQueue;if(a===null)return null;if(a=a.shared,ue&2){var n=a.pending;return n===null?l.next=l:(l.next=n.next,n.next=l),a.pending=l,l=Wn(e),co(e,null,t),l}return $n(e,a,l,t),Wn(e)}function an(e,l,t){if(l=l.updateQueue,l!==null&&(l=l.shared,(t&4194176)!==0)){var a=l.lanes;a&=e.pendingLanes,t|=a,l.lanes=t,gs(e,t)}}function oc(e,l){var t=e.updateQueue,a=e.alternate;if(a!==null&&(a=a.updateQueue,t===a)){var n=null,u=null;if(t=t.firstBaseUpdate,t!==null){do{var i={lane:t.lane,tag:t.tag,payload:t.payload,callback:null,next:null};u===null?n=u=i:u=u.next=i,t=t.next}while(t!==null);u===null?n=u=l:u=u.next=l}else n=u=l;t={baseState:a.baseState,firstBaseUpdate:n,lastBaseUpdate:u,shared:a.shared,callbacks:a.callbacks},e.updateQueue=t;return}e=t.lastBaseUpdate,e===null?t.firstBaseUpdate=l:e.next=l,t.lastBaseUpdate=l}var fc=!1;function nn(){if(fc){var e=sa;if(e!==null)throw e}}function un(e,l,t,a){fc=!1;var n=e.updateQueue;Fl=!1;var u=n.firstBaseUpdate,i=n.lastBaseUpdate,c=n.shared.pending;if(c!==null){n.shared.pending=null;var s=c,r=s.next;s.next=null,i===null?u=r:i.next=r,i=s;var v=e.alternate;v!==null&&(v=v.updateQueue,c=v.lastBaseUpdate,c!==i&&(c===null?v.firstBaseUpdate=r:c.next=r,v.lastBaseUpdate=s))}if(u!==null){var T=n.baseState;i=0,v=r=s=null,c=u;do{var p=c.lane&-536870913,y=p!==c.lane;if(y?(Q&p)===p:(a&p)===p){p!==0&&p===ca&&(fc=!0),v!==null&&(v=v.next={lane:0,tag:c.tag,payload:c.payload,callback:null,next:null});e:{var M=e,_=c;p=l;var se=t;switch(_.tag){case 1:if(M=_.payload,typeof M=="function"){T=M.call(se,T,p);break e}T=M;break e;case 3:M.flags=M.flags&-65537|128;case 0:if(M=_.payload,p=typeof M=="function"?M.call(se,T,p):M,p==null)break e;T=k({},T,p);break e;case 2:Fl=!0}}p=c.callback,p!==null&&(e.flags|=64,y&&(e.flags|=8192),y=n.callbacks,y===null?n.callbacks=[p]:y.push(p))}else y={lane:p,tag:c.tag,payload:c.payload,callback:c.callback,next:null},v===null?(r=v=y,s=T):v=v.next=y,i|=p;if(c=c.next,c===null){if(c=n.shared.pending,c===null)break;y=c,c=y.next,y.next=null,n.lastBaseUpdate=y,n.shared.pending=null}}while(!0);v===null&&(s=T),n.baseState=s,n.firstBaseUpdate=r,n.lastBaseUpdate=v,u===null&&(n.shared.lanes=0),tt|=i,e.lanes=i,e.memoizedState=T}}function Cf(e,l){if(typeof e!="function")throw Error(d(191,e));e.call(l)}function Of(e,l){var t=e.callbacks;if(t!==null)for(e.callbacks=null,e=0;e<t.length;e++)Cf(t[e],l)}function cn(e,l){try{var t=l.updateQueue,a=t!==null?t.lastEffect:null;if(a!==null){var n=a.next;t=n;do{if((t.tag&e)===e){a=void 0;var u=t.create,i=t.inst;a=u(),i.destroy=a}t=t.next}while(t!==n)}}catch(c){P(l,l.return,c)}}function Pl(e,l,t){try{var a=l.updateQueue,n=a!==null?a.lastEffect:null;if(n!==null){var u=n.next;a=u;do{if((a.tag&e)===e){var i=a.inst,c=i.destroy;if(c!==void 0){i.destroy=void 0,n=l;var s=t;try{c()}catch(r){P(n,s,r)}}}a=a.next}while(a!==u)}}catch(r){P(l,l.return,r)}}function jf(e){var l=e.updateQueue;if(l!==null){var t=e.stateNode;try{Of(l,t)}catch(a){P(e,e.return,a)}}}function Rf(e,l,t){t.props=Ot(e.type,e.memoizedProps),t.state=e.memoizedState;try{t.componentWillUnmount()}catch(a){P(e,l,a)}}function Ut(e,l){try{var t=e.ref;if(t!==null){var a=e.stateNode;switch(e.tag){case 26:case 27:case 5:var n=a;break;default:n=a}typeof t=="function"?e.refCleanup=t(n):t.current=n}}catch(u){P(e,l,u)}}function qe(e,l){var t=e.ref,a=e.refCleanup;if(t!==null)if(typeof a=="function")try{a()}catch(n){P(e,l,n)}finally{e.refCleanup=null,e=e.alternate,e!=null&&(e.refCleanup=null)}else if(typeof t=="function")try{t(null)}catch(n){P(e,l,n)}else t.current=null}function Uf(e){var l=e.type,t=e.memoizedProps,a=e.stateNode;try{e:switch(l){case"button":case"input":case"select":case"textarea":t.autoFocus&&a.focus();break e;case"img":t.src?a.src=t.src:t.srcSet&&(a.srcset=t.srcSet)}}catch(n){P(e,e.return,n)}}function _f(e,l,t){try{var a=e.stateNode;Hh(a,e.type,t,l),a[je]=l}catch(n){P(e,e.return,n)}}function Hf(e){return e.tag===5||e.tag===3||e.tag===26||e.tag===27||e.tag===4}function rc(e){e:for(;;){for(;e.sibling===null;){if(e.return===null||Hf(e.return))return null;e=e.return}for(e.sibling.return=e.return,e=e.sibling;e.tag!==5&&e.tag!==6&&e.tag!==27&&e.tag!==18;){if(e.flags&2||e.child===null||e.tag===4)continue e;e.child.return=e,e=e.child}if(!(e.flags&2))return e.stateNode}}function dc(e,l,t){var a=e.tag;if(a===5||a===6)e=e.stateNode,l?t.nodeType===8?t.parentNode.insertBefore(e,l):t.insertBefore(e,l):(t.nodeType===8?(l=t.parentNode,l.insertBefore(e,t)):(l=t,l.appendChild(e)),t=t._reactRootContainer,t!=null||l.onclick!==null||(l.onclick=ju));else if(a!==4&&a!==27&&(e=e.child,e!==null))for(dc(e,l,t),e=e.sibling;e!==null;)dc(e,l,t),e=e.sibling}function vu(e,l,t){var a=e.tag;if(a===5||a===6)e=e.stateNode,l?t.insertBefore(e,l):t.appendChild(e);else if(a!==4&&a!==27&&(e=e.child,e!==null))for(vu(e,l,t),e=e.sibling;e!==null;)vu(e,l,t),e=e.sibling}var zl=!1,ie=!1,mc=!1,Lf=typeof WeakSet=="function"?WeakSet:Set,xe=null,qf=!1;function rh(e,l){if(e=e.containerInfo,wc=Bu,e=Ps(e),gi(e)){if("selectionStart"in e)var t={start:e.selectionStart,end:e.selectionEnd};else e:{t=(t=e.ownerDocument)&&t.defaultView||window;var a=t.getSelection&&t.getSelection();if(a&&a.rangeCount!==0){t=a.anchorNode;var n=a.anchorOffset,u=a.focusNode;a=a.focusOffset;try{t.nodeType,u.nodeType}catch{t=null;break e}var i=0,c=-1,s=-1,r=0,v=0,T=e,p=null;l:for(;;){for(var y;T!==t||n!==0&&T.nodeType!==3||(c=i+n),T!==u||a!==0&&T.nodeType!==3||(s=i+a),T.nodeType===3&&(i+=T.nodeValue.length),(y=T.firstChild)!==null;)p=T,T=y;for(;;){if(T===e)break l;if(p===t&&++r===n&&(c=i),p===u&&++v===a&&(s=i),(y=T.nextSibling)!==null)break;T=p,p=T.parentNode}T=y}t=c===-1||s===-1?null:{start:c,end:s}}else t=null}t=t||{start:0,end:0}}else t=null;for(Xc={focusedElem:e,selectionRange:t},Bu=!1,xe=l;xe!==null;)if(l=xe,e=l.child,(l.subtreeFlags&1028)!==0&&e!==null)e.return=l,xe=e;else for(;xe!==null;){switch(l=xe,u=l.alternate,e=l.flags,l.tag){case 0:break;case 11:case 15:break;case 1:if(e&1024&&u!==null){e=void 0,t=l,n=u.memoizedProps,u=u.memoizedState,a=t.stateNode;try{var M=Ot(t.type,n,t.elementType===t.type);e=a.getSnapshotBeforeUpdate(M,u),a.__reactInternalSnapshotBeforeUpdate=e}catch(_){P(t,t.return,_)}}break;case 3:if(e&1024){if(e=l.stateNode.containerInfo,t=e.nodeType,t===9)Zc(e);else if(t===1)switch(e.nodeName){case"HEAD":case"HTML":case"BODY":Zc(e);break;default:e.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if(e&1024)throw Error(d(163))}if(e=l.sibling,e!==null){e.return=l.return,xe=e;break}xe=l.return}return M=qf,qf=!1,M}function Bf(e,l,t){var a=t.flags;switch(t.tag){case 0:case 11:case 15:Ol(e,t),a&4&&cn(5,t);break;case 1:if(Ol(e,t),a&4)if(e=t.stateNode,l===null)try{e.componentDidMount()}catch(c){P(t,t.return,c)}else{var n=Ot(t.type,l.memoizedProps);l=l.memoizedState;try{e.componentDidUpdate(n,l,e.__reactInternalSnapshotBeforeUpdate)}catch(c){P(t,t.return,c)}}a&64&&jf(t),a&512&&Ut(t,t.return);break;case 3:if(Ol(e,t),a&64&&(a=t.updateQueue,a!==null)){if(e=null,t.child!==null)switch(t.child.tag){case 27:case 5:e=t.child.stateNode;break;case 1:e=t.child.stateNode}try{Of(a,e)}catch(c){P(t,t.return,c)}}break;case 26:Ol(e,t),a&512&&Ut(t,t.return);break;case 27:case 5:Ol(e,t),l===null&&a&4&&Uf(t),a&512&&Ut(t,t.return);break;case 12:Ol(e,t);break;case 13:Ol(e,t),a&4&&wf(e,t);break;case 22:if(n=t.memoizedState!==null||zl,!n){l=l!==null&&l.memoizedState!==null||ie;var u=zl,i=ie;zl=n,(ie=l)&&!i?Il(e,t,(t.subtreeFlags&8772)!==0):Ol(e,t),zl=u,ie=i}a&512&&(t.memoizedProps.mode==="manual"?Ut(t,t.return):qe(t,t.return));break;default:Ol(e,t)}}function Gf(e){var l=e.alternate;l!==null&&(e.alternate=null,Gf(l)),e.child=null,e.deletions=null,e.sibling=null,e.tag===5&&(l=e.stateNode,l!==null&&Iu(l)),e.stateNode=null,e.return=null,e.dependencies=null,e.memoizedProps=null,e.memoizedState=null,e.pendingProps=null,e.stateNode=null,e.updateQueue=null}var me=null,Be=!1;function Cl(e,l,t){for(t=t.child;t!==null;)Yf(e,l,t),t=t.sibling}function Yf(e,l,t){if(_e&&typeof _e.onCommitFiberUnmount=="function")try{_e.onCommitFiberUnmount(Ca,t)}catch{}switch(t.tag){case 26:ie||qe(t,l),Cl(e,l,t),t.memoizedState?t.memoizedState.count--:t.stateNode&&(t=t.stateNode,t.parentNode.removeChild(t));break;case 27:ie||qe(t,l);var a=me,n=Be;for(me=t.stateNode,Cl(e,l,t),t=t.stateNode,l=t.attributes;l.length;)t.removeAttributeNode(l[0]);Iu(t),me=a,Be=n;break;case 5:ie||qe(t,l);case 6:n=me;var u=Be;if(me=null,Cl(e,l,t),me=n,Be=u,me!==null)if(Be)try{e=me,a=t.stateNode,e.nodeType===8?e.parentNode.removeChild(a):e.removeChild(a)}catch(i){P(t,l,i)}else try{me.removeChild(t.stateNode)}catch(i){P(t,l,i)}break;case 18:me!==null&&(Be?(l=me,t=t.stateNode,l.nodeType===8?Kc(l.parentNode,t):l.nodeType===1&&Kc(l,t),Nn(l)):Kc(me,t.stateNode));break;case 4:a=me,n=Be,me=t.stateNode.containerInfo,Be=!0,Cl(e,l,t),me=a,Be=n;break;case 0:case 11:case 14:case 15:ie||Pl(2,t,l),ie||Pl(4,t,l),Cl(e,l,t);break;case 1:ie||(qe(t,l),a=t.stateNode,typeof a.componentWillUnmount=="function"&&Rf(t,l,a)),Cl(e,l,t);break;case 21:Cl(e,l,t);break;case 22:ie||qe(t,l),ie=(a=ie)||t.memoizedState!==null,Cl(e,l,t),ie=a;break;default:Cl(e,l,t)}}function wf(e,l){if(l.memoizedState===null&&(e=l.alternate,e!==null&&(e=e.memoizedState,e!==null&&(e=e.dehydrated,e!==null))))try{Nn(e)}catch(t){P(l,l.return,t)}}function dh(e){switch(e.tag){case 13:case 19:var l=e.stateNode;return l===null&&(l=e.stateNode=new Lf),l;case 22:return e=e.stateNode,l=e._retryCache,l===null&&(l=e._retryCache=new Lf),l;default:throw Error(d(435,e.tag))}}function hc(e,l){var t=dh(e);l.forEach(function(a){var n=Nh.bind(null,e,a);t.has(a)||(t.add(a),a.then(n,n))})}function $e(e,l){var t=l.deletions;if(t!==null)for(var a=0;a<t.length;a++){var n=t[a],u=e,i=l,c=i;e:for(;c!==null;){switch(c.tag){case 27:case 5:me=c.stateNode,Be=!1;break e;case 3:me=c.stateNode.containerInfo,Be=!0;break e;case 4:me=c.stateNode.containerInfo,Be=!0;break e}c=c.return}if(me===null)throw Error(d(160));Yf(u,i,n),me=null,Be=!1,u=n.alternate,u!==null&&(u.return=null),n.return=null}if(l.subtreeFlags&13878)for(l=l.child;l!==null;)Xf(l,e),l=l.sibling}var cl=null;function Xf(e,l){var t=e.alternate,a=e.flags;switch(e.tag){case 0:case 11:case 14:case 15:$e(l,e),We(e),a&4&&(Pl(3,e,e.return),cn(3,e),Pl(5,e,e.return));break;case 1:$e(l,e),We(e),a&512&&(ie||t===null||qe(t,t.return)),a&64&&zl&&(e=e.updateQueue,e!==null&&(a=e.callbacks,a!==null&&(t=e.shared.hiddenCallbacks,e.shared.hiddenCallbacks=t===null?a:t.concat(a))));break;case 26:var n=cl;if($e(l,e),We(e),a&512&&(ie||t===null||qe(t,t.return)),a&4){var u=t!==null?t.memoizedState:null;if(a=e.memoizedState,t===null)if(a===null)if(e.stateNode===null){e:{a=e.type,t=e.memoizedProps,n=n.ownerDocument||n;l:switch(a){case"title":u=n.getElementsByTagName("title")[0],(!u||u[Ra]||u[Me]||u.namespaceURI==="http://www.w3.org/2000/svg"||u.hasAttribute("itemprop"))&&(u=n.createElement(a),n.head.insertBefore(u,n.querySelector("head > title"))),Ee(u,a,t),u[Me]=e,Se(u),a=u;break e;case"link":var i=Ur("link","href",n).get(a+(t.href||""));if(i){for(var c=0;c<i.length;c++)if(u=i[c],u.getAttribute("href")===(t.href==null?null:t.href)&&u.getAttribute("rel")===(t.rel==null?null:t.rel)&&u.getAttribute("title")===(t.title==null?null:t.title)&&u.getAttribute("crossorigin")===(t.crossOrigin==null?null:t.crossOrigin)){i.splice(c,1);break l}}u=n.createElement(a),Ee(u,a,t),n.head.appendChild(u);break;case"meta":if(i=Ur("meta","content",n).get(a+(t.content||""))){for(c=0;c<i.length;c++)if(u=i[c],u.getAttribute("content")===(t.content==null?null:""+t.content)&&u.getAttribute("name")===(t.name==null?null:t.name)&&u.getAttribute("property")===(t.property==null?null:t.property)&&u.getAttribute("http-equiv")===(t.httpEquiv==null?null:t.httpEquiv)&&u.getAttribute("charset")===(t.charSet==null?null:t.charSet)){i.splice(c,1);break l}}u=n.createElement(a),Ee(u,a,t),n.head.appendChild(u);break;default:throw Error(d(468,a))}u[Me]=e,Se(u),a=u}e.stateNode=a}else _r(n,e.type,e.stateNode);else e.stateNode=Rr(n,a,e.memoizedProps);else u!==a?(u===null?t.stateNode!==null&&(t=t.stateNode,t.parentNode.removeChild(t)):u.count--,a===null?_r(n,e.type,e.stateNode):Rr(n,a,e.memoizedProps)):a===null&&e.stateNode!==null&&_f(e,e.memoizedProps,t.memoizedProps)}break;case 27:if(a&4&&e.alternate===null){n=e.stateNode,u=e.memoizedProps;try{for(var s=n.firstChild;s;){var r=s.nextSibling,v=s.nodeName;s[Ra]||v==="HEAD"||v==="BODY"||v==="SCRIPT"||v==="STYLE"||v==="LINK"&&s.rel.toLowerCase()==="stylesheet"||n.removeChild(s),s=r}for(var T=e.type,p=n.attributes;p.length;)n.removeAttributeNode(p[0]);Ee(n,T,u),n[Me]=e,n[je]=u}catch(M){P(e,e.return,M)}}case 5:if($e(l,e),We(e),a&512&&(ie||t===null||qe(t,t.return)),e.flags&32){n=e.stateNode;try{Ft(n,"")}catch(M){P(e,e.return,M)}}a&4&&e.stateNode!=null&&(n=e.memoizedProps,_f(e,n,t!==null?t.memoizedProps:n)),a&1024&&(mc=!0);break;case 6:if($e(l,e),We(e),a&4){if(e.stateNode===null)throw Error(d(162));a=e.memoizedProps,t=e.stateNode;try{t.nodeValue=a}catch(M){P(e,e.return,M)}}break;case 3:if(Hu=null,n=cl,cl=Uu(l.containerInfo),$e(l,e),cl=n,We(e),a&4&&t!==null&&t.memoizedState.isDehydrated)try{Nn(l.containerInfo)}catch(M){P(e,e.return,M)}mc&&(mc=!1,Qf(e));break;case 4:a=cl,cl=Uu(e.stateNode.containerInfo),$e(l,e),We(e),cl=a;break;case 12:$e(l,e),We(e);break;case 13:$e(l,e),We(e),e.child.flags&8192&&e.memoizedState!==null!=(t!==null&&t.memoizedState!==null)&&(Ac=rl()),a&4&&(a=e.updateQueue,a!==null&&(e.updateQueue=null,hc(e,a)));break;case 22:if(a&512&&(ie||t===null||qe(t,t.return)),s=e.memoizedState!==null,r=t!==null&&t.memoizedState!==null,v=zl,T=ie,zl=v||s,ie=T||r,$e(l,e),ie=T,zl=v,We(e),l=e.stateNode,l._current=e,l._visibility&=-3,l._visibility|=l._pendingVisibility&2,a&8192&&(l._visibility=s?l._visibility&-2:l._visibility|1,s&&(l=zl||ie,t===null||r||l||da(e)),e.memoizedProps===null||e.memoizedProps.mode!=="manual"))e:for(t=null,l=e;;){if(l.tag===5||l.tag===26||l.tag===27){if(t===null){r=t=l;try{if(n=r.stateNode,s)u=n.style,typeof u.setProperty=="function"?u.setProperty("display","none","important"):u.display="none";else{i=r.stateNode,c=r.memoizedProps.style;var y=c!=null&&c.hasOwnProperty("display")?c.display:null;i.style.display=y==null||typeof y=="boolean"?"":(""+y).trim()}}catch(M){P(r,r.return,M)}}}else if(l.tag===6){if(t===null){r=l;try{r.stateNode.nodeValue=s?"":r.memoizedProps}catch(M){P(r,r.return,M)}}}else if((l.tag!==22&&l.tag!==23||l.memoizedState===null||l===e)&&l.child!==null){l.child.return=l,l=l.child;continue}if(l===e)break e;for(;l.sibling===null;){if(l.return===null||l.return===e)break e;t===l&&(t=null),l=l.return}t===l&&(t=null),l.sibling.return=l.return,l=l.sibling}a&4&&(a=e.updateQueue,a!==null&&(t=a.retryQueue,t!==null&&(a.retryQueue=null,hc(e,t))));break;case 19:$e(l,e),We(e),a&4&&(a=e.updateQueue,a!==null&&(e.updateQueue=null,hc(e,a)));break;case 21:break;default:$e(l,e),We(e)}}function We(e){var l=e.flags;if(l&2){try{if(e.tag!==27){e:{for(var t=e.return;t!==null;){if(Hf(t)){var a=t;break e}t=t.return}throw Error(d(160))}switch(a.tag){case 27:var n=a.stateNode,u=rc(e);vu(e,u,n);break;case 5:var i=a.stateNode;a.flags&32&&(Ft(i,""),a.flags&=-33);var c=rc(e);vu(e,c,i);break;case 3:case 4:var s=a.stateNode.containerInfo,r=rc(e);dc(e,r,s);break;default:throw Error(d(161))}}}catch(v){P(e,e.return,v)}e.flags&=-3}l&4096&&(e.flags&=-4097)}function Qf(e){if(e.subtreeFlags&1024)for(e=e.child;e!==null;){var l=e;Qf(l),l.tag===5&&l.flags&1024&&l.stateNode.reset(),e=e.sibling}}function Ol(e,l){if(l.subtreeFlags&8772)for(l=l.child;l!==null;)Bf(e,l.alternate,l),l=l.sibling}function da(e){for(e=e.child;e!==null;){var l=e;switch(l.tag){case 0:case 11:case 14:case 15:Pl(4,l,l.return),da(l);break;case 1:qe(l,l.return);var t=l.stateNode;typeof t.componentWillUnmount=="function"&&Rf(l,l.return,t),da(l);break;case 26:case 27:case 5:qe(l,l.return),da(l);break;case 22:qe(l,l.return),l.memoizedState===null&&da(l);break;default:da(l)}e=e.sibling}}function Il(e,l,t){for(t=t&&(l.subtreeFlags&8772)!==0,l=l.child;l!==null;){var a=l.alternate,n=e,u=l,i=u.flags;switch(u.tag){case 0:case 11:case 15:Il(n,u,t),cn(4,u);break;case 1:if(Il(n,u,t),a=u,n=a.stateNode,typeof n.componentDidMount=="function")try{n.componentDidMount()}catch(r){P(a,a.return,r)}if(a=u,n=a.updateQueue,n!==null){var c=a.stateNode;try{var s=n.shared.hiddenCallbacks;if(s!==null)for(n.shared.hiddenCallbacks=null,n=0;n<s.length;n++)Cf(s[n],c)}catch(r){P(a,a.return,r)}}t&&i&64&&jf(u),Ut(u,u.return);break;case 26:case 27:case 5:Il(n,u,t),t&&a===null&&i&4&&Uf(u),Ut(u,u.return);break;case 12:Il(n,u,t);break;case 13:Il(n,u,t),t&&i&4&&wf(n,u);break;case 22:u.memoizedState===null&&Il(n,u,t),Ut(u,u.return);break;default:Il(n,u,t)}l=l.sibling}}function gc(e,l){var t=null;e!==null&&e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(t=e.memoizedState.cachePool.pool),e=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(e=l.memoizedState.cachePool.pool),e!==t&&(e!=null&&e.refCount++,t!=null&&$a(t))}function pc(e,l){e=null,l.alternate!==null&&(e=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==e&&(l.refCount++,e!=null&&$a(e))}function et(e,l,t,a){if(l.subtreeFlags&10256)for(l=l.child;l!==null;)Vf(e,l,t,a),l=l.sibling}function Vf(e,l,t,a){var n=l.flags;switch(l.tag){case 0:case 11:case 15:et(e,l,t,a),n&2048&&cn(9,l);break;case 3:et(e,l,t,a),n&2048&&(e=null,l.alternate!==null&&(e=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==e&&(l.refCount++,e!=null&&$a(e)));break;case 12:if(n&2048){et(e,l,t,a),e=l.stateNode;try{var u=l.memoizedProps,i=u.id,c=u.onPostCommit;typeof c=="function"&&c(i,l.alternate===null?"mount":"update",e.passiveEffectDuration,-0)}catch(s){P(l,l.return,s)}}else et(e,l,t,a);break;case 23:break;case 22:u=l.stateNode,l.memoizedState!==null?u._visibility&4?et(e,l,t,a):sn(e,l):u._visibility&4?et(e,l,t,a):(u._visibility|=4,ma(e,l,t,a,(l.subtreeFlags&10256)!==0)),n&2048&&gc(l.alternate,l);break;case 24:et(e,l,t,a),n&2048&&pc(l.alternate,l);break;default:et(e,l,t,a)}}function ma(e,l,t,a,n){for(n=n&&(l.subtreeFlags&10256)!==0,l=l.child;l!==null;){var u=e,i=l,c=t,s=a,r=i.flags;switch(i.tag){case 0:case 11:case 15:ma(u,i,c,s,n),cn(8,i);break;case 23:break;case 22:var v=i.stateNode;i.memoizedState!==null?v._visibility&4?ma(u,i,c,s,n):sn(u,i):(v._visibility|=4,ma(u,i,c,s,n)),n&&r&2048&&gc(i.alternate,i);break;case 24:ma(u,i,c,s,n),n&&r&2048&&pc(i.alternate,i);break;default:ma(u,i,c,s,n)}l=l.sibling}}function sn(e,l){if(l.subtreeFlags&10256)for(l=l.child;l!==null;){var t=e,a=l,n=a.flags;switch(a.tag){case 22:sn(t,a),n&2048&&gc(a.alternate,a);break;case 24:sn(t,a),n&2048&&pc(a.alternate,a);break;default:sn(t,a)}l=l.sibling}}var on=8192;function ha(e){if(e.subtreeFlags&on)for(e=e.child;e!==null;)Kf(e),e=e.sibling}function Kf(e){switch(e.tag){case 26:ha(e),e.flags&on&&e.memoizedState!==null&&eg(cl,e.memoizedState,e.memoizedProps);break;case 5:ha(e);break;case 3:case 4:var l=cl;cl=Uu(e.stateNode.containerInfo),ha(e),cl=l;break;case 22:e.memoizedState===null&&(l=e.alternate,l!==null&&l.memoizedState!==null?(l=on,on=16777216,ha(e),on=l):ha(e));break;default:ha(e)}}function Zf(e){var l=e.alternate;if(l!==null&&(e=l.child,e!==null)){l.child=null;do l=e.sibling,e.sibling=null,e=l;while(e!==null)}}function fn(e){var l=e.deletions;if(e.flags&16){if(l!==null)for(var t=0;t<l.length;t++){var a=l[t];xe=a,Jf(a,e)}Zf(e)}if(e.subtreeFlags&10256)for(e=e.child;e!==null;)kf(e),e=e.sibling}function kf(e){switch(e.tag){case 0:case 11:case 15:fn(e),e.flags&2048&&Pl(9,e,e.return);break;case 3:fn(e);break;case 12:fn(e);break;case 22:var l=e.stateNode;e.memoizedState!==null&&l._visibility&4&&(e.return===null||e.return.tag!==13)?(l._visibility&=-5,bu(e)):fn(e);break;default:fn(e)}}function bu(e){var l=e.deletions;if(e.flags&16){if(l!==null)for(var t=0;t<l.length;t++){var a=l[t];xe=a,Jf(a,e)}Zf(e)}for(e=e.child;e!==null;){switch(l=e,l.tag){case 0:case 11:case 15:Pl(8,l,l.return),bu(l);break;case 22:t=l.stateNode,t._visibility&4&&(t._visibility&=-5,bu(l));break;default:bu(l)}e=e.sibling}}function Jf(e,l){for(;xe!==null;){var t=xe;switch(t.tag){case 0:case 11:case 15:Pl(8,t,l);break;case 23:case 22:if(t.memoizedState!==null&&t.memoizedState.cachePool!==null){var a=t.memoizedState.cachePool.pool;a!=null&&a.refCount++}break;case 24:$a(t.memoizedState.cache)}if(a=t.child,a!==null)a.return=t,xe=a;else e:for(t=e;xe!==null;){a=xe;var n=a.sibling,u=a.return;if(Gf(a),a===t){xe=null;break e}if(n!==null){n.return=u,xe=n;break e}xe=u}}}function mh(e,l,t,a){this.tag=e,this.key=t,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=l,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=a,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function Pe(e,l,t,a){return new mh(e,l,t,a)}function yc(e){return e=e.prototype,!(!e||!e.isReactComponent)}function lt(e,l){var t=e.alternate;return t===null?(t=Pe(e.tag,l,e.key,e.mode),t.elementType=e.elementType,t.type=e.type,t.stateNode=e.stateNode,t.alternate=e,e.alternate=t):(t.pendingProps=l,t.type=e.type,t.flags=0,t.subtreeFlags=0,t.deletions=null),t.flags=e.flags&31457280,t.childLanes=e.childLanes,t.lanes=e.lanes,t.child=e.child,t.memoizedProps=e.memoizedProps,t.memoizedState=e.memoizedState,t.updateQueue=e.updateQueue,l=e.dependencies,t.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext},t.sibling=e.sibling,t.index=e.index,t.ref=e.ref,t.refCleanup=e.refCleanup,t}function Ff(e,l){e.flags&=31457282;var t=e.alternate;return t===null?(e.childLanes=0,e.lanes=l,e.child=null,e.subtreeFlags=0,e.memoizedProps=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.stateNode=null):(e.childLanes=t.childLanes,e.lanes=t.lanes,e.child=t.child,e.subtreeFlags=0,e.deletions=null,e.memoizedProps=t.memoizedProps,e.memoizedState=t.memoizedState,e.updateQueue=t.updateQueue,e.type=t.type,l=t.dependencies,e.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext}),e}function Su(e,l,t,a,n,u){var i=0;if(a=e,typeof e=="function")yc(e)&&(i=1);else if(typeof e=="string")i=Ph(e,t,fl.current)?26:e==="html"||e==="head"||e==="body"?27:5;else e:switch(e){case Y:return _t(t.children,n,u,l);case W:i=8,n|=24;break;case he:return e=Pe(12,t,l,n|2),e.elementType=he,e.lanes=u,e;case C:return e=Pe(13,t,l,n),e.elementType=C,e.lanes=u,e;case tl:return e=Pe(19,t,l,n),e.elementType=tl,e.lanes=u,e;case ft:return $f(t,n,u,l);default:if(typeof e=="object"&&e!==null)switch(e.$$typeof){case ge:case pe:i=10;break e;case X:i=9;break e;case le:i=11;break e;case Gt:i=14;break e;case al:i=16,a=null;break e}i=29,t=Error(d(130,e===null?"null":typeof e,"")),a=null}return l=Pe(i,t,l,n),l.elementType=e,l.type=a,l.lanes=u,l}function _t(e,l,t,a){return e=Pe(7,e,a,l),e.lanes=t,e}function $f(e,l,t,a){e=Pe(22,e,a,l),e.elementType=ft,e.lanes=t;var n={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var u=n._current;if(u===null)throw Error(d(456));if(!(n._pendingVisibility&2)){var i=Ql(u,2);i!==null&&(n._pendingVisibility|=2,Oe(i,u,2))}},attach:function(){var u=n._current;if(u===null)throw Error(d(456));if(n._pendingVisibility&2){var i=Ql(u,2);i!==null&&(n._pendingVisibility&=-3,Oe(i,u,2))}}};return e.stateNode=n,e}function vc(e,l,t){return e=Pe(6,e,null,l),e.lanes=t,e}function bc(e,l,t){return l=Pe(4,e.children!==null?e.children:[],e.key,l),l.lanes=t,l.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},l}function jl(e){e.flags|=4}function Wf(e,l){if(l.type!=="stylesheet"||l.state.loading&4)e.flags&=-16777217;else if(e.flags|=16777216,!Hr(l)){if(l=Fe.current,l!==null&&((Q&4194176)===Q?ml!==null:(Q&62914560)!==Q&&!(Q&536870912)||l!==ml))throw ka=Di,mo;e.flags|=8192}}function Tu(e,l){l!==null&&(e.flags|=4),e.flags&16384&&(l=e.tag!==22?ms():536870912,e.lanes|=l,pa|=l)}function rn(e,l){if(!K)switch(e.tailMode){case"hidden":l=e.tail;for(var t=null;l!==null;)l.alternate!==null&&(t=l),l=l.sibling;t===null?e.tail=null:t.sibling=null;break;case"collapsed":t=e.tail;for(var a=null;t!==null;)t.alternate!==null&&(a=t),t=t.sibling;a===null?l||e.tail===null?e.tail=null:e.tail.sibling=null:a.sibling=null}}function ne(e){var l=e.alternate!==null&&e.alternate.child===e.child,t=0,a=0;if(l)for(var n=e.child;n!==null;)t|=n.lanes|n.childLanes,a|=n.subtreeFlags&31457280,a|=n.flags&31457280,n.return=e,n=n.sibling;else for(n=e.child;n!==null;)t|=n.lanes|n.childLanes,a|=n.subtreeFlags,a|=n.flags,n.return=e,n=n.sibling;return e.subtreeFlags|=a,e.childLanes=t,l}function hh(e,l,t){var a=l.pendingProps;switch(xi(l),l.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return ne(l),null;case 1:return ne(l),null;case 3:return t=l.stateNode,a=null,e!==null&&(a=e.memoizedState.cache),l.memoizedState.cache!==a&&(l.flags|=2048),Ml(ve),Qt(),t.pendingContext&&(t.context=t.pendingContext,t.pendingContext=null),(e===null||e.child===null)&&(Qa(l)?jl(l):e===null||e.memoizedState.isDehydrated&&!(l.flags&256)||(l.flags|=1024,il!==null&&(Mc(il),il=null))),ne(l),null;case 26:return t=l.memoizedState,e===null?(jl(l),t!==null?(ne(l),Wf(l,t)):(ne(l),l.flags&=-16777217)):t?t!==e.memoizedState?(jl(l),ne(l),Wf(l,t)):(ne(l),l.flags&=-16777217):(e.memoizedProps!==a&&jl(l),ne(l),l.flags&=-16777217),null;case 27:Un(l),t=Gl.current;var n=l.type;if(e!==null&&l.stateNode!=null)e.memoizedProps!==a&&jl(l);else{if(!a){if(l.stateNode===null)throw Error(d(166));return ne(l),null}e=fl.current,Qa(l)?fo(l):(e=Mr(n,a,t),l.stateNode=e,jl(l))}return ne(l),null;case 5:if(Un(l),t=l.type,e!==null&&l.stateNode!=null)e.memoizedProps!==a&&jl(l);else{if(!a){if(l.stateNode===null)throw Error(d(166));return ne(l),null}if(e=fl.current,Qa(l))fo(l);else{switch(n=Ru(Gl.current),e){case 1:e=n.createElementNS("http://www.w3.org/2000/svg",t);break;case 2:e=n.createElementNS("http://www.w3.org/1998/Math/MathML",t);break;default:switch(t){case"svg":e=n.createElementNS("http://www.w3.org/2000/svg",t);break;case"math":e=n.createElementNS("http://www.w3.org/1998/Math/MathML",t);break;case"script":e=n.createElement("div"),e.innerHTML="<script><\/script>",e=e.removeChild(e.firstChild);break;case"select":e=typeof a.is=="string"?n.createElement("select",{is:a.is}):n.createElement("select"),a.multiple?e.multiple=!0:a.size&&(e.size=a.size);break;default:e=typeof a.is=="string"?n.createElement(t,{is:a.is}):n.createElement(t)}}e[Me]=l,e[je]=a;e:for(n=l.child;n!==null;){if(n.tag===5||n.tag===6)e.appendChild(n.stateNode);else if(n.tag!==4&&n.tag!==27&&n.child!==null){n.child.return=n,n=n.child;continue}if(n===l)break e;for(;n.sibling===null;){if(n.return===null||n.return===l)break e;n=n.return}n.sibling.return=n.return,n=n.sibling}l.stateNode=e;e:switch(Ee(e,t,a),t){case"button":case"input":case"select":case"textarea":e=!!a.autoFocus;break e;case"img":e=!0;break e;default:e=!1}e&&jl(l)}}return ne(l),l.flags&=-16777217,null;case 6:if(e&&l.stateNode!=null)e.memoizedProps!==a&&jl(l);else{if(typeof a!="string"&&l.stateNode===null)throw Error(d(166));if(e=Gl.current,Qa(l)){if(e=l.stateNode,t=l.memoizedProps,a=null,n=Ce,n!==null)switch(n.tag){case 27:case 5:a=n.memoizedProps}e[Me]=l,e=!!(e.nodeValue===t||a!==null&&a.suppressHydrationWarning===!0||Tr(e.nodeValue,t)),e||Dt(l)}else e=Ru(e).createTextNode(a),e[Me]=l,l.stateNode=e}return ne(l),null;case 13:if(a=l.memoizedState,e===null||e.memoizedState!==null&&e.memoizedState.dehydrated!==null){if(n=Qa(l),a!==null&&a.dehydrated!==null){if(e===null){if(!n)throw Error(d(318));if(n=l.memoizedState,n=n!==null?n.dehydrated:null,!n)throw Error(d(317));n[Me]=l}else Va(),!(l.flags&128)&&(l.memoizedState=null),l.flags|=4;ne(l),n=!1}else il!==null&&(Mc(il),il=null),n=!0;if(!n)return l.flags&256?(Al(l),l):(Al(l),null)}if(Al(l),l.flags&128)return l.lanes=t,l;if(t=a!==null,e=e!==null&&e.memoizedState!==null,t){a=l.child,n=null,a.alternate!==null&&a.alternate.memoizedState!==null&&a.alternate.memoizedState.cachePool!==null&&(n=a.alternate.memoizedState.cachePool.pool);var u=null;a.memoizedState!==null&&a.memoizedState.cachePool!==null&&(u=a.memoizedState.cachePool.pool),u!==n&&(a.flags|=2048)}return t!==e&&t&&(l.child.flags|=8192),Tu(l,l.updateQueue),ne(l),null;case 4:return Qt(),e===null&&Bc(l.stateNode.containerInfo),ne(l),null;case 10:return Ml(l.type),ne(l),null;case 19:if(be(ye),n=l.memoizedState,n===null)return ne(l),null;if(a=(l.flags&128)!==0,u=n.rendering,u===null)if(a)rn(n,!1);else{if(ce!==0||e!==null&&e.flags&128)for(e=l.child;e!==null;){if(u=nu(e),u!==null){for(l.flags|=128,rn(n,!1),e=u.updateQueue,l.updateQueue=e,Tu(l,e),l.subtreeFlags=0,e=t,t=l.child;t!==null;)Ff(t,e),t=t.sibling;return ae(ye,ye.current&1|2),l.child}e=e.sibling}n.tail!==null&&rl()>xu&&(l.flags|=128,a=!0,rn(n,!1),l.lanes=4194304)}else{if(!a)if(e=nu(u),e!==null){if(l.flags|=128,a=!0,e=e.updateQueue,l.updateQueue=e,Tu(l,e),rn(n,!0),n.tail===null&&n.tailMode==="hidden"&&!u.alternate&&!K)return ne(l),null}else 2*rl()-n.renderingStartTime>xu&&t!==536870912&&(l.flags|=128,a=!0,rn(n,!1),l.lanes=4194304);n.isBackwards?(u.sibling=l.child,l.child=u):(e=n.last,e!==null?e.sibling=u:l.child=u,n.last=u)}return n.tail!==null?(l=n.tail,n.rendering=l,n.tail=l.sibling,n.renderingStartTime=rl(),l.sibling=null,e=ye.current,ae(ye,a?e&1|2:e&1),l):(ne(l),null);case 22:case 23:return Al(l),Ei(),a=l.memoizedState!==null,e!==null?e.memoizedState!==null!==a&&(l.flags|=8192):a&&(l.flags|=8192),a?t&536870912&&!(l.flags&128)&&(ne(l),l.subtreeFlags&6&&(l.flags|=8192)):ne(l),t=l.updateQueue,t!==null&&Tu(l,t.retryQueue),t=null,e!==null&&e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(t=e.memoizedState.cachePool.pool),a=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(a=l.memoizedState.cachePool.pool),a!==t&&(l.flags|=2048),e!==null&&be(Et),null;case 24:return t=null,e!==null&&(t=e.memoizedState.cache),l.memoizedState.cache!==t&&(l.flags|=2048),Ml(ve),ne(l),null;case 25:return null}throw Error(d(156,l.tag))}function gh(e,l){switch(xi(l),l.tag){case 1:return e=l.flags,e&65536?(l.flags=e&-65537|128,l):null;case 3:return Ml(ve),Qt(),e=l.flags,e&65536&&!(e&128)?(l.flags=e&-65537|128,l):null;case 26:case 27:case 5:return Un(l),null;case 13:if(Al(l),e=l.memoizedState,e!==null&&e.dehydrated!==null){if(l.alternate===null)throw Error(d(340));Va()}return e=l.flags,e&65536?(l.flags=e&-65537|128,l):null;case 19:return be(ye),null;case 4:return Qt(),null;case 10:return Ml(l.type),null;case 22:case 23:return Al(l),Ei(),e!==null&&be(Et),e=l.flags,e&65536?(l.flags=e&-65537|128,l):null;case 24:return Ml(ve),null;case 25:return null;default:return null}}function Pf(e,l){switch(xi(l),l.tag){case 3:Ml(ve),Qt();break;case 26:case 27:case 5:Un(l);break;case 4:Qt();break;case 13:Al(l);break;case 19:be(ye);break;case 10:Ml(l.type);break;case 22:case 23:Al(l),Ei(),e!==null&&be(Et);break;case 24:Ml(ve)}}var ph={getCacheForType:function(e){var l=ze(ve),t=l.data.get(e);return t===void 0&&(t=e(),l.data.set(e,t)),t}},yh=typeof WeakMap=="function"?WeakMap:Map,ue=0,I=null,G=null,Q=0,ee=0,Ge=null,Rl=!1,ga=!1,Sc=!1,Ul=0,ce=0,tt=0,Ht=0,Tc=0,Ie=0,pa=0,dn=null,gl=null,xc=!1,Ac=0,xu=1/0,Au=null,at=null,Du=!1,Lt=null,mn=0,Dc=0,Nc=null,hn=0,Ec=null;function Ye(){if(ue&2&&Q!==0)return Q&-Q;if(R.T!==null){var e=ca;return e!==0?e:_c()}return ys()}function If(){Ie===0&&(Ie=!(Q&536870912)||K?ds():536870912);var e=Fe.current;return e!==null&&(e.flags|=32),Ie}function Oe(e,l,t){(e===I&&ee===2||e.cancelPendingCommit!==null)&&(ya(e,0),_l(e,Q,Ie,!1)),ja(e,t),(!(ue&2)||e!==I)&&(e===I&&(!(ue&2)&&(Ht|=t),ce===4&&_l(e,Q,Ie,!1)),pl(e))}function er(e,l,t){if(ue&6)throw Error(d(327));var a=!t&&(l&60)===0&&(l&e.expiredLanes)===0||Oa(e,l),n=a?Sh(e,l):Oc(e,l,!0),u=a;do{if(n===0){ga&&!a&&_l(e,l,0,!1);break}else if(n===6)_l(e,l,0,!Rl);else{if(t=e.current.alternate,u&&!vh(t)){n=Oc(e,l,!1),u=!1;continue}if(n===2){if(u=l,e.errorRecoveryDisabledLanes&u)var i=0;else i=e.pendingLanes&-536870913,i=i!==0?i:i&536870912?536870912:0;if(i!==0){l=i;e:{var c=e;n=dn;var s=c.current.memoizedState.isDehydrated;if(s&&(ya(c,i).flags|=256),i=Oc(c,i,!1),i!==2){if(Sc&&!s){c.errorRecoveryDisabledLanes|=u,Ht|=u,n=4;break e}u=gl,gl=n,u!==null&&Mc(u)}n=i}if(u=!1,n!==2)continue}}if(n===1){ya(e,0),_l(e,l,0,!0);break}e:{switch(a=e,n){case 0:case 1:throw Error(d(345));case 4:if((l&4194176)===l){_l(a,l,Ie,!Rl);break e}break;case 2:gl=null;break;case 3:case 5:break;default:throw Error(d(329))}if(a.finishedWork=t,a.finishedLanes=l,(l&62914560)===l&&(u=Ac+300-rl(),10<u)){if(_l(a,l,Ie,!Rl),qn(a,0)!==0)break e;a.timeoutHandle=Dr(lr.bind(null,a,t,gl,Au,xc,l,Ie,Ht,pa,Rl,2,-0,0),u);break e}lr(a,t,gl,Au,xc,l,Ie,Ht,pa,Rl,0,-0,0)}}break}while(!0);pl(e)}function Mc(e){gl===null?gl=e:gl.push.apply(gl,e)}function lr(e,l,t,a,n,u,i,c,s,r,v,T,p){var y=l.subtreeFlags;if((y&8192||(y&16785408)===16785408)&&(Sn={stylesheets:null,count:0,unsuspend:Ih},Kf(l),l=lg(),l!==null)){e.cancelPendingCommit=l(sr.bind(null,e,t,a,n,i,c,s,1,T,p)),_l(e,u,i,!r);return}sr(e,t,a,n,i,c,s,v,T,p)}function vh(e){for(var l=e;;){var t=l.tag;if((t===0||t===11||t===15)&&l.flags&16384&&(t=l.updateQueue,t!==null&&(t=t.stores,t!==null)))for(var a=0;a<t.length;a++){var n=t[a],u=n.getSnapshot;n=n.value;try{if(!Le(u(),n))return!1}catch{return!1}}if(t=l.child,l.subtreeFlags&16384&&t!==null)t.return=l,l=t;else{if(l===e)break;for(;l.sibling===null;){if(l.return===null||l.return===e)return!0;l=l.return}l.sibling.return=l.return,l=l.sibling}}return!0}function _l(e,l,t,a){l&=~Tc,l&=~Ht,e.suspendedLanes|=l,e.pingedLanes&=~l,a&&(e.warmLanes|=l),a=e.expirationTimes;for(var n=l;0<n;){var u=31-He(n),i=1<<u;a[u]=-1,n&=~i}t!==0&&hs(e,t,l)}function Nu(){return ue&6?!0:(gn(0),!1)}function zc(){if(G!==null){if(ee===0)var e=G.return;else e=G,El=jt=null,_i(e),ua=null,Ja=0,e=G;for(;e!==null;)Pf(e.alternate,e),e=e.return;G=null}}function ya(e,l){e.finishedWork=null,e.finishedLanes=0;var t=e.timeoutHandle;t!==-1&&(e.timeoutHandle=-1,qh(t)),t=e.cancelPendingCommit,t!==null&&(e.cancelPendingCommit=null,t()),zc(),I=e,G=t=lt(e.current,null),Q=l,ee=0,Ge=null,Rl=!1,ga=Oa(e,l),Sc=!1,pa=Ie=Tc=Ht=tt=ce=0,gl=dn=null,xc=!1,l&8&&(l|=l&32);var a=e.entangledLanes;if(a!==0)for(e=e.entanglements,a&=l;0<a;){var n=31-He(a),u=1<<n;l|=e[n],a&=~u}return Ul=l,Fn(),t}function tr(e,l){q=null,R.H=hl,l===Za?(l=po(),ee=3):l===mo?(l=po(),ee=4):ee=l===gf?8:l!==null&&typeof l=="object"&&typeof l.then=="function"?6:1,Ge=l,G===null&&(ce=1,gu(e,Ze(l,e.current)))}function ar(){var e=R.H;return R.H=hl,e===null?hl:e}function nr(){var e=R.A;return R.A=ph,e}function Cc(){ce=4,Rl||(Q&4194176)!==Q&&Fe.current!==null||(ga=!0),!(tt&134217727)&&!(Ht&134217727)||I===null||_l(I,Q,Ie,!1)}function Oc(e,l,t){var a=ue;ue|=2;var n=ar(),u=nr();(I!==e||Q!==l)&&(Au=null,ya(e,l)),l=!1;var i=ce;e:do try{if(ee!==0&&G!==null){var c=G,s=Ge;switch(ee){case 8:zc(),i=6;break e;case 3:case 2:case 6:Fe.current===null&&(l=!0);var r=ee;if(ee=0,Ge=null,va(e,c,s,r),t&&ga){i=0;break e}break;default:r=ee,ee=0,Ge=null,va(e,c,s,r)}}bh(),i=ce;break}catch(v){tr(e,v)}while(!0);return l&&e.shellSuspendCounter++,El=jt=null,ue=a,R.H=n,R.A=u,G===null&&(I=null,Q=0,Fn()),i}function bh(){for(;G!==null;)ur(G)}function Sh(e,l){var t=ue;ue|=2;var a=ar(),n=nr();I!==e||Q!==l?(Au=null,xu=rl()+500,ya(e,l)):ga=Oa(e,l);e:do try{if(ee!==0&&G!==null){l=G;var u=Ge;l:switch(ee){case 1:ee=0,Ge=null,va(e,l,u,1);break;case 2:if(ho(u)){ee=0,Ge=null,ir(l);break}l=function(){ee===2&&I===e&&(ee=7),pl(e)},u.then(l,l);break e;case 3:ee=7;break e;case 4:ee=5;break e;case 7:ho(u)?(ee=0,Ge=null,ir(l)):(ee=0,Ge=null,va(e,l,u,7));break;case 5:var i=null;switch(G.tag){case 26:i=G.memoizedState;case 5:case 27:var c=G;if(!i||Hr(i)){ee=0,Ge=null;var s=c.sibling;if(s!==null)G=s;else{var r=c.return;r!==null?(G=r,Eu(r)):G=null}break l}}ee=0,Ge=null,va(e,l,u,5);break;case 6:ee=0,Ge=null,va(e,l,u,6);break;case 8:zc(),ce=6;break e;default:throw Error(d(462))}}Th();break}catch(v){tr(e,v)}while(!0);return El=jt=null,R.H=a,R.A=n,ue=t,G!==null?0:(I=null,Q=0,Fn(),ce)}function Th(){for(;G!==null&&!Qd();)ur(G)}function ur(e){var l=Mf(e.alternate,e,Ul);e.memoizedProps=e.pendingProps,l===null?Eu(e):G=l}function ir(e){var l=e,t=l.alternate;switch(l.tag){case 15:case 0:l=Tf(t,l,l.pendingProps,l.type,void 0,Q);break;case 11:l=Tf(t,l,l.pendingProps,l.type.render,l.ref,Q);break;case 5:_i(l);default:Pf(t,l),l=G=Ff(l,Ul),l=Mf(t,l,Ul)}e.memoizedProps=e.pendingProps,l===null?Eu(e):G=l}function va(e,l,t,a){El=jt=null,_i(l),ua=null,Ja=0;var n=l.return;try{if(oh(e,n,l,t,Q)){ce=1,gu(e,Ze(t,e.current)),G=null;return}}catch(u){if(n!==null)throw G=n,u;ce=1,gu(e,Ze(t,e.current)),G=null;return}l.flags&32768?(K||a===1?e=!0:ga||Q&536870912?e=!1:(Rl=e=!0,(a===2||a===3||a===6)&&(a=Fe.current,a!==null&&a.tag===13&&(a.flags|=16384))),cr(l,e)):Eu(l)}function Eu(e){var l=e;do{if(l.flags&32768){cr(l,Rl);return}e=l.return;var t=hh(l.alternate,l,Ul);if(t!==null){G=t;return}if(l=l.sibling,l!==null){G=l;return}G=l=e}while(l!==null);ce===0&&(ce=5)}function cr(e,l){do{var t=gh(e.alternate,e);if(t!==null){t.flags&=32767,G=t;return}if(t=e.return,t!==null&&(t.flags|=32768,t.subtreeFlags=0,t.deletions=null),!l&&(e=e.sibling,e!==null)){G=e;return}G=e=t}while(e!==null);ce=6,G=null}function sr(e,l,t,a,n,u,i,c,s,r){var v=R.T,T=V.p;try{V.p=2,R.T=null,xh(e,l,t,a,T,n,u,i,c,s,r)}finally{R.T=v,V.p=T}}function xh(e,l,t,a,n,u,i,c){do ba();while(Lt!==null);if(ue&6)throw Error(d(327));var s=e.finishedWork;if(a=e.finishedLanes,s===null)return null;if(e.finishedWork=null,e.finishedLanes=0,s===e.current)throw Error(d(177));e.callbackNode=null,e.callbackPriority=0,e.cancelPendingCommit=null;var r=s.lanes|s.childLanes;if(r|=bi,em(e,a,r,u,i,c),e===I&&(G=I=null,Q=0),!(s.subtreeFlags&10256)&&!(s.flags&10256)||Du||(Du=!0,Dc=r,Nc=t,Eh(_n,function(){return ba(),null})),t=(s.flags&15990)!==0,s.subtreeFlags&15990||t?(t=R.T,R.T=null,u=V.p,V.p=2,i=ue,ue|=4,rh(e,s),Xf(s,e),Zm(Xc,e.containerInfo),Bu=!!wc,Xc=wc=null,e.current=s,Bf(e,s.alternate,s),Vd(),ue=i,V.p=u,R.T=t):e.current=s,Du?(Du=!1,Lt=e,mn=a):or(e,r),r=e.pendingLanes,r===0&&(at=null),Fd(s.stateNode),pl(e),l!==null)for(n=e.onRecoverableError,s=0;s<l.length;s++)r=l[s],n(r.value,{componentStack:r.stack});return mn&3&&ba(),r=e.pendingLanes,a&4194218&&r&42?e===Ec?hn++:(hn=0,Ec=e):hn=0,gn(0),null}function or(e,l){(e.pooledCacheLanes&=l)===0&&(l=e.pooledCache,l!=null&&(e.pooledCache=null,$a(l)))}function ba(){if(Lt!==null){var e=Lt,l=Dc;Dc=0;var t=ps(mn),a=R.T,n=V.p;try{if(V.p=32>t?32:t,R.T=null,Lt===null)var u=!1;else{t=Nc,Nc=null;var i=Lt,c=mn;if(Lt=null,mn=0,ue&6)throw Error(d(331));var s=ue;if(ue|=4,kf(i.current),Vf(i,i.current,c,t),ue=s,gn(0,!1),_e&&typeof _e.onPostCommitFiberRoot=="function")try{_e.onPostCommitFiberRoot(Ca,i)}catch{}u=!0}return u}finally{V.p=n,R.T=a,or(e,l)}}return!1}function fr(e,l,t){l=Ze(t,l),l=Ji(e.stateNode,l,2),e=Wl(e,l,2),e!==null&&(ja(e,2),pl(e))}function P(e,l,t){if(e.tag===3)fr(e,e,t);else for(;l!==null;){if(l.tag===3){fr(l,e,t);break}else if(l.tag===1){var a=l.stateNode;if(typeof l.type.getDerivedStateFromError=="function"||typeof a.componentDidCatch=="function"&&(at===null||!at.has(a))){e=Ze(t,e),t=mf(2),a=Wl(l,t,2),a!==null&&(hf(t,a,l,e),ja(a,2),pl(a));break}}l=l.return}}function jc(e,l,t){var a=e.pingCache;if(a===null){a=e.pingCache=new yh;var n=new Set;a.set(l,n)}else n=a.get(l),n===void 0&&(n=new Set,a.set(l,n));n.has(t)||(Sc=!0,n.add(t),e=Ah.bind(null,e,l,t),l.then(e,e))}function Ah(e,l,t){var a=e.pingCache;a!==null&&a.delete(l),e.pingedLanes|=e.suspendedLanes&t,e.warmLanes&=~t,I===e&&(Q&t)===t&&(ce===4||ce===3&&(Q&62914560)===Q&&300>rl()-Ac?!(ue&2)&&ya(e,0):Tc|=t,pa===Q&&(pa=0)),pl(e)}function rr(e,l){l===0&&(l=ms()),e=Ql(e,l),e!==null&&(ja(e,l),pl(e))}function Dh(e){var l=e.memoizedState,t=0;l!==null&&(t=l.retryLane),rr(e,t)}function Nh(e,l){var t=0;switch(e.tag){case 13:var a=e.stateNode,n=e.memoizedState;n!==null&&(t=n.retryLane);break;case 19:a=e.stateNode;break;case 22:a=e.stateNode._retryCache;break;default:throw Error(d(314))}a!==null&&a.delete(l),rr(e,t)}function Eh(e,l){return Fu(e,l)}var Mu=null,Sa=null,Rc=!1,zu=!1,Uc=!1,qt=0;function pl(e){e!==Sa&&e.next===null&&(Sa===null?Mu=Sa=e:Sa=Sa.next=e),zu=!0,Rc||(Rc=!0,zh(Mh))}function gn(e,l){if(!Uc&&zu){Uc=!0;do for(var t=!1,a=Mu;a!==null;){if(e!==0){var n=a.pendingLanes;if(n===0)var u=0;else{var i=a.suspendedLanes,c=a.pingedLanes;u=(1<<31-He(42|e)+1)-1,u&=n&~(i&~c),u=u&201326677?u&201326677|1:u?u|2:0}u!==0&&(t=!0,hr(a,u))}else u=Q,u=qn(a,a===I?u:0),!(u&3)||Oa(a,u)||(t=!0,hr(a,u));a=a.next}while(t);Uc=!1}}function Mh(){zu=Rc=!1;var e=0;qt!==0&&(Lh()&&(e=qt),qt=0);for(var l=rl(),t=null,a=Mu;a!==null;){var n=a.next,u=dr(a,l);u===0?(a.next=null,t===null?Mu=n:t.next=n,n===null&&(Sa=t)):(t=a,(e!==0||u&3)&&(zu=!0)),a=n}gn(e)}function dr(e,l){for(var t=e.suspendedLanes,a=e.pingedLanes,n=e.expirationTimes,u=e.pendingLanes&-62914561;0<u;){var i=31-He(u),c=1<<i,s=n[i];s===-1?(!(c&t)||c&a)&&(n[i]=Id(c,l)):s<=l&&(e.expiredLanes|=c),u&=~c}if(l=I,t=Q,t=qn(e,e===l?t:0),a=e.callbackNode,t===0||e===l&&ee===2||e.cancelPendingCommit!==null)return a!==null&&a!==null&&$u(a),e.callbackNode=null,e.callbackPriority=0;if(!(t&3)||Oa(e,t)){if(l=t&-t,l===e.callbackPriority)return l;switch(a!==null&&$u(a),ps(t)){case 2:case 8:t=fs;break;case 32:t=_n;break;case 268435456:t=rs;break;default:t=_n}return a=mr.bind(null,e),t=Fu(t,a),e.callbackPriority=l,e.callbackNode=t,l}return a!==null&&a!==null&&$u(a),e.callbackPriority=2,e.callbackNode=null,2}function mr(e,l){var t=e.callbackNode;if(ba()&&e.callbackNode!==t)return null;var a=Q;return a=qn(e,e===I?a:0),a===0?null:(er(e,a,l),dr(e,rl()),e.callbackNode!=null&&e.callbackNode===t?mr.bind(null,e):null)}function hr(e,l){if(ba())return null;er(e,l,!0)}function zh(e){Bh(function(){ue&6?Fu(os,e):e()})}function _c(){return qt===0&&(qt=ds()),qt}function gr(e){return e==null||typeof e=="symbol"||typeof e=="boolean"?null:typeof e=="function"?e:Xn(""+e)}function pr(e,l){var t=l.ownerDocument.createElement("input");return t.name=l.name,t.value=l.value,e.id&&t.setAttribute("form",e.id),l.parentNode.insertBefore(t,l),e=new FormData(e),t.parentNode.removeChild(t),e}function Ch(e,l,t,a,n){if(l==="submit"&&t&&t.stateNode===n){var u=gr((n[je]||null).action),i=a.submitter;i&&(l=(l=i[je]||null)?gr(l.formAction):i.getAttribute("formAction"),l!==null&&(u=l,i=null));var c=new Zn("action","action",null,a,n);e.push({event:c,listeners:[{instance:null,listener:function(){if(a.defaultPrevented){if(qt!==0){var s=i?pr(n,i):new FormData(n);Qi(t,{pending:!0,data:s,method:n.method,action:u},null,s)}}else typeof u=="function"&&(c.preventDefault(),s=i?pr(n,i):new FormData(n),Qi(t,{pending:!0,data:s,method:n.method,action:u},u,s))},currentTarget:n}]})}}for(var Hc=0;Hc<io.length;Hc++){var Lc=io[Hc],Oh=Lc.toLowerCase(),jh=Lc[0].toUpperCase()+Lc.slice(1);ul(Oh,"on"+jh)}ul(lo,"onAnimationEnd"),ul(to,"onAnimationIteration"),ul(ao,"onAnimationStart"),ul("dblclick","onDoubleClick"),ul("focusin","onFocus"),ul("focusout","onBlur"),ul(Jm,"onTransitionRun"),ul(Fm,"onTransitionStart"),ul($m,"onTransitionCancel"),ul(no,"onTransitionEnd"),kt("onMouseEnter",["mouseout","mouseover"]),kt("onMouseLeave",["mouseout","mouseover"]),kt("onPointerEnter",["pointerout","pointerover"]),kt("onPointerLeave",["pointerout","pointerover"]),vt("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),vt("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),vt("onBeforeInput",["compositionend","keypress","textInput","paste"]),vt("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),vt("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),vt("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var pn="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),Rh=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(pn));function yr(e,l){l=(l&4)!==0;for(var t=0;t<e.length;t++){var a=e[t],n=a.event;a=a.listeners;e:{var u=void 0;if(l)for(var i=a.length-1;0<=i;i--){var c=a[i],s=c.instance,r=c.currentTarget;if(c=c.listener,s!==u&&n.isPropagationStopped())break e;u=c,n.currentTarget=r;try{u(n)}catch(v){hu(v)}n.currentTarget=null,u=s}else for(i=0;i<a.length;i++){if(c=a[i],s=c.instance,r=c.currentTarget,c=c.listener,s!==u&&n.isPropagationStopped())break e;u=c,n.currentTarget=r;try{u(n)}catch(v){hu(v)}n.currentTarget=null,u=s}}}}function w(e,l){var t=l[Pu];t===void 0&&(t=l[Pu]=new Set);var a=e+"__bubble";t.has(a)||(vr(l,e,2,!1),t.add(a))}function qc(e,l,t){var a=0;l&&(a|=4),vr(t,e,a,l)}var Cu="_reactListening"+Math.random().toString(36).slice(2);function Bc(e){if(!e[Cu]){e[Cu]=!0,bs.forEach(function(t){t!=="selectionchange"&&(Rh.has(t)||qc(t,!1,e),qc(t,!0,e))});var l=e.nodeType===9?e:e.ownerDocument;l===null||l[Cu]||(l[Cu]=!0,qc("selectionchange",!1,l))}}function vr(e,l,t,a){switch(wr(l)){case 2:var n=ng;break;case 8:n=ug;break;default:n=Wc}t=n.bind(null,l,t,e),n=void 0,!ii||l!=="touchstart"&&l!=="touchmove"&&l!=="wheel"||(n=!0),a?n!==void 0?e.addEventListener(l,t,{capture:!0,passive:n}):e.addEventListener(l,t,!0):n!==void 0?e.addEventListener(l,t,{passive:n}):e.addEventListener(l,t,!1)}function Gc(e,l,t,a,n){var u=a;if(!(l&1)&&!(l&2)&&a!==null)e:for(;;){if(a===null)return;var i=a.tag;if(i===3||i===4){var c=a.stateNode.containerInfo;if(c===n||c.nodeType===8&&c.parentNode===n)break;if(i===4)for(i=a.return;i!==null;){var s=i.tag;if((s===3||s===4)&&(s=i.stateNode.containerInfo,s===n||s.nodeType===8&&s.parentNode===n))return;i=i.return}for(;c!==null;){if(i=yt(c),i===null)return;if(s=i.tag,s===5||s===6||s===26||s===27){a=u=i;continue e}c=c.parentNode}}a=a.return}js(function(){var r=u,v=ni(t),T=[];e:{var p=uo.get(e);if(p!==void 0){var y=Zn,M=e;switch(e){case"keypress":if(Vn(t)===0)break e;case"keydown":case"keyup":y=Nm;break;case"focusin":M="focus",y=fi;break;case"focusout":M="blur",y=fi;break;case"beforeblur":case"afterblur":y=fi;break;case"click":if(t.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":y=_s;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":y=mm;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":y=zm;break;case lo:case to:case ao:y=pm;break;case no:y=Om;break;case"scroll":case"scrollend":y=rm;break;case"wheel":y=Rm;break;case"copy":case"cut":case"paste":y=vm;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":y=Ls;break;case"toggle":case"beforetoggle":y=_m}var _=(l&4)!==0,se=!_&&(e==="scroll"||e==="scrollend"),m=_?p!==null?p+"Capture":null:p;_=[];for(var f=r,h;f!==null;){var S=f;if(h=S.stateNode,S=S.tag,S!==5&&S!==26&&S!==27||h===null||m===null||(S=_a(f,m),S!=null&&_.push(yn(f,S,h))),se)break;f=f.return}0<_.length&&(p=new y(p,M,null,t,v),T.push({event:p,listeners:_}))}}if(!(l&7)){e:{if(p=e==="mouseover"||e==="pointerover",y=e==="mouseout"||e==="pointerout",p&&t!==ai&&(M=t.relatedTarget||t.fromElement)&&(yt(M)||M[Vt]))break e;if((y||p)&&(p=v.window===v?v:(p=v.ownerDocument)?p.defaultView||p.parentWindow:window,y?(M=t.relatedTarget||t.toElement,y=r,M=M?yt(M):null,M!==null&&(se=U(M),_=M.tag,M!==se||_!==5&&_!==27&&_!==6)&&(M=null)):(y=null,M=r),y!==M)){if(_=_s,S="onMouseLeave",m="onMouseEnter",f="mouse",(e==="pointerout"||e==="pointerover")&&(_=Ls,S="onPointerLeave",m="onPointerEnter",f="pointer"),se=y==null?p:Ua(y),h=M==null?p:Ua(M),p=new _(S,f+"leave",y,t,v),p.target=se,p.relatedTarget=h,S=null,yt(v)===r&&(_=new _(m,f+"enter",M,t,v),_.target=h,_.relatedTarget=se,S=_),se=S,y&&M)l:{for(_=y,m=M,f=0,h=_;h;h=Ta(h))f++;for(h=0,S=m;S;S=Ta(S))h++;for(;0<f-h;)_=Ta(_),f--;for(;0<h-f;)m=Ta(m),h--;for(;f--;){if(_===m||m!==null&&_===m.alternate)break l;_=Ta(_),m=Ta(m)}_=null}else _=null;y!==null&&br(T,p,y,_,!1),M!==null&&se!==null&&br(T,se,M,_,!0)}}e:{if(p=r?Ua(r):window,y=p.nodeName&&p.nodeName.toLowerCase(),y==="select"||y==="input"&&p.type==="file")var D=Vs;else if(Xs(p))if(Ks)D=Vm;else{D=Xm;var B=wm}else y=p.nodeName,!y||y.toLowerCase()!=="input"||p.type!=="checkbox"&&p.type!=="radio"?r&&ti(r.elementType)&&(D=Vs):D=Qm;if(D&&(D=D(e,r))){Qs(T,D,t,v);break e}B&&B(e,p,r),e==="focusout"&&r&&p.type==="number"&&r.memoizedProps.value!=null&&li(p,"number",p.value)}switch(B=r?Ua(r):window,e){case"focusin":(Xs(B)||B.contentEditable==="true")&&(It=B,pi=r,Xa=null);break;case"focusout":Xa=pi=It=null;break;case"mousedown":yi=!0;break;case"contextmenu":case"mouseup":case"dragend":yi=!1,Is(T,t,v);break;case"selectionchange":if(km)break;case"keydown":case"keyup":Is(T,t,v)}var z;if(di)e:{switch(e){case"compositionstart":var O="onCompositionStart";break e;case"compositionend":O="onCompositionEnd";break e;case"compositionupdate":O="onCompositionUpdate";break e}O=void 0}else Pt?Ys(e,t)&&(O="onCompositionEnd"):e==="keydown"&&t.keyCode===229&&(O="onCompositionStart");O&&(qs&&t.locale!=="ko"&&(Pt||O!=="onCompositionStart"?O==="onCompositionEnd"&&Pt&&(z=Rs()):(Xl=v,ci="value"in Xl?Xl.value:Xl.textContent,Pt=!0)),B=Ou(r,O),0<B.length&&(O=new Hs(O,e,null,t,v),T.push({event:O,listeners:B}),z?O.data=z:(z=ws(t),z!==null&&(O.data=z)))),(z=Lm?qm(e,t):Bm(e,t))&&(O=Ou(r,"onBeforeInput"),0<O.length&&(B=new Hs("onBeforeInput","beforeinput",null,t,v),T.push({event:B,listeners:O}),B.data=z)),Ch(T,e,r,t,v)}yr(T,l)})}function yn(e,l,t){return{instance:e,listener:l,currentTarget:t}}function Ou(e,l){for(var t=l+"Capture",a=[];e!==null;){var n=e,u=n.stateNode;n=n.tag,n!==5&&n!==26&&n!==27||u===null||(n=_a(e,t),n!=null&&a.unshift(yn(e,n,u)),n=_a(e,l),n!=null&&a.push(yn(e,n,u))),e=e.return}return a}function Ta(e){if(e===null)return null;do e=e.return;while(e&&e.tag!==5&&e.tag!==27);return e||null}function br(e,l,t,a,n){for(var u=l._reactName,i=[];t!==null&&t!==a;){var c=t,s=c.alternate,r=c.stateNode;if(c=c.tag,s!==null&&s===a)break;c!==5&&c!==26&&c!==27||r===null||(s=r,n?(r=_a(t,u),r!=null&&i.unshift(yn(t,r,s))):n||(r=_a(t,u),r!=null&&i.push(yn(t,r,s)))),t=t.return}i.length!==0&&e.push({event:l,listeners:i})}var Uh=/\r\n?/g,_h=/\u0000|\uFFFD/g;function Sr(e){return(typeof e=="string"?e:""+e).replace(Uh,`
`).replace(_h,"")}function Tr(e,l){return l=Sr(l),Sr(e)===l}function ju(){}function F(e,l,t,a,n,u){switch(t){case"children":typeof a=="string"?l==="body"||l==="textarea"&&a===""||Ft(e,a):(typeof a=="number"||typeof a=="bigint")&&l!=="body"&&Ft(e,""+a);break;case"className":Gn(e,"class",a);break;case"tabIndex":Gn(e,"tabindex",a);break;case"dir":case"role":case"viewBox":case"width":case"height":Gn(e,t,a);break;case"style":Cs(e,a,u);break;case"data":if(l!=="object"){Gn(e,"data",a);break}case"src":case"href":if(a===""&&(l!=="a"||t!=="href")){e.removeAttribute(t);break}if(a==null||typeof a=="function"||typeof a=="symbol"||typeof a=="boolean"){e.removeAttribute(t);break}a=Xn(""+a),e.setAttribute(t,a);break;case"action":case"formAction":if(typeof a=="function"){e.setAttribute(t,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof u=="function"&&(t==="formAction"?(l!=="input"&&F(e,l,"name",n.name,n,null),F(e,l,"formEncType",n.formEncType,n,null),F(e,l,"formMethod",n.formMethod,n,null),F(e,l,"formTarget",n.formTarget,n,null)):(F(e,l,"encType",n.encType,n,null),F(e,l,"method",n.method,n,null),F(e,l,"target",n.target,n,null)));if(a==null||typeof a=="symbol"||typeof a=="boolean"){e.removeAttribute(t);break}a=Xn(""+a),e.setAttribute(t,a);break;case"onClick":a!=null&&(e.onclick=ju);break;case"onScroll":a!=null&&w("scroll",e);break;case"onScrollEnd":a!=null&&w("scrollend",e);break;case"dangerouslySetInnerHTML":if(a!=null){if(typeof a!="object"||!("__html"in a))throw Error(d(61));if(t=a.__html,t!=null){if(n.children!=null)throw Error(d(60));e.innerHTML=t}}break;case"multiple":e.multiple=a&&typeof a!="function"&&typeof a!="symbol";break;case"muted":e.muted=a&&typeof a!="function"&&typeof a!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(a==null||typeof a=="function"||typeof a=="boolean"||typeof a=="symbol"){e.removeAttribute("xlink:href");break}t=Xn(""+a),e.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",t);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":a!=null&&typeof a!="function"&&typeof a!="symbol"?e.setAttribute(t,""+a):e.removeAttribute(t);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":a&&typeof a!="function"&&typeof a!="symbol"?e.setAttribute(t,""):e.removeAttribute(t);break;case"capture":case"download":a===!0?e.setAttribute(t,""):a!==!1&&a!=null&&typeof a!="function"&&typeof a!="symbol"?e.setAttribute(t,a):e.removeAttribute(t);break;case"cols":case"rows":case"size":case"span":a!=null&&typeof a!="function"&&typeof a!="symbol"&&!isNaN(a)&&1<=a?e.setAttribute(t,a):e.removeAttribute(t);break;case"rowSpan":case"start":a==null||typeof a=="function"||typeof a=="symbol"||isNaN(a)?e.removeAttribute(t):e.setAttribute(t,a);break;case"popover":w("beforetoggle",e),w("toggle",e),Bn(e,"popover",a);break;case"xlinkActuate":Sl(e,"http://www.w3.org/1999/xlink","xlink:actuate",a);break;case"xlinkArcrole":Sl(e,"http://www.w3.org/1999/xlink","xlink:arcrole",a);break;case"xlinkRole":Sl(e,"http://www.w3.org/1999/xlink","xlink:role",a);break;case"xlinkShow":Sl(e,"http://www.w3.org/1999/xlink","xlink:show",a);break;case"xlinkTitle":Sl(e,"http://www.w3.org/1999/xlink","xlink:title",a);break;case"xlinkType":Sl(e,"http://www.w3.org/1999/xlink","xlink:type",a);break;case"xmlBase":Sl(e,"http://www.w3.org/XML/1998/namespace","xml:base",a);break;case"xmlLang":Sl(e,"http://www.w3.org/XML/1998/namespace","xml:lang",a);break;case"xmlSpace":Sl(e,"http://www.w3.org/XML/1998/namespace","xml:space",a);break;case"is":Bn(e,"is",a);break;case"innerText":case"textContent":break;default:(!(2<t.length)||t[0]!=="o"&&t[0]!=="O"||t[1]!=="n"&&t[1]!=="N")&&(t=om.get(t)||t,Bn(e,t,a))}}function Yc(e,l,t,a,n,u){switch(t){case"style":Cs(e,a,u);break;case"dangerouslySetInnerHTML":if(a!=null){if(typeof a!="object"||!("__html"in a))throw Error(d(61));if(t=a.__html,t!=null){if(n.children!=null)throw Error(d(60));e.innerHTML=t}}break;case"children":typeof a=="string"?Ft(e,a):(typeof a=="number"||typeof a=="bigint")&&Ft(e,""+a);break;case"onScroll":a!=null&&w("scroll",e);break;case"onScrollEnd":a!=null&&w("scrollend",e);break;case"onClick":a!=null&&(e.onclick=ju);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!Ss.hasOwnProperty(t))e:{if(t[0]==="o"&&t[1]==="n"&&(n=t.endsWith("Capture"),l=t.slice(2,n?t.length-7:void 0),u=e[je]||null,u=u!=null?u[t]:null,typeof u=="function"&&e.removeEventListener(l,u,n),typeof a=="function")){typeof u!="function"&&u!==null&&(t in e?e[t]=null:e.hasAttribute(t)&&e.removeAttribute(t)),e.addEventListener(l,a,n);break e}t in e?e[t]=a:a===!0?e.setAttribute(t,""):Bn(e,t,a)}}}function Ee(e,l,t){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":w("error",e),w("load",e);var a=!1,n=!1,u;for(u in t)if(t.hasOwnProperty(u)){var i=t[u];if(i!=null)switch(u){case"src":a=!0;break;case"srcSet":n=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(d(137,l));default:F(e,l,u,i,t,null)}}n&&F(e,l,"srcSet",t.srcSet,t,null),a&&F(e,l,"src",t.src,t,null);return;case"input":w("invalid",e);var c=u=i=n=null,s=null,r=null;for(a in t)if(t.hasOwnProperty(a)){var v=t[a];if(v!=null)switch(a){case"name":n=v;break;case"type":i=v;break;case"checked":s=v;break;case"defaultChecked":r=v;break;case"value":u=v;break;case"defaultValue":c=v;break;case"children":case"dangerouslySetInnerHTML":if(v!=null)throw Error(d(137,l));break;default:F(e,l,a,v,t,null)}}Ns(e,u,c,s,r,i,n,!1),Yn(e);return;case"select":w("invalid",e),a=i=u=null;for(n in t)if(t.hasOwnProperty(n)&&(c=t[n],c!=null))switch(n){case"value":u=c;break;case"defaultValue":i=c;break;case"multiple":a=c;default:F(e,l,n,c,t,null)}l=u,t=i,e.multiple=!!a,l!=null?Jt(e,!!a,l,!1):t!=null&&Jt(e,!!a,t,!0);return;case"textarea":w("invalid",e),u=n=a=null;for(i in t)if(t.hasOwnProperty(i)&&(c=t[i],c!=null))switch(i){case"value":a=c;break;case"defaultValue":n=c;break;case"children":u=c;break;case"dangerouslySetInnerHTML":if(c!=null)throw Error(d(91));break;default:F(e,l,i,c,t,null)}Ms(e,a,n,u),Yn(e);return;case"option":for(s in t)if(t.hasOwnProperty(s)&&(a=t[s],a!=null))switch(s){case"selected":e.selected=a&&typeof a!="function"&&typeof a!="symbol";break;default:F(e,l,s,a,t,null)}return;case"dialog":w("cancel",e),w("close",e);break;case"iframe":case"object":w("load",e);break;case"video":case"audio":for(a=0;a<pn.length;a++)w(pn[a],e);break;case"image":w("error",e),w("load",e);break;case"details":w("toggle",e);break;case"embed":case"source":case"link":w("error",e),w("load",e);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(r in t)if(t.hasOwnProperty(r)&&(a=t[r],a!=null))switch(r){case"children":case"dangerouslySetInnerHTML":throw Error(d(137,l));default:F(e,l,r,a,t,null)}return;default:if(ti(l)){for(v in t)t.hasOwnProperty(v)&&(a=t[v],a!==void 0&&Yc(e,l,v,a,t,void 0));return}}for(c in t)t.hasOwnProperty(c)&&(a=t[c],a!=null&&F(e,l,c,a,t,null))}function Hh(e,l,t,a){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var n=null,u=null,i=null,c=null,s=null,r=null,v=null;for(y in t){var T=t[y];if(t.hasOwnProperty(y)&&T!=null)switch(y){case"checked":break;case"value":break;case"defaultValue":s=T;default:a.hasOwnProperty(y)||F(e,l,y,null,a,T)}}for(var p in a){var y=a[p];if(T=t[p],a.hasOwnProperty(p)&&(y!=null||T!=null))switch(p){case"type":u=y;break;case"name":n=y;break;case"checked":r=y;break;case"defaultChecked":v=y;break;case"value":i=y;break;case"defaultValue":c=y;break;case"children":case"dangerouslySetInnerHTML":if(y!=null)throw Error(d(137,l));break;default:y!==T&&F(e,l,p,y,a,T)}}ei(e,i,c,s,r,v,u,n);return;case"select":y=i=c=p=null;for(u in t)if(s=t[u],t.hasOwnProperty(u)&&s!=null)switch(u){case"value":break;case"multiple":y=s;default:a.hasOwnProperty(u)||F(e,l,u,null,a,s)}for(n in a)if(u=a[n],s=t[n],a.hasOwnProperty(n)&&(u!=null||s!=null))switch(n){case"value":p=u;break;case"defaultValue":c=u;break;case"multiple":i=u;default:u!==s&&F(e,l,n,u,a,s)}l=c,t=i,a=y,p!=null?Jt(e,!!t,p,!1):!!a!=!!t&&(l!=null?Jt(e,!!t,l,!0):Jt(e,!!t,t?[]:"",!1));return;case"textarea":y=p=null;for(c in t)if(n=t[c],t.hasOwnProperty(c)&&n!=null&&!a.hasOwnProperty(c))switch(c){case"value":break;case"children":break;default:F(e,l,c,null,a,n)}for(i in a)if(n=a[i],u=t[i],a.hasOwnProperty(i)&&(n!=null||u!=null))switch(i){case"value":p=n;break;case"defaultValue":y=n;break;case"children":break;case"dangerouslySetInnerHTML":if(n!=null)throw Error(d(91));break;default:n!==u&&F(e,l,i,n,a,u)}Es(e,p,y);return;case"option":for(var M in t)if(p=t[M],t.hasOwnProperty(M)&&p!=null&&!a.hasOwnProperty(M))switch(M){case"selected":e.selected=!1;break;default:F(e,l,M,null,a,p)}for(s in a)if(p=a[s],y=t[s],a.hasOwnProperty(s)&&p!==y&&(p!=null||y!=null))switch(s){case"selected":e.selected=p&&typeof p!="function"&&typeof p!="symbol";break;default:F(e,l,s,p,a,y)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var _ in t)p=t[_],t.hasOwnProperty(_)&&p!=null&&!a.hasOwnProperty(_)&&F(e,l,_,null,a,p);for(r in a)if(p=a[r],y=t[r],a.hasOwnProperty(r)&&p!==y&&(p!=null||y!=null))switch(r){case"children":case"dangerouslySetInnerHTML":if(p!=null)throw Error(d(137,l));break;default:F(e,l,r,p,a,y)}return;default:if(ti(l)){for(var se in t)p=t[se],t.hasOwnProperty(se)&&p!==void 0&&!a.hasOwnProperty(se)&&Yc(e,l,se,void 0,a,p);for(v in a)p=a[v],y=t[v],!a.hasOwnProperty(v)||p===y||p===void 0&&y===void 0||Yc(e,l,v,p,a,y);return}}for(var m in t)p=t[m],t.hasOwnProperty(m)&&p!=null&&!a.hasOwnProperty(m)&&F(e,l,m,null,a,p);for(T in a)p=a[T],y=t[T],!a.hasOwnProperty(T)||p===y||p==null&&y==null||F(e,l,T,p,a,y)}var wc=null,Xc=null;function Ru(e){return e.nodeType===9?e:e.ownerDocument}function xr(e){switch(e){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function Ar(e,l){if(e===0)switch(l){case"svg":return 1;case"math":return 2;default:return 0}return e===1&&l==="foreignObject"?0:e}function Qc(e,l){return e==="textarea"||e==="noscript"||typeof l.children=="string"||typeof l.children=="number"||typeof l.children=="bigint"||typeof l.dangerouslySetInnerHTML=="object"&&l.dangerouslySetInnerHTML!==null&&l.dangerouslySetInnerHTML.__html!=null}var Vc=null;function Lh(){var e=window.event;return e&&e.type==="popstate"?e===Vc?!1:(Vc=e,!0):(Vc=null,!1)}var Dr=typeof setTimeout=="function"?setTimeout:void 0,qh=typeof clearTimeout=="function"?clearTimeout:void 0,Nr=typeof Promise=="function"?Promise:void 0,Bh=typeof queueMicrotask=="function"?queueMicrotask:typeof Nr<"u"?function(e){return Nr.resolve(null).then(e).catch(Gh)}:Dr;function Gh(e){setTimeout(function(){throw e})}function Kc(e,l){var t=l,a=0;do{var n=t.nextSibling;if(e.removeChild(t),n&&n.nodeType===8)if(t=n.data,t==="/$"){if(a===0){e.removeChild(n),Nn(l);return}a--}else t!=="$"&&t!=="$?"&&t!=="$!"||a++;t=n}while(t);Nn(l)}function Zc(e){var l=e.firstChild;for(l&&l.nodeType===10&&(l=l.nextSibling);l;){var t=l;switch(l=l.nextSibling,t.nodeName){case"HTML":case"HEAD":case"BODY":Zc(t),Iu(t);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(t.rel.toLowerCase()==="stylesheet")continue}e.removeChild(t)}}function Yh(e,l,t,a){for(;e.nodeType===1;){var n=t;if(e.nodeName.toLowerCase()!==l.toLowerCase()){if(!a&&(e.nodeName!=="INPUT"||e.type!=="hidden"))break}else if(a){if(!e[Ra])switch(l){case"meta":if(!e.hasAttribute("itemprop"))break;return e;case"link":if(u=e.getAttribute("rel"),u==="stylesheet"&&e.hasAttribute("data-precedence"))break;if(u!==n.rel||e.getAttribute("href")!==(n.href==null?null:n.href)||e.getAttribute("crossorigin")!==(n.crossOrigin==null?null:n.crossOrigin)||e.getAttribute("title")!==(n.title==null?null:n.title))break;return e;case"style":if(e.hasAttribute("data-precedence"))break;return e;case"script":if(u=e.getAttribute("src"),(u!==(n.src==null?null:n.src)||e.getAttribute("type")!==(n.type==null?null:n.type)||e.getAttribute("crossorigin")!==(n.crossOrigin==null?null:n.crossOrigin))&&u&&e.hasAttribute("async")&&!e.hasAttribute("itemprop"))break;return e;default:return e}}else if(l==="input"&&e.type==="hidden"){var u=n.name==null?null:""+n.name;if(n.type==="hidden"&&e.getAttribute("name")===u)return e}else return e;if(e=sl(e.nextSibling),e===null)break}return null}function wh(e,l,t){if(l==="")return null;for(;e.nodeType!==3;)if((e.nodeType!==1||e.nodeName!=="INPUT"||e.type!=="hidden")&&!t||(e=sl(e.nextSibling),e===null))return null;return e}function sl(e){for(;e!=null;e=e.nextSibling){var l=e.nodeType;if(l===1||l===3)break;if(l===8){if(l=e.data,l==="$"||l==="$!"||l==="$?"||l==="F!"||l==="F")break;if(l==="/$")return null}}return e}function Er(e){e=e.previousSibling;for(var l=0;e;){if(e.nodeType===8){var t=e.data;if(t==="$"||t==="$!"||t==="$?"){if(l===0)return e;l--}else t==="/$"&&l++}e=e.previousSibling}return null}function Mr(e,l,t){switch(l=Ru(t),e){case"html":if(e=l.documentElement,!e)throw Error(d(452));return e;case"head":if(e=l.head,!e)throw Error(d(453));return e;case"body":if(e=l.body,!e)throw Error(d(454));return e;default:throw Error(d(451))}}var el=new Map,zr=new Set;function Uu(e){return typeof e.getRootNode=="function"?e.getRootNode():e.ownerDocument}var Hl=V.d;V.d={f:Xh,r:Qh,D:Vh,C:Kh,L:Zh,m:kh,X:Fh,S:Jh,M:$h};function Xh(){var e=Hl.f(),l=Nu();return e||l}function Qh(e){var l=Kt(e);l!==null&&l.tag===5&&l.type==="form"?ef(l):Hl.r(e)}var xa=typeof document>"u"?null:document;function Cr(e,l,t){var a=xa;if(a&&typeof l=="string"&&l){var n=Ve(l);n='link[rel="'+e+'"][href="'+n+'"]',typeof t=="string"&&(n+='[crossorigin="'+t+'"]'),zr.has(n)||(zr.add(n),e={rel:e,crossOrigin:t,href:l},a.querySelector(n)===null&&(l=a.createElement("link"),Ee(l,"link",e),Se(l),a.head.appendChild(l)))}}function Vh(e){Hl.D(e),Cr("dns-prefetch",e,null)}function Kh(e,l){Hl.C(e,l),Cr("preconnect",e,l)}function Zh(e,l,t){Hl.L(e,l,t);var a=xa;if(a&&e&&l){var n='link[rel="preload"][as="'+Ve(l)+'"]';l==="image"&&t&&t.imageSrcSet?(n+='[imagesrcset="'+Ve(t.imageSrcSet)+'"]',typeof t.imageSizes=="string"&&(n+='[imagesizes="'+Ve(t.imageSizes)+'"]')):n+='[href="'+Ve(e)+'"]';var u=n;switch(l){case"style":u=Aa(e);break;case"script":u=Da(e)}el.has(u)||(e=k({rel:"preload",href:l==="image"&&t&&t.imageSrcSet?void 0:e,as:l},t),el.set(u,e),a.querySelector(n)!==null||l==="style"&&a.querySelector(vn(u))||l==="script"&&a.querySelector(bn(u))||(l=a.createElement("link"),Ee(l,"link",e),Se(l),a.head.appendChild(l)))}}function kh(e,l){Hl.m(e,l);var t=xa;if(t&&e){var a=l&&typeof l.as=="string"?l.as:"script",n='link[rel="modulepreload"][as="'+Ve(a)+'"][href="'+Ve(e)+'"]',u=n;switch(a){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":u=Da(e)}if(!el.has(u)&&(e=k({rel:"modulepreload",href:e},l),el.set(u,e),t.querySelector(n)===null)){switch(a){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(t.querySelector(bn(u)))return}a=t.createElement("link"),Ee(a,"link",e),Se(a),t.head.appendChild(a)}}}function Jh(e,l,t){Hl.S(e,l,t);var a=xa;if(a&&e){var n=Zt(a).hoistableStyles,u=Aa(e);l=l||"default";var i=n.get(u);if(!i){var c={loading:0,preload:null};if(i=a.querySelector(vn(u)))c.loading=5;else{e=k({rel:"stylesheet",href:e,"data-precedence":l},t),(t=el.get(u))&&kc(e,t);var s=i=a.createElement("link");Se(s),Ee(s,"link",e),s._p=new Promise(function(r,v){s.onload=r,s.onerror=v}),s.addEventListener("load",function(){c.loading|=1}),s.addEventListener("error",function(){c.loading|=2}),c.loading|=4,_u(i,l,a)}i={type:"stylesheet",instance:i,count:1,state:c},n.set(u,i)}}}function Fh(e,l){Hl.X(e,l);var t=xa;if(t&&e){var a=Zt(t).hoistableScripts,n=Da(e),u=a.get(n);u||(u=t.querySelector(bn(n)),u||(e=k({src:e,async:!0},l),(l=el.get(n))&&Jc(e,l),u=t.createElement("script"),Se(u),Ee(u,"link",e),t.head.appendChild(u)),u={type:"script",instance:u,count:1,state:null},a.set(n,u))}}function $h(e,l){Hl.M(e,l);var t=xa;if(t&&e){var a=Zt(t).hoistableScripts,n=Da(e),u=a.get(n);u||(u=t.querySelector(bn(n)),u||(e=k({src:e,async:!0,type:"module"},l),(l=el.get(n))&&Jc(e,l),u=t.createElement("script"),Se(u),Ee(u,"link",e),t.head.appendChild(u)),u={type:"script",instance:u,count:1,state:null},a.set(n,u))}}function Or(e,l,t,a){var n=(n=Gl.current)?Uu(n):null;if(!n)throw Error(d(446));switch(e){case"meta":case"title":return null;case"style":return typeof t.precedence=="string"&&typeof t.href=="string"?(l=Aa(t.href),t=Zt(n).hoistableStyles,a=t.get(l),a||(a={type:"style",instance:null,count:0,state:null},t.set(l,a)),a):{type:"void",instance:null,count:0,state:null};case"link":if(t.rel==="stylesheet"&&typeof t.href=="string"&&typeof t.precedence=="string"){e=Aa(t.href);var u=Zt(n).hoistableStyles,i=u.get(e);if(i||(n=n.ownerDocument||n,i={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},u.set(e,i),(u=n.querySelector(vn(e)))&&!u._p&&(i.instance=u,i.state.loading=5),el.has(e)||(t={rel:"preload",as:"style",href:t.href,crossOrigin:t.crossOrigin,integrity:t.integrity,media:t.media,hrefLang:t.hrefLang,referrerPolicy:t.referrerPolicy},el.set(e,t),u||Wh(n,e,t,i.state))),l&&a===null)throw Error(d(528,""));return i}if(l&&a!==null)throw Error(d(529,""));return null;case"script":return l=t.async,t=t.src,typeof t=="string"&&l&&typeof l!="function"&&typeof l!="symbol"?(l=Da(t),t=Zt(n).hoistableScripts,a=t.get(l),a||(a={type:"script",instance:null,count:0,state:null},t.set(l,a)),a):{type:"void",instance:null,count:0,state:null};default:throw Error(d(444,e))}}function Aa(e){return'href="'+Ve(e)+'"'}function vn(e){return'link[rel="stylesheet"]['+e+"]"}function jr(e){return k({},e,{"data-precedence":e.precedence,precedence:null})}function Wh(e,l,t,a){e.querySelector('link[rel="preload"][as="style"]['+l+"]")?a.loading=1:(l=e.createElement("link"),a.preload=l,l.addEventListener("load",function(){return a.loading|=1}),l.addEventListener("error",function(){return a.loading|=2}),Ee(l,"link",t),Se(l),e.head.appendChild(l))}function Da(e){return'[src="'+Ve(e)+'"]'}function bn(e){return"script[async]"+e}function Rr(e,l,t){if(l.count++,l.instance===null)switch(l.type){case"style":var a=e.querySelector('style[data-href~="'+Ve(t.href)+'"]');if(a)return l.instance=a,Se(a),a;var n=k({},t,{"data-href":t.href,"data-precedence":t.precedence,href:null,precedence:null});return a=(e.ownerDocument||e).createElement("style"),Se(a),Ee(a,"style",n),_u(a,t.precedence,e),l.instance=a;case"stylesheet":n=Aa(t.href);var u=e.querySelector(vn(n));if(u)return l.state.loading|=4,l.instance=u,Se(u),u;a=jr(t),(n=el.get(n))&&kc(a,n),u=(e.ownerDocument||e).createElement("link"),Se(u);var i=u;return i._p=new Promise(function(c,s){i.onload=c,i.onerror=s}),Ee(u,"link",a),l.state.loading|=4,_u(u,t.precedence,e),l.instance=u;case"script":return u=Da(t.src),(n=e.querySelector(bn(u)))?(l.instance=n,Se(n),n):(a=t,(n=el.get(u))&&(a=k({},t),Jc(a,n)),e=e.ownerDocument||e,n=e.createElement("script"),Se(n),Ee(n,"link",a),e.head.appendChild(n),l.instance=n);case"void":return null;default:throw Error(d(443,l.type))}else l.type==="stylesheet"&&!(l.state.loading&4)&&(a=l.instance,l.state.loading|=4,_u(a,t.precedence,e));return l.instance}function _u(e,l,t){for(var a=t.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),n=a.length?a[a.length-1]:null,u=n,i=0;i<a.length;i++){var c=a[i];if(c.dataset.precedence===l)u=c;else if(u!==n)break}u?u.parentNode.insertBefore(e,u.nextSibling):(l=t.nodeType===9?t.head:t,l.insertBefore(e,l.firstChild))}function kc(e,l){e.crossOrigin==null&&(e.crossOrigin=l.crossOrigin),e.referrerPolicy==null&&(e.referrerPolicy=l.referrerPolicy),e.title==null&&(e.title=l.title)}function Jc(e,l){e.crossOrigin==null&&(e.crossOrigin=l.crossOrigin),e.referrerPolicy==null&&(e.referrerPolicy=l.referrerPolicy),e.integrity==null&&(e.integrity=l.integrity)}var Hu=null;function Ur(e,l,t){if(Hu===null){var a=new Map,n=Hu=new Map;n.set(t,a)}else n=Hu,a=n.get(t),a||(a=new Map,n.set(t,a));if(a.has(e))return a;for(a.set(e,null),t=t.getElementsByTagName(e),n=0;n<t.length;n++){var u=t[n];if(!(u[Ra]||u[Me]||e==="link"&&u.getAttribute("rel")==="stylesheet")&&u.namespaceURI!=="http://www.w3.org/2000/svg"){var i=u.getAttribute(l)||"";i=e+i;var c=a.get(i);c?c.push(u):a.set(i,[u])}}return a}function _r(e,l,t){e=e.ownerDocument||e,e.head.insertBefore(t,l==="title"?e.querySelector("head > title"):null)}function Ph(e,l,t){if(t===1||l.itemProp!=null)return!1;switch(e){case"meta":case"title":return!0;case"style":if(typeof l.precedence!="string"||typeof l.href!="string"||l.href==="")break;return!0;case"link":if(typeof l.rel!="string"||typeof l.href!="string"||l.href===""||l.onLoad||l.onError)break;switch(l.rel){case"stylesheet":return e=l.disabled,typeof l.precedence=="string"&&e==null;default:return!0}case"script":if(l.async&&typeof l.async!="function"&&typeof l.async!="symbol"&&!l.onLoad&&!l.onError&&l.src&&typeof l.src=="string")return!0}return!1}function Hr(e){return!(e.type==="stylesheet"&&!(e.state.loading&3))}var Sn=null;function Ih(){}function eg(e,l,t){if(Sn===null)throw Error(d(475));var a=Sn;if(l.type==="stylesheet"&&(typeof t.media!="string"||matchMedia(t.media).matches!==!1)&&!(l.state.loading&4)){if(l.instance===null){var n=Aa(t.href),u=e.querySelector(vn(n));if(u){e=u._p,e!==null&&typeof e=="object"&&typeof e.then=="function"&&(a.count++,a=Lu.bind(a),e.then(a,a)),l.state.loading|=4,l.instance=u,Se(u);return}u=e.ownerDocument||e,t=jr(t),(n=el.get(n))&&kc(t,n),u=u.createElement("link"),Se(u);var i=u;i._p=new Promise(function(c,s){i.onload=c,i.onerror=s}),Ee(u,"link",t),l.instance=u}a.stylesheets===null&&(a.stylesheets=new Map),a.stylesheets.set(l,e),(e=l.state.preload)&&!(l.state.loading&3)&&(a.count++,l=Lu.bind(a),e.addEventListener("load",l),e.addEventListener("error",l))}}function lg(){if(Sn===null)throw Error(d(475));var e=Sn;return e.stylesheets&&e.count===0&&Fc(e,e.stylesheets),0<e.count?function(l){var t=setTimeout(function(){if(e.stylesheets&&Fc(e,e.stylesheets),e.unsuspend){var a=e.unsuspend;e.unsuspend=null,a()}},6e4);return e.unsuspend=l,function(){e.unsuspend=null,clearTimeout(t)}}:null}function Lu(){if(this.count--,this.count===0){if(this.stylesheets)Fc(this,this.stylesheets);else if(this.unsuspend){var e=this.unsuspend;this.unsuspend=null,e()}}}var qu=null;function Fc(e,l){e.stylesheets=null,e.unsuspend!==null&&(e.count++,qu=new Map,l.forEach(tg,e),qu=null,Lu.call(e))}function tg(e,l){if(!(l.state.loading&4)){var t=qu.get(e);if(t)var a=t.get(null);else{t=new Map,qu.set(e,t);for(var n=e.querySelectorAll("link[data-precedence],style[data-precedence]"),u=0;u<n.length;u++){var i=n[u];(i.nodeName==="LINK"||i.getAttribute("media")!=="not all")&&(t.set(i.dataset.precedence,i),a=i)}a&&t.set(null,a)}n=l.instance,i=n.getAttribute("data-precedence"),u=t.get(i)||a,u===a&&t.set(null,n),t.set(i,n),this.count++,a=Lu.bind(this),n.addEventListener("load",a),n.addEventListener("error",a),u?u.parentNode.insertBefore(n,u.nextSibling):(e=e.nodeType===9?e.head:e,e.insertBefore(n,e.firstChild)),l.state.loading|=4}}var Tn={$$typeof:pe,Provider:null,Consumer:null,_currentValue:Xe,_currentValue2:Xe,_threadCount:0};function ag(e,l,t,a,n,u,i,c){this.tag=1,this.containerInfo=e,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Wu(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Wu(0),this.hiddenUpdates=Wu(null),this.identifierPrefix=a,this.onUncaughtError=n,this.onCaughtError=u,this.onRecoverableError=i,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=c,this.incompleteTransitions=new Map}function Lr(e,l,t,a,n,u,i,c,s,r,v,T){return e=new ag(e,l,t,i,c,s,r,T),l=1,u===!0&&(l|=24),u=Pe(3,null,null,l),e.current=u,u.stateNode=e,l=Mi(),l.refCount++,e.pooledCache=l,l.refCount++,u.memoizedState={element:a,isDehydrated:t,cache:l},cc(u),e}function qr(e){return e?(e=ta,e):ta}function Br(e,l,t,a,n,u){n=qr(n),a.context===null?a.context=n:a.pendingContext=n,a=$l(l),a.payload={element:t},u=u===void 0?null:u,u!==null&&(a.callback=u),t=Wl(e,a,l),t!==null&&(Oe(t,e,l),an(t,e,l))}function Gr(e,l){if(e=e.memoizedState,e!==null&&e.dehydrated!==null){var t=e.retryLane;e.retryLane=t!==0&&t<l?t:l}}function $c(e,l){Gr(e,l),(e=e.alternate)&&Gr(e,l)}function Yr(e){if(e.tag===13){var l=Ql(e,67108864);l!==null&&Oe(l,e,67108864),$c(e,67108864)}}var Bu=!0;function ng(e,l,t,a){var n=R.T;R.T=null;var u=V.p;try{V.p=2,Wc(e,l,t,a)}finally{V.p=u,R.T=n}}function ug(e,l,t,a){var n=R.T;R.T=null;var u=V.p;try{V.p=8,Wc(e,l,t,a)}finally{V.p=u,R.T=n}}function Wc(e,l,t,a){if(Bu){var n=Pc(a);if(n===null)Gc(e,l,a,Gu,t),Xr(e,a);else if(cg(n,e,l,t,a))a.stopPropagation();else if(Xr(e,a),l&4&&-1<ig.indexOf(e)){for(;n!==null;){var u=Kt(n);if(u!==null)switch(u.tag){case 3:if(u=u.stateNode,u.current.memoizedState.isDehydrated){var i=pt(u.pendingLanes);if(i!==0){var c=u;for(c.pendingLanes|=2,c.entangledLanes|=2;i;){var s=1<<31-He(i);c.entanglements[1]|=s,i&=~s}pl(u),!(ue&6)&&(xu=rl()+500,gn(0))}}break;case 13:c=Ql(u,2),c!==null&&Oe(c,u,2),Nu(),$c(u,2)}if(u=Pc(a),u===null&&Gc(e,l,a,Gu,t),u===n)break;n=u}n!==null&&a.stopPropagation()}else Gc(e,l,a,null,t)}}function Pc(e){return e=ni(e),Ic(e)}var Gu=null;function Ic(e){if(Gu=null,e=yt(e),e!==null){var l=U(e);if(l===null)e=null;else{var t=l.tag;if(t===13){if(e=te(l),e!==null)return e;e=null}else if(t===3){if(l.stateNode.current.memoizedState.isDehydrated)return l.tag===3?l.stateNode.containerInfo:null;e=null}else l!==e&&(e=null)}}return Gu=e,null}function wr(e){switch(e){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(Kd()){case os:return 2;case fs:return 8;case _n:case Zd:return 32;case rs:return 268435456;default:return 32}default:return 32}}var es=!1,nt=null,ut=null,it=null,xn=new Map,An=new Map,ct=[],ig="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function Xr(e,l){switch(e){case"focusin":case"focusout":nt=null;break;case"dragenter":case"dragleave":ut=null;break;case"mouseover":case"mouseout":it=null;break;case"pointerover":case"pointerout":xn.delete(l.pointerId);break;case"gotpointercapture":case"lostpointercapture":An.delete(l.pointerId)}}function Dn(e,l,t,a,n,u){return e===null||e.nativeEvent!==u?(e={blockedOn:l,domEventName:t,eventSystemFlags:a,nativeEvent:u,targetContainers:[n]},l!==null&&(l=Kt(l),l!==null&&Yr(l)),e):(e.eventSystemFlags|=a,l=e.targetContainers,n!==null&&l.indexOf(n)===-1&&l.push(n),e)}function cg(e,l,t,a,n){switch(l){case"focusin":return nt=Dn(nt,e,l,t,a,n),!0;case"dragenter":return ut=Dn(ut,e,l,t,a,n),!0;case"mouseover":return it=Dn(it,e,l,t,a,n),!0;case"pointerover":var u=n.pointerId;return xn.set(u,Dn(xn.get(u)||null,e,l,t,a,n)),!0;case"gotpointercapture":return u=n.pointerId,An.set(u,Dn(An.get(u)||null,e,l,t,a,n)),!0}return!1}function Qr(e){var l=yt(e.target);if(l!==null){var t=U(l);if(t!==null){if(l=t.tag,l===13){if(l=te(t),l!==null){e.blockedOn=l,lm(e.priority,function(){if(t.tag===13){var a=Ye(),n=Ql(t,a);n!==null&&Oe(n,t,a),$c(t,a)}});return}}else if(l===3&&t.stateNode.current.memoizedState.isDehydrated){e.blockedOn=t.tag===3?t.stateNode.containerInfo:null;return}}}e.blockedOn=null}function Yu(e){if(e.blockedOn!==null)return!1;for(var l=e.targetContainers;0<l.length;){var t=Pc(e.nativeEvent);if(t===null){t=e.nativeEvent;var a=new t.constructor(t.type,t);ai=a,t.target.dispatchEvent(a),ai=null}else return l=Kt(t),l!==null&&Yr(l),e.blockedOn=t,!1;l.shift()}return!0}function Vr(e,l,t){Yu(e)&&t.delete(l)}function sg(){es=!1,nt!==null&&Yu(nt)&&(nt=null),ut!==null&&Yu(ut)&&(ut=null),it!==null&&Yu(it)&&(it=null),xn.forEach(Vr),An.forEach(Vr)}function wu(e,l){e.blockedOn===l&&(e.blockedOn=null,es||(es=!0,g.unstable_scheduleCallback(g.unstable_NormalPriority,sg)))}var Xu=null;function Kr(e){Xu!==e&&(Xu=e,g.unstable_scheduleCallback(g.unstable_NormalPriority,function(){Xu===e&&(Xu=null);for(var l=0;l<e.length;l+=3){var t=e[l],a=e[l+1],n=e[l+2];if(typeof a!="function"){if(Ic(a||t)===null)continue;break}var u=Kt(t);u!==null&&(e.splice(l,3),l-=3,Qi(u,{pending:!0,data:n,method:t.method,action:a},a,n))}}))}function Nn(e){function l(s){return wu(s,e)}nt!==null&&wu(nt,e),ut!==null&&wu(ut,e),it!==null&&wu(it,e),xn.forEach(l),An.forEach(l);for(var t=0;t<ct.length;t++){var a=ct[t];a.blockedOn===e&&(a.blockedOn=null)}for(;0<ct.length&&(t=ct[0],t.blockedOn===null);)Qr(t),t.blockedOn===null&&ct.shift();if(t=(e.ownerDocument||e).$$reactFormReplay,t!=null)for(a=0;a<t.length;a+=3){var n=t[a],u=t[a+1],i=n[je]||null;if(typeof u=="function")i||Kr(t);else if(i){var c=null;if(u&&u.hasAttribute("formAction")){if(n=u,i=u[je]||null)c=i.formAction;else if(Ic(n)!==null)continue}else c=i.action;typeof c=="function"?t[a+1]=c:(t.splice(a,3),a-=3),Kr(t)}}}function ls(e){this._internalRoot=e}Qu.prototype.render=ls.prototype.render=function(e){var l=this._internalRoot;if(l===null)throw Error(d(409));var t=l.current,a=Ye();Br(t,a,e,l,null,null)},Qu.prototype.unmount=ls.prototype.unmount=function(){var e=this._internalRoot;if(e!==null){this._internalRoot=null;var l=e.containerInfo;e.tag===0&&ba(),Br(e.current,2,null,e,null,null),Nu(),l[Vt]=null}};function Qu(e){this._internalRoot=e}Qu.prototype.unstable_scheduleHydration=function(e){if(e){var l=ys();e={blockedOn:null,target:e,priority:l};for(var t=0;t<ct.length&&l!==0&&l<ct[t].priority;t++);ct.splice(t,0,e),t===0&&Qr(e)}};var Zr=b.version;if(Zr!=="19.0.0")throw Error(d(527,Zr,"19.0.0"));V.findDOMNode=function(e){var l=e._reactInternals;if(l===void 0)throw typeof e.render=="function"?Error(d(188)):(e=Object.keys(e).join(","),Error(d(268,e)));return e=wt(l),e=e!==null?gt(e):null,e=e===null?null:e.stateNode,e};var og={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:R,findFiberByHostInstance:yt,reconcilerVersion:"19.0.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var Vu=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!Vu.isDisabled&&Vu.supportsFiber)try{Ca=Vu.inject(og),_e=Vu}catch{}}return Mn.createRoot=function(e,l){if(!N(e))throw Error(d(299));var t=!1,a="",n=of,u=ff,i=rf,c=null;return l!=null&&(l.unstable_strictMode===!0&&(t=!0),l.identifierPrefix!==void 0&&(a=l.identifierPrefix),l.onUncaughtError!==void 0&&(n=l.onUncaughtError),l.onCaughtError!==void 0&&(u=l.onCaughtError),l.onRecoverableError!==void 0&&(i=l.onRecoverableError),l.unstable_transitionCallbacks!==void 0&&(c=l.unstable_transitionCallbacks)),l=Lr(e,1,!1,null,null,t,a,n,u,i,c,null),e[Vt]=l.current,Bc(e.nodeType===8?e.parentNode:e),new ls(l)},Mn.hydrateRoot=function(e,l,t){if(!N(e))throw Error(d(299));var a=!1,n="",u=of,i=ff,c=rf,s=null,r=null;return t!=null&&(t.unstable_strictMode===!0&&(a=!0),t.identifierPrefix!==void 0&&(n=t.identifierPrefix),t.onUncaughtError!==void 0&&(u=t.onUncaughtError),t.onCaughtError!==void 0&&(i=t.onCaughtError),t.onRecoverableError!==void 0&&(c=t.onRecoverableError),t.unstable_transitionCallbacks!==void 0&&(s=t.unstable_transitionCallbacks),t.formState!==void 0&&(r=t.formState)),l=Lr(e,1,!0,l,t??null,a,n,u,i,c,s,r),l.context=qr(null),t=l.current,a=Ye(),n=$l(a),n.callback=null,Wl(t,n,a),l.current.lanes=a,ja(l,a),pl(l),e[Vt]=l.current,Bc(e),new Qu(l)},Mn.version="19.0.0",Mn}var ld;function tp(){if(ld)return ts.exports;ld=1;function g(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(g)}catch(b){console.error(b)}}return g(),ts.exports=lp(),ts.exports}var ap=tp();const np={visibleTabs:{},setTabVisibility:()=>{},isTabVisible:()=>!1},pd=E.createContext(np),up=({children:g})=>{const b=ll.use.currentTab(),[x,d]=E.useState(()=>({documents:!0,"knowledge-graph":!0,retrieval:!0,api:!0}));E.useEffect(()=>{d(j=>({...j,documents:!0,"knowledge-graph":!0,retrieval:!0,api:!0}))},[b]);const N=E.useMemo(()=>({visibleTabs:x,setTabVisibility:(j,H)=>{d($=>({...$,[j]:H}))},isTabVisible:j=>!!x[j]}),[x]);return o.jsx(pd.Provider,{value:N,children:g})};var yd="AlertDialog",[ip,Yy]=hg(yd,[td]),Bl=td(),vd=g=>{const{__scopeAlertDialog:b,...x}=g,d=Bl(b);return o.jsx(Sg,{...d,...x,modal:!0})};vd.displayName=yd;var cp="AlertDialogTrigger",sp=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,N=Bl(x);return o.jsx(Tg,{...N,...d,ref:b})});sp.displayName=cp;var op="AlertDialogPortal",bd=g=>{const{__scopeAlertDialog:b,...x}=g,d=Bl(b);return o.jsx(dg,{...d,...x})};bd.displayName=op;var fp="AlertDialogOverlay",Sd=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,N=Bl(x);return o.jsx(rg,{...N,...d,ref:b})});Sd.displayName=fp;var Na="AlertDialogContent",[rp,dp]=ip(Na),Td=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,children:d,...N}=g,j=Bl(x),H=E.useRef(null),$=ad(b,H),Y=E.useRef(null);return o.jsx(mg,{contentName:Na,titleName:xd,docsSlug:"alert-dialog",children:o.jsx(rp,{scope:x,cancelRef:Y,children:o.jsxs(gg,{role:"alertdialog",...j,...N,ref:$,onOpenAutoFocus:pg(N.onOpenAutoFocus,W=>{var he;W.preventDefault(),(he=Y.current)==null||he.focus({preventScroll:!0})}),onPointerDownOutside:W=>W.preventDefault(),onInteractOutside:W=>W.preventDefault(),children:[o.jsx(yg,{children:d}),o.jsx(hp,{contentRef:H})]})})})});Td.displayName=Na;var xd="AlertDialogTitle",Ad=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,N=Bl(x);return o.jsx(vg,{...N,...d,ref:b})});Ad.displayName=xd;var Dd="AlertDialogDescription",Nd=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,N=Bl(x);return o.jsx(bg,{...N,...d,ref:b})});Nd.displayName=Dd;var mp="AlertDialogAction",Ed=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,N=Bl(x);return o.jsx(nd,{...N,...d,ref:b})});Ed.displayName=mp;var Md="AlertDialogCancel",zd=E.forwardRef((g,b)=>{const{__scopeAlertDialog:x,...d}=g,{cancelRef:N}=dp(Md,x),j=Bl(x),H=ad(b,N);return o.jsx(nd,{...j,...d,ref:H})});zd.displayName=Md;var hp=({contentRef:g})=>{const b=`\`${Na}\` requires a description for the component to be accessible for screen reader users.

You can add a description to the \`${Na}\` by passing a \`${Dd}\` component as a child, which also benefits sighted users by adding visible context to the dialog.

Alternatively, you can use your own component as a description by assigning it an \`id\` and passing the same value to the \`aria-describedby\` prop in \`${Na}\`. If the description is confusing or duplicative for sighted users, you can use the \`@radix-ui/react-visually-hidden\` primitive as a wrapper around your description component.

For more information, see https://radix-ui.com/primitives/docs/components/alert-dialog`;return E.useEffect(()=>{var d;document.getElementById((d=g.current)==null?void 0:d.getAttribute("aria-describedby"))||console.warn(b)},[b,g]),null},gp=vd,pp=bd,Cd=Sd,Od=Td,jd=Ed,Rd=zd,Ud=Ad,_d=Nd;const yp=gp,vp=pp,Hd=E.forwardRef(({className:g,...b},x)=>o.jsx(Cd,{className:we("data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 fixed inset-0 z-50 bg-black/50",g),...b,ref:x}));Hd.displayName=Cd.displayName;const Ld=E.forwardRef(({className:g,...b},x)=>o.jsxs(vp,{children:[o.jsx(Hd,{}),o.jsx(Od,{ref:x,className:we("bg-background data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[state=closed]:slide-out-to-top-[48%] data-[state=open]:slide-in-from-top-[48%] fixed top-[50%] left-[50%] z-50 grid w-full max-w-lg translate-x-[-50%] translate-y-[-50%] gap-4 border p-6 shadow-lg duration-200 sm:rounded-lg",g),...b})]}));Ld.displayName=Od.displayName;const qd=({className:g,...b})=>o.jsx("div",{className:we("flex flex-col space-y-2 text-center sm:text-left",g),...b});qd.displayName="AlertDialogHeader";const Bd=E.forwardRef(({className:g,...b},x)=>o.jsx(Ud,{ref:x,className:we("text-lg font-semibold",g),...b}));Bd.displayName=Ud.displayName;const Gd=E.forwardRef(({className:g,...b},x)=>o.jsx(_d,{ref:x,className:we("text-muted-foreground text-sm",g),...b}));Gd.displayName=_d.displayName;const bp=E.forwardRef(({className:g,...b},x)=>o.jsx(jd,{ref:x,className:we(od(),g),...b}));bp.displayName=jd.displayName;const Sp=E.forwardRef(({className:g,...b},x)=>o.jsx(Rd,{ref:x,className:we(od({variant:"outline"}),"mt-2 sm:mt-0",g),...b}));Sp.displayName=Rd.displayName;const Tp=({open:g,onOpenChange:b})=>{const{t:x}=ql(),d=ll.use.apiKey(),[N,j]=E.useState(""),H=Bt.use.message();E.useEffect(()=>{j(d||"")},[d,g]),E.useEffect(()=>{H&&(H.includes(fd)||H.includes(rd))&&b(!0)},[H,b]);const $=E.useCallback(()=>{ll.setState({apiKey:N||null}),b(!1)},[N,b]),Y=E.useCallback(W=>{j(W.target.value)},[j]);return o.jsx(yp,{open:g,onOpenChange:b,children:o.jsxs(Ld,{children:[o.jsxs(qd,{children:[o.jsx(Bd,{children:x("apiKeyAlert.title")}),o.jsx(Gd,{children:x("apiKeyAlert.description")})]}),o.jsxs("div",{className:"flex flex-col gap-4",children:[o.jsxs("form",{className:"flex gap-2",onSubmit:W=>W.preventDefault(),children:[o.jsx(us,{type:"password",value:N,onChange:Y,placeholder:x("apiKeyAlert.placeholder"),className:"max-h-full w-full min-w-0",autoComplete:"off"}),o.jsx(Cn,{onClick:$,variant:"outline",size:"sm",children:x("apiKeyAlert.save")})]}),H&&o.jsx("div",{className:"text-sm text-red-500",children:H})]})]})})},xp=({status:g})=>{const{t:b}=ql();return g?o.jsxs("div",{className:"min-w-[300px] space-y-2 text-xs",children:[o.jsxs("div",{className:"space-y-1",children:[o.jsx("h4",{className:"font-medium",children:b("graphPanel.statusCard.storageInfo")}),o.jsxs("div",{className:"text-foreground grid grid-cols-[120px_1fr] gap-1",children:[o.jsxs("span",{children:[b("graphPanel.statusCard.workingDirectory"),":"]}),o.jsx("span",{className:"truncate",children:g.working_directory}),o.jsxs("span",{children:[b("graphPanel.statusCard.inputDirectory"),":"]}),o.jsx("span",{className:"truncate",children:g.input_directory})]})]}),o.jsxs("div",{className:"space-y-1",children:[o.jsx("h4",{className:"font-medium",children:b("graphPanel.statusCard.llmConfig")}),o.jsxs("div",{className:"text-foreground grid grid-cols-[120px_1fr] gap-1",children:[o.jsxs("span",{children:[b("graphPanel.statusCard.llmBinding"),":"]}),o.jsx("span",{children:g.configuration.llm_binding}),o.jsxs("span",{children:[b("graphPanel.statusCard.llmBindingHost"),":"]}),o.jsx("span",{children:g.configuration.llm_binding_host}),o.jsxs("span",{children:[b("graphPanel.statusCard.llmModel"),":"]}),o.jsx("span",{children:g.configuration.llm_model}),o.jsxs("span",{children:[b("graphPanel.statusCard.maxTokens"),":"]}),o.jsx("span",{children:g.configuration.max_tokens})]})]}),o.jsxs("div",{className:"space-y-1",children:[o.jsx("h4",{className:"font-medium",children:b("graphPanel.statusCard.embeddingConfig")}),o.jsxs("div",{className:"text-foreground grid grid-cols-[120px_1fr] gap-1",children:[o.jsxs("span",{children:[b("graphPanel.statusCard.embeddingBinding"),":"]}),o.jsx("span",{children:g.configuration.embedding_binding}),o.jsxs("span",{children:[b("graphPanel.statusCard.embeddingBindingHost"),":"]}),o.jsx("span",{children:g.configuration.embedding_binding_host}),o.jsxs("span",{children:[b("graphPanel.statusCard.embeddingModel"),":"]}),o.jsx("span",{children:g.configuration.embedding_model})]})]}),o.jsxs("div",{className:"space-y-1",children:[o.jsx("h4",{className:"font-medium",children:b("graphPanel.statusCard.storageConfig")}),o.jsxs("div",{className:"text-foreground grid grid-cols-[120px_1fr] gap-1",children:[o.jsxs("span",{children:[b("graphPanel.statusCard.kvStorage"),":"]}),o.jsx("span",{children:g.configuration.kv_storage}),o.jsxs("span",{children:[b("graphPanel.statusCard.docStatusStorage"),":"]}),o.jsx("span",{children:g.configuration.doc_status_storage}),o.jsxs("span",{children:[b("graphPanel.statusCard.graphStorage"),":"]}),o.jsx("span",{children:g.configuration.graph_storage}),o.jsxs("span",{children:[b("graphPanel.statusCard.vectorStorage"),":"]}),o.jsx("span",{children:g.configuration.vector_storage})]})]})]}):o.jsx("div",{className:"text-foreground text-xs",children:b("graphPanel.statusCard.unavailable")})},Ap=({open:g,onOpenChange:b,status:x})=>{const{t:d}=ql();return o.jsx(Mg,{open:g,onOpenChange:b,children:o.jsxs(zg,{className:"sm:max-w-[500px]",children:[o.jsxs(Cg,{children:[o.jsx(Og,{children:d("graphPanel.statusDialog.title")}),o.jsx(jg,{children:d("graphPanel.statusDialog.description")})]}),o.jsx(xp,{status:x})]})})},Dp=()=>{const{t:g}=ql(),b=Bt.use.health(),x=Bt.use.lastCheckTime(),d=Bt.use.status(),[N,j]=E.useState(!1),[H,$]=E.useState(!1);return E.useEffect(()=>{j(!0);const Y=setTimeout(()=>j(!1),300);return()=>clearTimeout(Y)},[x]),o.jsxs("div",{className:"fixed right-4 bottom-4 flex items-center gap-2 opacity-80 select-none",children:[o.jsxs("div",{className:"flex cursor-pointer items-center gap-2",onClick:()=>$(!0),children:[o.jsx("div",{className:we("h-3 w-3 rounded-full transition-all duration-300","shadow-[0_0_8px_rgba(0,0,0,0.2)]",b?"bg-green-500":"bg-red-500",N&&"scale-125",N&&b&&"shadow-[0_0_12px_rgba(34,197,94,0.4)]",N&&!b&&"shadow-[0_0_12px_rgba(239,68,68,0.4)]")}),o.jsx("span",{className:"text-muted-foreground text-xs",children:g(b?"graphPanel.statusIndicator.connected":"graphPanel.statusIndicator.disconnected")})]}),o.jsx(Ap,{open:H,onOpenChange:$,status:d})]})};function Yd({className:g}){const[b,x]=E.useState(!1),{t:d}=ql(),N=ll.use.language(),j=ll.use.setLanguage(),H=ll.use.theme(),$=ll.use.setTheme(),Y=E.useCallback(he=>{j(he)},[j]),W=E.useCallback(he=>{$(he)},[$]);return o.jsxs(Rg,{open:b,onOpenChange:x,children:[o.jsx(Ug,{asChild:!0,children:o.jsx(Cn,{variant:"ghost",size:"icon",className:we("h-9 w-9",g),children:o.jsx(_g,{className:"h-5 w-5"})})}),o.jsx(Hg,{side:"bottom",align:"end",className:"w-56",children:o.jsxs("div",{className:"flex flex-col gap-4",children:[o.jsxs("div",{className:"flex flex-col gap-2",children:[o.jsx("label",{className:"text-sm font-medium",children:d("settings.language")}),o.jsxs(Jr,{value:N,onValueChange:Y,children:[o.jsx(Fr,{children:o.jsx($r,{})}),o.jsxs(Wr,{children:[o.jsx(ot,{value:"en",children:"English"}),o.jsx(ot,{value:"zh",children:"中文"}),o.jsx(ot,{value:"fr",children:"Français"}),o.jsx(ot,{value:"ar",children:"العربية"}),o.jsx(ot,{value:"zh_TW",children:"繁體中文"})]})]})]}),o.jsxs("div",{className:"flex flex-col gap-2",children:[o.jsx("label",{className:"text-sm font-medium",children:d("settings.theme")}),o.jsxs(Jr,{value:H,onValueChange:W,children:[o.jsx(Fr,{children:o.jsx($r,{})}),o.jsxs(Wr,{children:[o.jsx(ot,{value:"light",children:d("settings.light")}),o.jsx(ot,{value:"dark",children:d("settings.dark")}),o.jsx(ot,{value:"system",children:d("settings.system")})]})]})]})]})})]})}const Np=xg,wd=E.forwardRef(({className:g,...b},x)=>o.jsx(ud,{ref:x,className:we("bg-muted text-muted-foreground inline-flex h-10 items-center justify-center rounded-md p-1",g),...b}));wd.displayName=ud.displayName;const Xd=E.forwardRef(({className:g,...b},x)=>o.jsx(id,{ref:x,className:we("ring-offset-background focus-visible:ring-ring data-[state=active]:bg-background data-[state=active]:text-foreground inline-flex items-center justify-center rounded-sm px-3 py-1.5 text-sm font-medium whitespace-nowrap transition-all focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none disabled:pointer-events-none disabled:opacity-50 data-[state=active]:shadow-sm",g),...b}));Xd.displayName=id.displayName;const zn=E.forwardRef(({className:g,...b},x)=>o.jsx(cd,{ref:x,className:we("ring-offset-background focus-visible:ring-ring focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none","data-[state=inactive]:invisible data-[state=active]:visible","h-full w-full",g),forceMount:!0,...b}));zn.displayName=cd.displayName;function Ku({value:g,currentTab:b,children:x}){return o.jsx(Xd,{value:g,className:we("cursor-pointer px-2 py-1 transition-all",b===g?"!bg-emerald-400 !text-zinc-50":"hover:bg-background/60"),children:x})}function Ep(){const g=ll.use.currentTab(),{t:b}=ql();return o.jsx("div",{className:"flex h-8 self-center",children:o.jsxs(wd,{className:"h-full gap-2",children:[o.jsx(Ku,{value:"documents",currentTab:g,children:b("header.documents")}),o.jsx(Ku,{value:"knowledge-graph",currentTab:g,children:b("header.knowledgeGraph")}),o.jsx(Ku,{value:"retrieval",currentTab:g,children:b("header.retrieval")}),o.jsx(Ku,{value:"api",currentTab:g,children:b("header.api")})]})})}function Mp(){const{t:g}=ql(),{isGuestMode:b,coreVersion:x,apiVersion:d,username:N,webuiTitle:j,webuiDescription:H}=Ll(),$=x&&d?`${x}/${d}`:null,Y=()=>{md.navigateToLogin()};return o.jsxs("header",{className:"border-border/40 bg-background/95 supports-[backdrop-filter]:bg-background/60 sticky top-0 z-50 flex h-10 w-full border-b px-4 backdrop-blur",children:[o.jsxs("div",{className:"min-w-[200px] w-auto flex items-center",children:[o.jsxs("a",{href:dd,className:"flex items-center gap-2",children:[o.jsx(ss,{className:"size-4 text-emerald-400","aria-hidden":"true"}),o.jsx("span",{className:"font-bold md:inline-block",children:is.name})]}),j&&o.jsxs("div",{className:"flex items-center",children:[o.jsx("span",{className:"mx-1 text-xs text-gray-500 dark:text-gray-400",children:"|"}),o.jsx(Lg,{children:o.jsxs(qg,{children:[o.jsx(Bg,{asChild:!0,children:o.jsx("span",{className:"font-medium text-sm cursor-default",children:j})}),H&&o.jsx(Gg,{side:"bottom",children:H})]})})]})]}),o.jsxs("div",{className:"flex h-10 flex-1 items-center justify-center",children:[o.jsx(Ep,{}),b&&o.jsx("div",{className:"ml-2 self-center px-2 py-1 text-xs bg-amber-100 text-amber-800 dark:bg-amber-900 dark:text-amber-200 rounded-md",children:g("login.guestMode","Guest Mode")})]}),o.jsx("nav",{className:"w-[200px] flex items-center justify-end",children:o.jsxs("div",{className:"flex items-center gap-2",children:[$&&o.jsxs("span",{className:"text-xs text-gray-500 dark:text-gray-400 mr-1",children:["v",$]}),o.jsx(Cn,{variant:"ghost",size:"icon",side:"bottom",tooltip:g("header.projectRepository"),children:o.jsx("a",{href:is.github,target:"_blank",rel:"noopener noreferrer",children:o.jsx(Yg,{className:"size-4","aria-hidden":"true"})})}),o.jsx(Yd,{}),!b&&o.jsx(Cn,{variant:"ghost",size:"icon",side:"bottom",tooltip:`${g("header.logout")} (${N})`,onClick:Y,children:o.jsx(wg,{className:"size-4","aria-hidden":"true"})})]})})]})}const zp=()=>{const g=E.useContext(pd);if(!g)throw new Error("useTabVisibility must be used within a TabVisibilityProvider");return g};function Cp(){const{t:g}=ql(),{isTabVisible:b}=zp(),x=b("api"),[d,N]=E.useState(!1);return E.useEffect(()=>{d||N(!0)},[d]),o.jsx("div",{className:`size-full ${x?"":"hidden"}`,children:d?o.jsx("iframe",{src:Xg+"/docs",className:"size-full w-full h-full",style:{width:"100%",height:"100%",border:"none"}},"api-docs-iframe"):o.jsx("div",{className:"flex h-full w-full items-center justify-center bg-background",children:o.jsxs("div",{className:"text-center",children:[o.jsx("div",{className:"mb-2 h-8 w-8 animate-spin rounded-full border-4 border-primary border-t-transparent"}),o.jsx("p",{children:g("apiSite.loading")})]})})})}function Op(){const g=Bt.use.message(),b=ll.use.enableHealthCheck(),x=ll.use.currentTab(),[d,N]=E.useState(!1),[j,H]=E.useState(!0),$=E.useRef(!1),Y=E.useRef(!1),W=E.useCallback(X=>{N(X),X||Bt.getState().clear()},[]),he=E.useRef(!0);E.useEffect(()=>{he.current=!0;const X=()=>{he.current=!1};return window.addEventListener("beforeunload",X),()=>{he.current=!1,window.removeEventListener("beforeunload",X)}},[]),E.useEffect(()=>{if(!b||d)return;const X=async()=>{try{he.current&&await Bt.getState().check()}catch(le){console.error("Health check error:",le)}};Y.current||(Y.current=!0,X());const pe=setInterval(X,Qg*1e3);return()=>clearInterval(pe)},[b,d]),E.useEffect(()=>{(async()=>{if($.current)return;if($.current=!0,sessionStorage.getItem("VERSION_CHECKED_FROM_LOGIN")==="true"){H(!1);return}try{H(!0);const le=localStorage.getItem("LIGHTRAG-API-TOKEN"),C=await gd();if(!C.auth_configured&&C.access_token)Ll.getState().login(C.access_token,!0,C.core_version,C.api_version,C.webui_title||null,C.webui_description||null);else if(le&&(C.core_version||C.api_version||C.webui_title||C.webui_description)){const tl=C.auth_mode==="disabled"||Ll.getState().isGuestMode;Ll.getState().login(le,tl,C.core_version,C.api_version,C.webui_title||null,C.webui_description||null)}sessionStorage.setItem("VERSION_CHECKED_FROM_LOGIN","true")}catch(le){console.error("Failed to get version info:",le)}finally{H(!1)}})()},[]);const ge=E.useCallback(X=>ll.getState().setCurrentTab(X),[]);return E.useEffect(()=>{g&&(g.includes(fd)||g.includes(rd))&&N(!0)},[g]),o.jsx(hd,{children:o.jsx(up,{children:j?o.jsxs("div",{className:"flex h-screen w-screen flex-col",children:[o.jsxs("header",{className:"border-border/40 bg-background/95 supports-[backdrop-filter]:bg-background/60 sticky top-0 z-50 flex h-10 w-full border-b px-4 backdrop-blur",children:[o.jsx("div",{className:"min-w-[200px] w-auto flex items-center",children:o.jsxs("a",{href:dd,className:"flex items-center gap-2",children:[o.jsx(ss,{className:"size-4 text-emerald-400","aria-hidden":"true"}),o.jsx("span",{className:"font-bold md:inline-block",children:is.name})]})}),o.jsx("div",{className:"flex h-10 flex-1 items-center justify-center"}),o.jsx("nav",{className:"w-[200px] flex items-center justify-end"})]}),o.jsx("div",{className:"flex flex-1 items-center justify-center",children:o.jsxs("div",{className:"text-center",children:[o.jsx("div",{className:"mb-2 h-8 w-8 animate-spin rounded-full border-4 border-primary border-t-transparent"}),o.jsx("p",{children:"Initializing..."})]})})]}):o.jsxs("main",{className:"flex h-screen w-screen overflow-hidden",children:[o.jsxs(Np,{defaultValue:x,className:"!m-0 flex grow flex-col !p-0 overflow-hidden",onValueChange:ge,children:[o.jsx(Mp,{}),o.jsxs("div",{className:"relative grow",children:[o.jsx(zn,{value:"documents",className:"absolute top-0 right-0 bottom-0 left-0 overflow-auto",children:o.jsx(Pg,{})}),o.jsx(zn,{value:"knowledge-graph",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:o.jsx(Vg,{})}),o.jsx(zn,{value:"retrieval",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:o.jsx(Wg,{})}),o.jsx(zn,{value:"api",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:o.jsx(Cp,{})})]})]}),b&&o.jsx(Dp,{}),o.jsx(Tp,{open:d,onOpenChange:W})]})})})}const jp=()=>{const g=sd(),{login:b,isAuthenticated:x}=Ll(),{t:d}=ql(),[N,j]=E.useState(!1),[H,$]=E.useState(""),[Y,W]=E.useState(""),[he,ge]=E.useState(!0),X=E.useRef(!1);if(E.useEffect(()=>{console.log("LoginPage mounted")},[]),E.useEffect(()=>((async()=>{if(!X.current){X.current=!0;try{if(x){g("/");return}const C=await gd();if((C.core_version||C.api_version)&&sessionStorage.setItem("VERSION_CHECKED_FROM_LOGIN","true"),!C.auth_configured&&C.access_token){b(C.access_token,!0,C.core_version,C.api_version,C.webui_title||null,C.webui_description||null),C.message&&En.info(C.message),g("/");return}ge(!1)}catch(C){console.error("Failed to check auth configuration:",C),ge(!1)}}})(),()=>{}),[x,b,g]),he)return null;const pe=async le=>{if(le.preventDefault(),!H||!Y){En.error(d("login.errorEmptyFields"));return}try{j(!0);const C=await Jg(H,Y),tl=C.auth_mode==="disabled";b(C.access_token,tl,C.core_version,C.api_version,C.webui_title||null,C.webui_description||null),(C.core_version||C.api_version)&&sessionStorage.setItem("VERSION_CHECKED_FROM_LOGIN","true"),tl?En.info(C.message||d("login.authDisabled","Authentication is disabled. Using guest access.")):En.success(d("login.successMessage")),g("/")}catch(C){console.error("Login failed...",C),En.error(d("login.errorInvalidCredentials")),Ll.getState().logout(),localStorage.removeItem("LIGHTRAG-API-TOKEN")}finally{j(!1)}};return o.jsxs("div",{className:"flex h-screen w-screen items-center justify-center bg-gradient-to-br from-emerald-50 to-teal-100 dark:from-gray-900 dark:to-gray-800",children:[o.jsx("div",{className:"absolute top-4 right-4 flex items-center gap-2",children:o.jsx(Yd,{className:"bg-white/30 dark:bg-gray-800/30 backdrop-blur-sm rounded-md"})}),o.jsxs(Kg,{className:"w-full max-w-[480px] shadow-lg mx-4",children:[o.jsx(Zg,{className:"flex items-center justify-center space-y-2 pb-8 pt-6",children:o.jsxs("div",{className:"flex flex-col items-center space-y-4",children:[o.jsxs("div",{className:"flex items-center gap-3",children:[o.jsx("img",{src:"logo.png",alt:"LightRAG Logo",className:"h-12 w-12"}),o.jsx(ss,{className:"size-10 text-emerald-400","aria-hidden":"true"})]}),o.jsxs("div",{className:"text-center space-y-2",children:[o.jsx("h1",{className:"text-3xl font-bold tracking-tight",children:"LightRAG"}),o.jsx("p",{className:"text-muted-foreground text-sm",children:d("login.description")})]})]})}),o.jsx(kg,{className:"px-8 pb-8",children:o.jsxs("form",{onSubmit:pe,className:"space-y-6",children:[o.jsxs("div",{className:"flex items-center gap-4",children:[o.jsx("label",{htmlFor:"username-input",className:"text-sm font-medium w-16 shrink-0",children:d("login.username")}),o.jsx(us,{id:"username-input",placeholder:d("login.usernamePlaceholder"),value:H,onChange:le=>$(le.target.value),required:!0,className:"h-11 flex-1"})]}),o.jsxs("div",{className:"flex items-center gap-4",children:[o.jsx("label",{htmlFor:"password-input",className:"text-sm font-medium w-16 shrink-0",children:d("login.password")}),o.jsx(us,{id:"password-input",type:"password",placeholder:d("login.passwordPlaceholder"),value:Y,onChange:le=>W(le.target.value),required:!0,className:"h-11 flex-1"})]}),o.jsx(Cn,{type:"submit",className:"w-full h-11 text-base font-medium mt-2",disabled:N,children:d(N?"login.loggingIn":"login.loginButton")})]})})]})]})},Rp=()=>{const[g,b]=E.useState(!0),{isAuthenticated:x}=Ll(),d=sd();return E.useEffect(()=>{md.setNavigate(d)},[d]),E.useEffect(()=>((async()=>{try{const j=localStorage.getItem("LIGHTRAG-API-TOKEN");if(j&&x){b(!1);return}j||Ll.getState().logout()}catch(j){console.error("Auth initialization error:",j),x||Ll.getState().logout()}finally{b(!1)}})(),()=>{}),[x]),E.useEffect(()=>{!g&&!x&&window.location.hash.slice(1)!=="/login"&&(console.log("Not authenticated, redirecting to login"),d("/login"))},[g,x,d]),g?null:o.jsxs(Eg,{children:[o.jsx(kr,{path:"/login",element:o.jsx(jp,{})}),o.jsx(kr,{path:"/*",element:x?o.jsx(Op,{}):null})]})},Up=()=>o.jsx(hd,{children:o.jsxs(Ng,{children:[o.jsx(Rp,{}),o.jsx(Fg,{position:"bottom-center",theme:"system",closeButton:!0,richColors:!0})]})}),_p={language:"Language",theme:"Theme",light:"Light",dark:"Dark",system:"System"},Hp={documents:"Documents",knowledgeGraph:"Knowledge Graph",retrieval:"Retrieval",api:"API",projectRepository:"Project Repository",logout:"Logout",themeToggle:{switchToLight:"Switch to light theme",switchToDark:"Switch to dark theme"}},Lp={description:"Please enter your account and password to log in to the system",username:"Username",usernamePlaceholder:"Please input a username",password:"Password",passwordPlaceholder:"Please input a password",loginButton:"Login",loggingIn:"Logging in...",successMessage:"Login succeeded",errorEmptyFields:"Please enter your username and password",errorInvalidCredentials:"Login failed, please check username and password",authDisabled:"Authentication is disabled. Using login free mode.",guestMode:"Login Free"},qp={cancel:"Cancel",save:"Save"},Bp={clearDocuments:{button:"Clear",tooltip:"Clear documents",title:"Clear Documents",description:"This will remove all documents from the system",warning:"WARNING: This action will permanently delete all documents and cannot be undone!",confirm:"Do you really want to clear all documents?",confirmPrompt:"Type 'yes' to confirm this action",confirmPlaceholder:"Type yes to confirm",clearCache:"Clear LLM cache",confirmButton:"YES",success:"Documents cleared successfully",cacheCleared:"Cache cleared successfully",cacheClearFailed:`Failed to clear cache:
{{error}}`,failed:`Clear Documents Failed:
{{message}}`,error:`Clear Documents Failed:
{{error}}`},uploadDocuments:{button:"Upload",tooltip:"Upload documents",title:"Upload Documents",description:"Drag and drop your documents here or click to browse.",single:{uploading:"Uploading {{name}}: {{percent}}%",success:`Upload Success:
{{name}} uploaded successfully`,failed:`Upload Failed:
{{name}}
{{message}}`,error:`Upload Failed:
{{name}}
{{error}}`},batch:{uploading:"Uploading files...",success:"Files uploaded successfully",error:"Some files failed to upload"},generalError:`Upload Failed
{{error}}`,fileTypes:"Supported types: TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS",fileUploader:{singleFileLimit:"Cannot upload more than 1 file at a time",maxFilesLimit:"Cannot upload more than {{count}} files",fileRejected:"File {{name}} was rejected",unsupportedType:"Unsupported file type",fileTooLarge:"File too large, maximum size is {{maxSize}}",dropHere:"Drop the files here",dragAndDrop:"Drag and drop files here, or click to select files",removeFile:"Remove file",uploadDescription:"You can upload {{isMultiple ? 'multiple' : count}} files (up to {{maxSize}} each)",duplicateFile:"File name already exists in server cache"}},documentManager:{title:"Document Management",scanButton:"Scan",scanTooltip:"Scan documents in input folder",pipelineStatusButton:"Pipeline Status",pipelineStatusTooltip:"View pipeline status",uploadedTitle:"Uploaded Documents",uploadedDescription:"List of uploaded documents and their statuses.",emptyTitle:"No Documents",emptyDescription:"There are no uploaded documents yet.",columns:{id:"ID",summary:"Summary",status:"Status",length:"Length",chunks:"Chunks",created:"Created",updated:"Updated",metadata:"Metadata"},status:{all:"All",completed:"Completed",processing:"Processing",pending:"Pending",failed:"Failed"},errors:{loadFailed:`Failed to load documents
{{error}}`,scanFailed:`Failed to scan documents
{{error}}`,scanProgressFailed:`Failed to get scan progress
{{error}}`},fileNameLabel:"File Name",showButton:"Show",hideButton:"Hide",showFileNameTooltip:"Show file name",hideFileNameTooltip:"Hide file name"},pipelineStatus:{title:"Pipeline Status",busy:"Pipeline Busy",requestPending:"Request Pending",jobName:"Job Name",startTime:"Start Time",progress:"Progress",unit:"batch",latestMessage:"Latest Message",historyMessages:"History Messages",errors:{fetchFailed:`Failed to get pipeline status
{{error}}`}}},Gp={dataIsTruncated:"Graph data is truncated to Max Nodes",statusDialog:{title:"LightRAG Server Settings",description:"View current system status and connection information"},legend:"Legend",nodeTypes:{person:"Person",category:"Category",geo:"Geographic",location:"Location",organization:"Organization",event:"Event",equipment:"Equipment",weapon:"Weapon",animal:"Animal",unknown:"Unknown",object:"Object",group:"Group",technology:"Technology"},sideBar:{settings:{settings:"Settings",healthCheck:"Health Check",showPropertyPanel:"Show Property Panel",showSearchBar:"Show Search Bar",showNodeLabel:"Show Node Label",nodeDraggable:"Node Draggable",showEdgeLabel:"Show Edge Label",hideUnselectedEdges:"Hide Unselected Edges",edgeEvents:"Edge Events",maxQueryDepth:"Max Query Depth",maxNodes:"Max Nodes",maxLayoutIterations:"Max Layout Iterations",resetToDefault:"Reset to default",edgeSizeRange:"Edge Size Range",depth:"D",max:"Max",degree:"Degree",apiKey:"API Key",enterYourAPIkey:"Enter your API key",save:"Save",refreshLayout:"Refresh Layout"},zoomControl:{zoomIn:"Zoom In",zoomOut:"Zoom Out",resetZoom:"Reset Zoom",rotateCamera:"Clockwise Rotate",rotateCameraCounterClockwise:"Counter-Clockwise Rotate"},layoutsControl:{startAnimation:"Continue layout animation",stopAnimation:"Stop layout animation",layoutGraph:"Layout Graph",layouts:{Circular:"Circular",Circlepack:"Circlepack",Random:"Random",Noverlaps:"Noverlaps","Force Directed":"Force Directed","Force Atlas":"Force Atlas"}},fullScreenControl:{fullScreen:"Full Screen",windowed:"Windowed"},legendControl:{toggleLegend:"Toggle Legend"}},statusIndicator:{connected:"Connected",disconnected:"Disconnected"},statusCard:{unavailable:"Status information unavailable",storageInfo:"Storage Info",workingDirectory:"Working Directory",inputDirectory:"Input Directory",llmConfig:"LLM Configuration",llmBinding:"LLM Binding",llmBindingHost:"LLM Binding Host",llmModel:"LLM Model",maxTokens:"Max Tokens",embeddingConfig:"Embedding Configuration",embeddingBinding:"Embedding Binding",embeddingBindingHost:"Embedding Binding Host",embeddingModel:"Embedding Model",storageConfig:"Storage Configuration",kvStorage:"KV Storage",docStatusStorage:"Doc Status Storage",graphStorage:"Graph Storage",vectorStorage:"Vector Storage"},propertiesView:{editProperty:"Edit {{property}}",editPropertyDescription:"Edit the property value in the text area below.",errors:{duplicateName:"Node name already exists",updateFailed:"Failed to update node",tryAgainLater:"Please try again later"},success:{entityUpdated:"Node updated successfully",relationUpdated:"Relation updated successfully"},node:{title:"Node",id:"ID",labels:"Labels",degree:"Degree",properties:"Properties",relationships:"Relations(within subgraph)",expandNode:"Expand Node",pruneNode:"Prune Node",deleteAllNodesError:"Refuse to delete all nodes in the graph",nodesRemoved:"{{count}} nodes removed, including orphan nodes",noNewNodes:"No expandable nodes found",propertyNames:{description:"Description",entity_id:"Name",entity_type:"Type",source_id:"SrcID",Neighbour:"Neigh",file_path:"Source",keywords:"Keys",weight:"Weight"}},edge:{title:"Relationship",id:"ID",type:"Type",source:"Source",target:"Target",properties:"Properties"}},search:{placeholder:"Search nodes...",message:"And {count} others"},graphLabels:{selectTooltip:"Select query label",noLabels:"No labels found",label:"Label",placeholder:"Search labels...",andOthers:"And {count} others",refreshTooltip:"Reload data(After file added)"},emptyGraph:"Empty(Try Reload Again)"},Yp={chatMessage:{copyTooltip:"Copy to clipboard",copyError:"Failed to copy text to clipboard"},retrieval:{startPrompt:"Start a retrieval by typing your query below",clear:"Clear",send:"Send",placeholder:"Enter your query (Support prefix: /<Query Mode>)",error:"Error: Failed to get response",queryModeError:"Only supports the following query modes: {{modes}}",queryModePrefixInvalid:"Invalid query mode prefix. Use: /<mode> [space] your query"},querySettings:{parametersTitle:"Parameters",parametersDescription:"Configure your query parameters",queryMode:"Query Mode",queryModeTooltip:`Select the retrieval strategy:
• Naive: Basic search without advanced techniques
• Local: Context-dependent information retrieval
• Global: Utilizes global knowledge base
• Hybrid: Combines local and global retrieval
• Mix: Integrates knowledge graph with vector retrieval
• Bypass: Passes query directly to LLM without retrieval`,queryModeOptions:{naive:"Naive",local:"Local",global:"Global",hybrid:"Hybrid",mix:"Mix",bypass:"Bypass"},responseFormat:"Response Format",responseFormatTooltip:`Defines the response format. Examples:
• Multiple Paragraphs
• Single Paragraph
• Bullet Points`,responseFormatOptions:{multipleParagraphs:"Multiple Paragraphs",singleParagraph:"Single Paragraph",bulletPoints:"Bullet Points"},topK:"Top K Results",topKTooltip:"Number of top items to retrieve. Represents entities in 'local' mode and relationships in 'global' mode",topKPlaceholder:"Number of results",maxTokensTextUnit:"Max Tokens for Text Unit",maxTokensTextUnitTooltip:"Maximum number of tokens allowed for each retrieved text chunk",maxTokensGlobalContext:"Max Tokens for Global Context",maxTokensGlobalContextTooltip:"Maximum number of tokens allocated for relationship descriptions in global retrieval",maxTokensLocalContext:"Max Tokens for Local Context",maxTokensLocalContextTooltip:"Maximum number of tokens allocated for entity descriptions in local retrieval",historyTurns:"History Turns",historyTurnsTooltip:"Number of complete conversation turns (user-assistant pairs) to consider in the response context",historyTurnsPlaceholder:"Number of history turns",hlKeywords:"High-Level Keywords",hlKeywordsTooltip:"List of high-level keywords to prioritize in retrieval. Separate with commas",hlkeywordsPlaceHolder:"Enter keywords",llKeywords:"Low-Level Keywords",llKeywordsTooltip:"List of low-level keywords to refine retrieval focus. Separate with commas",onlyNeedContext:"Only Need Context",onlyNeedContextTooltip:"If True, only returns the retrieved context without generating a response",onlyNeedPrompt:"Only Need Prompt",onlyNeedPromptTooltip:"If True, only returns the generated prompt without producing a response",streamResponse:"Stream Response",streamResponseTooltip:"If True, enables streaming output for real-time responses"}},wp={loading:"Loading API Documentation..."},Xp={title:"API Key is required",description:"Please enter your API key to access the service",placeholder:"Enter your API key",save:"Save"},Qp={settings:_p,header:Hp,login:Lp,common:qp,documentPanel:Bp,graphPanel:Gp,retrievePanel:Yp,apiSite:wp,apiKeyAlert:Xp},Vp={language:"语言",theme:"主题",light:"浅色",dark:"深色",system:"系统"},Kp={documents:"文档",knowledgeGraph:"知识图谱",retrieval:"检索",api:"API",projectRepository:"项目仓库",logout:"退出登录",themeToggle:{switchToLight:"切换到浅色主题",switchToDark:"切换到深色主题"}},Zp={description:"请输入您的账号和密码登录系统",username:"用户名",usernamePlaceholder:"请输入用户名",password:"密码",passwordPlaceholder:"请输入密码",loginButton:"登录",loggingIn:"登录中...",successMessage:"登录成功",errorEmptyFields:"请输入您的用户名和密码",errorInvalidCredentials:"登录失败，请检查用户名和密码",authDisabled:"认证已禁用，使用无需登陆模式。",guestMode:"无需登陆"},kp={cancel:"取消",save:"保存"},Jp={clearDocuments:{button:"清空",tooltip:"清空文档",title:"清空文档",description:"此操作将从系统中移除所有文档",warning:"警告：此操作将永久删除所有文档，无法恢复！",confirm:"确定要清空所有文档吗？",confirmPrompt:"请输入 yes 确认操作",confirmPlaceholder:"输入 yes 确认",clearCache:"清空LLM缓存",confirmButton:"确定",success:"文档清空成功",cacheCleared:"缓存清空成功",cacheClearFailed:`清空缓存失败：
{{error}}`,failed:`清空文档失败：
{{message}}`,error:`清空文档失败：
{{error}}`},uploadDocuments:{button:"上传",tooltip:"上传文档",title:"上传文档",description:"拖拽文件到此处或点击浏览",single:{uploading:"正在上传 {{name}}：{{percent}}%",success:`上传成功：
{{name}} 上传完成`,failed:`上传失败：
{{name}}
{{message}}`,error:`上传失败：
{{name}}
{{error}}`},batch:{uploading:"正在上传文件...",success:"文件上传完成",error:"部分文件上传失败"},generalError:`上传失败
{{error}}`,fileTypes:"支持的文件类型：TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS",fileUploader:{singleFileLimit:"一次只能上传一个文件",maxFilesLimit:"最多只能上传 {{count}} 个文件",fileRejected:"文件 {{name}} 被拒绝",unsupportedType:"不支持的文件类型",fileTooLarge:"文件过大，最大允许 {{maxSize}}",dropHere:"将文件拖放到此处",dragAndDrop:"拖放文件到此处，或点击选择文件",removeFile:"移除文件",uploadDescription:"您可以上传{{isMultiple ? '多个' : count}}个文件（每个文件最大{{maxSize}}）",duplicateFile:"文件名与服务器上的缓存重复"}},documentManager:{title:"文档管理",scanButton:"扫描",scanTooltip:"扫描输入目录中的文档",pipelineStatusButton:"流水线状态",pipelineStatusTooltip:"查看流水线状态",uploadedTitle:"已上传文档",uploadedDescription:"已上传文档列表及其状态",emptyTitle:"无文档",emptyDescription:"还没有上传任何文档",columns:{id:"ID",summary:"摘要",status:"状态",length:"长度",chunks:"分块",created:"创建时间",updated:"更新时间",metadata:"元数据"},status:{all:"全部",completed:"已完成",processing:"处理中",pending:"等待中",failed:"失败"},errors:{loadFailed:`加载文档失败
{{error}}`,scanFailed:`扫描文档失败
{{error}}`,scanProgressFailed:`获取扫描进度失败
{{error}}`},fileNameLabel:"文件名",showButton:"显示",hideButton:"隐藏",showFileNameTooltip:"显示文件名",hideFileNameTooltip:"隐藏文件名"},pipelineStatus:{title:"流水线状态",busy:"流水线忙碌",requestPending:"待处理请求",jobName:"作业名称",startTime:"开始时间",progress:"进度",unit:"批",latestMessage:"最新消息",historyMessages:"历史消息",errors:{fetchFailed:`获取流水线状态失败
{{error}}`}}},Fp={dataIsTruncated:"图数据已截断至最大返回节点数",statusDialog:{title:"LightRAG 服务器设置",description:"查看当前系统状态和连接信息"},legend:"图例",nodeTypes:{person:"人物角色",category:"分类",geo:"地理名称",location:"位置",organization:"组织机构",event:"事件",equipment:"装备",weapon:"武器",animal:"动物",unknown:"未知",object:"物品",group:"群组",technology:"技术"},sideBar:{settings:{settings:"设置",healthCheck:"健康检查",showPropertyPanel:"显示属性面板",showSearchBar:"显示搜索栏",showNodeLabel:"显示节点标签",nodeDraggable:"节点可拖动",showEdgeLabel:"显示边标签",hideUnselectedEdges:"隐藏未选中的边",edgeEvents:"边事件",maxQueryDepth:"最大查询深度",maxNodes:"最大返回节点数",maxLayoutIterations:"最大布局迭代次数",resetToDefault:"重置为默认值",edgeSizeRange:"边粗细范围",depth:"深",max:"Max",degree:"邻边",apiKey:"API密钥",enterYourAPIkey:"输入您的API密钥",save:"保存",refreshLayout:"刷新布局"},zoomControl:{zoomIn:"放大",zoomOut:"缩小",resetZoom:"重置缩放",rotateCamera:"顺时针旋转图形",rotateCameraCounterClockwise:"逆时针旋转图形"},layoutsControl:{startAnimation:"继续布局动画",stopAnimation:"停止布局动画",layoutGraph:"图布局",layouts:{Circular:"环形",Circlepack:"圆形打包",Random:"随机",Noverlaps:"无重叠","Force Directed":"力导向","Force Atlas":"力地图"}},fullScreenControl:{fullScreen:"全屏",windowed:"窗口"},legendControl:{toggleLegend:"切换图例显示"}},statusIndicator:{connected:"已连接",disconnected:"未连接"},statusCard:{unavailable:"状态信息不可用",storageInfo:"存储信息",workingDirectory:"工作目录",inputDirectory:"输入目录",llmConfig:"LLM配置",llmBinding:"LLM绑定",llmBindingHost:"LLM绑定主机",llmModel:"LLM模型",maxTokens:"最大令牌数",embeddingConfig:"嵌入配置",embeddingBinding:"嵌入绑定",embeddingBindingHost:"嵌入绑定主机",embeddingModel:"嵌入模型",storageConfig:"存储配置",kvStorage:"KV存储",docStatusStorage:"文档状态存储",graphStorage:"图存储",vectorStorage:"向量存储"},propertiesView:{editProperty:"编辑{{property}}",editPropertyDescription:"在下方文本区域编辑属性值。",errors:{duplicateName:"节点名称已存在",updateFailed:"更新节点失败",tryAgainLater:"请稍后重试"},success:{entityUpdated:"节点更新成功",relationUpdated:"关系更新成功"},node:{title:"节点",id:"ID",labels:"标签",degree:"度数",properties:"属性",relationships:"关系(子图内)",expandNode:"扩展节点",pruneNode:"修剪节点",deleteAllNodesError:"拒绝删除图中的所有节点",nodesRemoved:"已删除 {{count}} 个节点，包括孤立节点",noNewNodes:"没有发现可以扩展的节点",propertyNames:{description:"描述",entity_id:"名称",entity_type:"类型",source_id:"信源ID",Neighbour:"邻接",file_path:"信源",keywords:"Keys",weight:"权重"}},edge:{title:"关系",id:"ID",type:"类型",source:"源节点",target:"目标节点",properties:"属性"}},search:{placeholder:"搜索节点...",message:"还有 {count} 个"},graphLabels:{selectTooltip:"选择查询标签",noLabels:"未找到标签",label:"标签",placeholder:"搜索标签...",andOthers:"还有 {count} 个",refreshTooltip:"重载图形数据(添加文件后需重载)"},emptyGraph:"无数据(请重载图形数据)"},$p={chatMessage:{copyTooltip:"复制到剪贴板",copyError:"复制文本到剪贴板失败"},retrieval:{startPrompt:"输入查询开始检索",clear:"清空",send:"发送",placeholder:"输入查询内容 (支持模式前缀: /<Query Mode>)",error:"错误：获取响应失败",queryModeError:"仅支持以下查询模式：{{modes}}",queryModePrefixInvalid:"无效的查询模式前缀。请使用：/<模式> [空格] 查询内容"},querySettings:{parametersTitle:"参数",parametersDescription:"配置查询参数",queryMode:"查询模式",queryModeTooltip:`选择检索策略：
• Naive：基础搜索，无高级技术
• Local：上下文相关信息检索
• Global：利用全局知识库
• Hybrid：结合本地和全局检索
• Mix：整合知识图谱和向量检索
• Bypass：直接传递查询到LLM，不进行检索`,queryModeOptions:{naive:"Naive",local:"Local",global:"Global",hybrid:"Hybrid",mix:"Mix",bypass:"Bypass"},responseFormat:"响应格式",responseFormatTooltip:`定义响应格式。例如：
• 多段落
• 单段落
• 要点`,responseFormatOptions:{multipleParagraphs:"多段落",singleParagraph:"单段落",bulletPoints:"要点"},topK:"Top K结果",topKTooltip:"检索的顶部项目数。在'local'模式下表示实体，在'global'模式下表示关系",topKPlaceholder:"结果数量",maxTokensTextUnit:"文本单元最大令牌数",maxTokensTextUnitTooltip:"每个检索文本块允许的最大令牌数",maxTokensGlobalContext:"全局上下文最大令牌数",maxTokensGlobalContextTooltip:"全局检索中关系描述的最大令牌数",maxTokensLocalContext:"本地上下文最大令牌数",maxTokensLocalContextTooltip:"本地检索中实体描述的最大令牌数",historyTurns:"历史轮次",historyTurnsTooltip:"响应上下文中考虑的完整对话轮次（用户-助手对）数量",historyTurnsPlaceholder:"历史轮次数",hlKeywords:"高级关键词",hlKeywordsTooltip:"检索中优先考虑的高级关键词列表。用逗号分隔",hlkeywordsPlaceHolder:"输入关键词",llKeywords:"低级关键词",llKeywordsTooltip:"用于细化检索重点的低级关键词列表。用逗号分隔",onlyNeedContext:"仅需上下文",onlyNeedContextTooltip:"如果为True，仅返回检索到的上下文而不生成响应",onlyNeedPrompt:"仅需提示",onlyNeedPromptTooltip:"如果为True，仅返回生成的提示而不产生响应",streamResponse:"流式响应",streamResponseTooltip:"如果为True，启用实时流式输出响应"}},Wp={loading:"正在加载 API 文档..."},Pp={title:"需要 API Key",description:"请输入您的 API Key 以访问服务",placeholder:"请输入 API Key",save:"保存"},Ip={settings:Vp,header:Kp,login:Zp,common:kp,documentPanel:Jp,graphPanel:Fp,retrievePanel:$p,apiSite:Wp,apiKeyAlert:Pp},ey={language:"Langue",theme:"Thème",light:"Clair",dark:"Sombre",system:"Système"},ly={documents:"Documents",knowledgeGraph:"Graphe de connaissances",retrieval:"Récupération",api:"API",projectRepository:"Référentiel du projet",logout:"Déconnexion",themeToggle:{switchToLight:"Passer au thème clair",switchToDark:"Passer au thème sombre"}},ty={description:"Veuillez entrer votre compte et mot de passe pour vous connecter au système",username:"Nom d'utilisateur",usernamePlaceholder:"Veuillez saisir un nom d'utilisateur",password:"Mot de passe",passwordPlaceholder:"Veuillez saisir un mot de passe",loginButton:"Connexion",loggingIn:"Connexion en cours...",successMessage:"Connexion réussie",errorEmptyFields:"Veuillez saisir votre nom d'utilisateur et mot de passe",errorInvalidCredentials:"Échec de la connexion, veuillez vérifier le nom d'utilisateur et le mot de passe",authDisabled:"L'authentification est désactivée. Utilisation du mode sans connexion.",guestMode:"Mode sans connexion"},ay={cancel:"Annuler",save:"Sauvegarder"},ny={clearDocuments:{button:"Effacer",tooltip:"Effacer les documents",title:"Effacer les documents",description:"Cette action supprimera tous les documents du système",warning:"ATTENTION : Cette action supprimera définitivement tous les documents et ne peut pas être annulée !",confirm:"Voulez-vous vraiment effacer tous les documents ?",confirmPrompt:"Tapez 'yes' pour confirmer cette action",confirmPlaceholder:"Tapez yes pour confirmer",clearCache:"Effacer le cache LLM",confirmButton:"OUI",success:"Documents effacés avec succès",cacheCleared:"Cache effacé avec succès",cacheClearFailed:`Échec de l'effacement du cache :
{{error}}`,failed:`Échec de l'effacement des documents :
{{message}}`,error:`Échec de l'effacement des documents :
{{error}}`},uploadDocuments:{button:"Télécharger",tooltip:"Télécharger des documents",title:"Télécharger des documents",description:"Glissez-déposez vos documents ici ou cliquez pour parcourir.",single:{uploading:"Téléchargement de {{name}} : {{percent}}%",success:`Succès du téléchargement :
{{name}} téléchargé avec succès`,failed:`Échec du téléchargement :
{{name}}
{{message}}`,error:`Échec du téléchargement :
{{name}}
{{error}}`},batch:{uploading:"Téléchargement des fichiers...",success:"Fichiers téléchargés avec succès",error:"Certains fichiers n'ont pas pu être téléchargés"},generalError:`Échec du téléchargement
{{error}}`,fileTypes:"Types pris en charge : TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS",fileUploader:{singleFileLimit:"Impossible de télécharger plus d'un fichier à la fois",maxFilesLimit:"Impossible de télécharger plus de {{count}} fichiers",fileRejected:"Le fichier {{name}} a été rejeté",unsupportedType:"Type de fichier non pris en charge",fileTooLarge:"Fichier trop volumineux, taille maximale {{maxSize}}",dropHere:"Déposez les fichiers ici",dragAndDrop:"Glissez et déposez les fichiers ici, ou cliquez pour sélectionner",removeFile:"Supprimer le fichier",uploadDescription:"Vous pouvez télécharger {{isMultiple ? 'plusieurs' : count}} fichiers (jusqu'à {{maxSize}} chacun)",duplicateFile:"Le nom du fichier existe déjà dans le cache du serveur"}},documentManager:{title:"Gestion des documents",scanButton:"Scanner",scanTooltip:"Scanner les documents dans le dossier d'entrée",pipelineStatusButton:"État du Pipeline",pipelineStatusTooltip:"Voir l'état du pipeline",uploadedTitle:"Documents téléchargés",uploadedDescription:"Liste des documents téléchargés et leurs statuts.",emptyTitle:"Aucun document",emptyDescription:"Il n'y a pas encore de documents téléchargés.",columns:{id:"ID",summary:"Résumé",status:"Statut",length:"Longueur",chunks:"Fragments",created:"Créé",updated:"Mis à jour",metadata:"Métadonnées"},status:{all:"Tous",completed:"Terminé",processing:"En traitement",pending:"En attente",failed:"Échoué"},errors:{loadFailed:`Échec du chargement des documents
{{error}}`,scanFailed:`Échec de la numérisation des documents
{{error}}`,scanProgressFailed:`Échec de l'obtention de la progression de la numérisation
{{error}}`},fileNameLabel:"Nom du fichier",showButton:"Afficher",hideButton:"Masquer",showFileNameTooltip:"Afficher le nom du fichier",hideFileNameTooltip:"Masquer le nom du fichier"},pipelineStatus:{title:"État du Pipeline",busy:"Pipeline occupé",requestPending:"Requête en attente",jobName:"Nom du travail",startTime:"Heure de début",progress:"Progression",unit:"lot",latestMessage:"Dernier message",historyMessages:"Historique des messages",errors:{fetchFailed:`Échec de la récupération de l'état du pipeline
{{error}}`}}},uy={dataIsTruncated:"Les données du graphe sont tronquées au nombre maximum de nœuds",statusDialog:{title:"Paramètres du Serveur LightRAG",description:"Afficher l'état actuel du système et les informations de connexion"},legend:"Légende",nodeTypes:{person:"Personne",category:"Catégorie",geo:"Géographique",location:"Emplacement",organization:"Organisation",event:"Événement",equipment:"Équipement",weapon:"Arme",animal:"Animal",unknown:"Inconnu",object:"Objet",group:"Groupe",technology:"Technologie"},sideBar:{settings:{settings:"Paramètres",healthCheck:"Vérification de l'état",showPropertyPanel:"Afficher le panneau des propriétés",showSearchBar:"Afficher la barre de recherche",showNodeLabel:"Afficher l'étiquette du nœud",nodeDraggable:"Nœud déplaçable",showEdgeLabel:"Afficher l'étiquette de l'arête",hideUnselectedEdges:"Masquer les arêtes non sélectionnées",edgeEvents:"Événements des arêtes",maxQueryDepth:"Profondeur maximale de la requête",maxNodes:"Nombre maximum de nœuds",maxLayoutIterations:"Itérations maximales de mise en page",resetToDefault:"Réinitialiser par défaut",edgeSizeRange:"Plage de taille des arêtes",depth:"D",max:"Max",degree:"Degré",apiKey:"Clé API",enterYourAPIkey:"Entrez votre clé API",save:"Sauvegarder",refreshLayout:"Actualiser la mise en page"},zoomControl:{zoomIn:"Zoom avant",zoomOut:"Zoom arrière",resetZoom:"Réinitialiser le zoom",rotateCamera:"Rotation horaire",rotateCameraCounterClockwise:"Rotation antihoraire"},layoutsControl:{startAnimation:"Démarrer l'animation de mise en page",stopAnimation:"Arrêter l'animation de mise en page",layoutGraph:"Mettre en page le graphe",layouts:{Circular:"Circulaire",Circlepack:"Paquet circulaire",Random:"Aléatoire",Noverlaps:"Sans chevauchement","Force Directed":"Dirigé par la force","Force Atlas":"Atlas de force"}},fullScreenControl:{fullScreen:"Plein écran",windowed:"Fenêtré"},legendControl:{toggleLegend:"Basculer la légende"}},statusIndicator:{connected:"Connecté",disconnected:"Déconnecté"},statusCard:{unavailable:"Informations sur l'état indisponibles",storageInfo:"Informations de stockage",workingDirectory:"Répertoire de travail",inputDirectory:"Répertoire d'entrée",llmConfig:"Configuration du modèle de langage",llmBinding:"Liaison du modèle de langage",llmBindingHost:"Hôte de liaison du modèle de langage",llmModel:"Modèle de langage",maxTokens:"Nombre maximum de jetons",embeddingConfig:"Configuration d'incorporation",embeddingBinding:"Liaison d'incorporation",embeddingBindingHost:"Hôte de liaison d'incorporation",embeddingModel:"Modèle d'incorporation",storageConfig:"Configuration de stockage",kvStorage:"Stockage clé-valeur",docStatusStorage:"Stockage de l'état des documents",graphStorage:"Stockage du graphe",vectorStorage:"Stockage vectoriel"},propertiesView:{editProperty:"Modifier {{property}}",editPropertyDescription:"Modifiez la valeur de la propriété dans la zone de texte ci-dessous.",errors:{duplicateName:"Le nom du nœud existe déjà",updateFailed:"Échec de la mise à jour du nœud",tryAgainLater:"Veuillez réessayer plus tard"},success:{entityUpdated:"Nœud mis à jour avec succès",relationUpdated:"Relation mise à jour avec succès"},node:{title:"Nœud",id:"ID",labels:"Étiquettes",degree:"Degré",properties:"Propriétés",relationships:"Relations(dans le sous-graphe)",expandNode:"Développer le nœud",pruneNode:"Élaguer le nœud",deleteAllNodesError:"Refus de supprimer tous les nœuds du graphe",nodesRemoved:"{{count}} nœuds supprimés, y compris les nœuds orphelins",noNewNodes:"Aucun nœud développable trouvé",propertyNames:{description:"Description",entity_id:"Nom",entity_type:"Type",source_id:"ID source",Neighbour:"Voisin",file_path:"Source",keywords:"Keys",weight:"Poids"}},edge:{title:"Relation",id:"ID",type:"Type",source:"Source",target:"Cible",properties:"Propriétés"}},search:{placeholder:"Rechercher des nœuds...",message:"Et {{count}} autres"},graphLabels:{selectTooltip:"Sélectionner l'étiquette de la requête",noLabels:"Aucune étiquette trouvée",label:"Étiquette",placeholder:"Rechercher des étiquettes...",andOthers:"Et {{count}} autres",refreshTooltip:"Recharger les données (Après l'ajout de fichier)"},emptyGraph:"Vide (Essayez de recharger)"},iy={chatMessage:{copyTooltip:"Copier dans le presse-papiers",copyError:"Échec de la copie du texte dans le presse-papiers"},retrieval:{startPrompt:"Démarrez une récupération en tapant votre requête ci-dessous",clear:"Effacer",send:"Envoyer",placeholder:"Tapez votre requête (Préfixe de requête : /<Query Mode>)",error:"Erreur : Échec de l'obtention de la réponse",queryModeError:"Seuls les modes de requête suivants sont pris en charge : {{modes}}",queryModePrefixInvalid:"Préfixe de mode de requête invalide. Utilisez : /<mode> [espace] votre requête"},querySettings:{parametersTitle:"Paramètres",parametersDescription:"Configurez vos paramètres de requête",queryMode:"Mode de requête",queryModeTooltip:`Sélectionnez la stratégie de récupération :
• Naïf : Recherche de base sans techniques avancées
• Local : Récupération d'informations dépendante du contexte
• Global : Utilise une base de connaissances globale
• Hybride : Combine récupération locale et globale
• Mixte : Intègre le graphe de connaissances avec la récupération vectorielle
• Bypass : Transmet directement la requête au LLM sans récupération`,queryModeOptions:{naive:"Naïf",local:"Local",global:"Global",hybrid:"Hybride",mix:"Mixte",bypass:"Bypass"},responseFormat:"Format de réponse",responseFormatTooltip:`Définit le format de la réponse. Exemples :
• Plusieurs paragraphes
• Paragraphe unique
• Points à puces`,responseFormatOptions:{multipleParagraphs:"Plusieurs paragraphes",singleParagraph:"Paragraphe unique",bulletPoints:"Points à puces"},topK:"Top K résultats",topKTooltip:"Nombre d'éléments supérieurs à récupérer. Représente les entités en mode 'local' et les relations en mode 'global'",topKPlaceholder:"Nombre de résultats",maxTokensTextUnit:"Nombre maximum de jetons pour l'unité de texte",maxTokensTextUnitTooltip:"Nombre maximum de jetons autorisés pour chaque fragment de texte récupéré",maxTokensGlobalContext:"Nombre maximum de jetons pour le contexte global",maxTokensGlobalContextTooltip:"Nombre maximum de jetons alloués pour les descriptions des relations dans la récupération globale",maxTokensLocalContext:"Nombre maximum de jetons pour le contexte local",maxTokensLocalContextTooltip:"Nombre maximum de jetons alloués pour les descriptions des entités dans la récupération locale",historyTurns:"Tours d'historique",historyTurnsTooltip:"Nombre de tours complets de conversation (paires utilisateur-assistant) à prendre en compte dans le contexte de la réponse",historyTurnsPlaceholder:"Nombre de tours d'historique",hlKeywords:"Mots-clés de haut niveau",hlKeywordsTooltip:"Liste de mots-clés de haut niveau à prioriser dans la récupération. Séparez par des virgules",hlkeywordsPlaceHolder:"Entrez les mots-clés",llKeywords:"Mots-clés de bas niveau",llKeywordsTooltip:"Liste de mots-clés de bas niveau pour affiner la focalisation de la récupération. Séparez par des virgules",onlyNeedContext:"Besoin uniquement du contexte",onlyNeedContextTooltip:"Si vrai, ne renvoie que le contexte récupéré sans générer de réponse",onlyNeedPrompt:"Besoin uniquement de l'invite",onlyNeedPromptTooltip:"Si vrai, ne renvoie que l'invite générée sans produire de réponse",streamResponse:"Réponse en flux",streamResponseTooltip:"Si vrai, active la sortie en flux pour des réponses en temps réel"}},cy={loading:"Chargement de la documentation de l'API..."},sy={title:"Clé API requise",description:"Veuillez entrer votre clé API pour accéder au service",placeholder:"Entrez votre clé API",save:"Sauvegarder"},oy={settings:ey,header:ly,login:ty,common:ay,documentPanel:ny,graphPanel:uy,retrievePanel:iy,apiSite:cy,apiKeyAlert:sy},fy={language:"اللغة",theme:"السمة",light:"فاتح",dark:"داكن",system:"النظام"},ry={documents:"المستندات",knowledgeGraph:"شبكة المعرفة",retrieval:"الاسترجاع",api:"واجهة برمجة التطبيقات",projectRepository:"مستودع المشروع",logout:"تسجيل الخروج",themeToggle:{switchToLight:"التحويل إلى السمة الفاتحة",switchToDark:"التحويل إلى السمة الداكنة"}},dy={description:"الرجاء إدخال حسابك وكلمة المرور لتسجيل الدخول إلى النظام",username:"اسم المستخدم",usernamePlaceholder:"الرجاء إدخال اسم المستخدم",password:"كلمة المرور",passwordPlaceholder:"الرجاء إدخال كلمة المرور",loginButton:"تسجيل الدخول",loggingIn:"جاري تسجيل الدخول...",successMessage:"تم تسجيل الدخول بنجاح",errorEmptyFields:"الرجاء إدخال اسم المستخدم وكلمة المرور",errorInvalidCredentials:"فشل تسجيل الدخول، يرجى التحقق من اسم المستخدم وكلمة المرور",authDisabled:"تم تعطيل المصادقة. استخدام وضع بدون تسجيل دخول.",guestMode:"وضع بدون تسجيل دخول"},my={cancel:"إلغاء",save:"حفظ"},hy={clearDocuments:{button:"مسح",tooltip:"مسح المستندات",title:"مسح المستندات",description:"سيؤدي هذا إلى إزالة جميع المستندات من النظام",warning:"تحذير: سيؤدي هذا الإجراء إلى حذف جميع المستندات بشكل دائم ولا يمكن التراجع عنه!",confirm:"هل تريد حقًا مسح جميع المستندات؟",confirmPrompt:"اكتب 'yes' لتأكيد هذا الإجراء",confirmPlaceholder:"اكتب yes للتأكيد",clearCache:"مسح كاش نموذج اللغة",confirmButton:"نعم",success:"تم مسح المستندات بنجاح",cacheCleared:"تم مسح ذاكرة التخزين المؤقت بنجاح",cacheClearFailed:`فشل مسح ذاكرة التخزين المؤقت:
{{error}}`,failed:`فشل مسح المستندات:
{{message}}`,error:`فشل مسح المستندات:
{{error}}`},uploadDocuments:{button:"رفع",tooltip:"رفع المستندات",title:"رفع المستندات",description:"اسحب وأفلت مستنداتك هنا أو انقر للتصفح.",single:{uploading:"جارٍ الرفع {{name}}: {{percent}}%",success:`نجاح الرفع:
تم رفع {{name}} بنجاح`,failed:`فشل الرفع:
{{name}}
{{message}}`,error:`فشل الرفع:
{{name}}
{{error}}`},batch:{uploading:"جارٍ رفع الملفات...",success:"تم رفع الملفات بنجاح",error:"فشل رفع بعض الملفات"},generalError:`فشل الرفع
{{error}}`,fileTypes:"الأنواع المدعومة: TXT، MD، DOCX، PDF، PPTX، RTF، ODT، EPUB، HTML، HTM، TEX، JSON، XML، YAML، YML، CSV، LOG، CONF، INI، PROPERTIES، SQL، BAT، SH، C، CPP، PY، JAVA، JS، TS، SWIFT، GO، RB، PHP، CSS، SCSS، LESS",fileUploader:{singleFileLimit:"لا يمكن رفع أكثر من ملف واحد في المرة الواحدة",maxFilesLimit:"لا يمكن رفع أكثر من {{count}} ملفات",fileRejected:"تم رفض الملف {{name}}",unsupportedType:"نوع الملف غير مدعوم",fileTooLarge:"حجم الملف كبير جدًا، الحد الأقصى {{maxSize}}",dropHere:"أفلت الملفات هنا",dragAndDrop:"اسحب وأفلت الملفات هنا، أو انقر للاختيار",removeFile:"إزالة الملف",uploadDescription:"يمكنك رفع {{isMultiple ? 'عدة' : count}} ملفات (حتى {{maxSize}} لكل منها)",duplicateFile:"اسم الملف موجود بالفعل في ذاكرة التخزين المؤقت للخادم"}},documentManager:{title:"إدارة المستندات",scanButton:"مسح ضوئي",scanTooltip:"مسح المستندات ضوئيًا في مجلد الإدخال",pipelineStatusButton:"حالة خط المعالجة",pipelineStatusTooltip:"عرض حالة خط المعالجة",uploadedTitle:"المستندات المرفوعة",uploadedDescription:"قائمة المستندات المرفوعة وحالاتها.",emptyTitle:"لا توجد مستندات",emptyDescription:"لا توجد مستندات مرفوعة بعد.",columns:{id:"المعرف",summary:"الملخص",status:"الحالة",length:"الطول",chunks:"الأجزاء",created:"تم الإنشاء",updated:"تم التحديث",metadata:"البيانات الوصفية"},status:{all:"الكل",completed:"مكتمل",processing:"قيد المعالجة",pending:"معلق",failed:"فشل"},errors:{loadFailed:`فشل تحميل المستندات
{{error}}`,scanFailed:`فشل مسح المستندات
{{error}}`,scanProgressFailed:`فشل الحصول على تقدم المسح
{{error}}`},fileNameLabel:"اسم الملف",showButton:"عرض",hideButton:"إخفاء",showFileNameTooltip:"عرض اسم الملف",hideFileNameTooltip:"إخفاء اسم الملف"},pipelineStatus:{title:"حالة خط المعالجة",busy:"خط المعالجة مشغول",requestPending:"الطلب معلق",jobName:"اسم المهمة",startTime:"وقت البدء",progress:"التقدم",unit:"دفعة",latestMessage:"آخر رسالة",historyMessages:"سجل الرسائل",errors:{fetchFailed:`فشل في جلب حالة خط المعالجة
{{error}}`}}},gy={dataIsTruncated:"تم اقتصار بيانات الرسم البياني على الحد الأقصى للعقد",statusDialog:{title:"إعدادات خادم LightRAG",description:"عرض حالة النظام الحالية ومعلومات الاتصال"},legend:"المفتاح",nodeTypes:{person:"شخص",category:"فئة",geo:"كيان جغرافي",location:"موقع",organization:"منظمة",event:"حدث",equipment:"معدات",weapon:"سلاح",animal:"حيوان",unknown:"غير معروف",object:"مصنوع",group:"مجموعة",technology:"العلوم"},sideBar:{settings:{settings:"الإعدادات",healthCheck:"فحص الحالة",showPropertyPanel:"إظهار لوحة الخصائص",showSearchBar:"إظهار شريط البحث",showNodeLabel:"إظهار تسمية العقدة",nodeDraggable:"العقدة قابلة للسحب",showEdgeLabel:"إظهار تسمية الحافة",hideUnselectedEdges:"إخفاء الحواف غير المحددة",edgeEvents:"أحداث الحافة",maxQueryDepth:"أقصى عمق للاستعلام",maxNodes:"الحد الأقصى للعقد",maxLayoutIterations:"أقصى تكرارات التخطيط",resetToDefault:"إعادة التعيين إلى الافتراضي",edgeSizeRange:"نطاق حجم الحافة",depth:"D",max:"Max",degree:"الدرجة",apiKey:"مفتاح واجهة برمجة التطبيقات",enterYourAPIkey:"أدخل مفتاح واجهة برمجة التطبيقات الخاص بك",save:"حفظ",refreshLayout:"تحديث التخطيط"},zoomControl:{zoomIn:"تكبير",zoomOut:"تصغير",resetZoom:"إعادة تعيين التكبير",rotateCamera:"تدوير في اتجاه عقارب الساعة",rotateCameraCounterClockwise:"تدوير عكس اتجاه عقارب الساعة"},layoutsControl:{startAnimation:"بدء حركة التخطيط",stopAnimation:"إيقاف حركة التخطيط",layoutGraph:"تخطيط الرسم البياني",layouts:{Circular:"دائري",Circlepack:"حزمة دائرية",Random:"عشوائي",Noverlaps:"بدون تداخل","Force Directed":"موجه بالقوة","Force Atlas":"أطلس القوة"}},fullScreenControl:{fullScreen:"شاشة كاملة",windowed:"نوافذ"},legendControl:{toggleLegend:"تبديل المفتاح"}},statusIndicator:{connected:"متصل",disconnected:"غير متصل"},statusCard:{unavailable:"معلومات الحالة غير متوفرة",storageInfo:"معلومات التخزين",workingDirectory:"دليل العمل",inputDirectory:"دليل الإدخال",llmConfig:"تكوين نموذج اللغة الكبير",llmBinding:"ربط نموذج اللغة الكبير",llmBindingHost:"مضيف ربط نموذج اللغة الكبير",llmModel:"نموذج اللغة الكبير",maxTokens:"أقصى عدد من الرموز",embeddingConfig:"تكوين التضمين",embeddingBinding:"ربط التضمين",embeddingBindingHost:"مضيف ربط التضمين",embeddingModel:"نموذج التضمين",storageConfig:"تكوين التخزين",kvStorage:"تخزين المفتاح-القيمة",docStatusStorage:"تخزين حالة المستند",graphStorage:"تخزين الرسم البياني",vectorStorage:"تخزين المتجهات"},propertiesView:{editProperty:"تعديل {{property}}",editPropertyDescription:"قم بتحرير قيمة الخاصية في منطقة النص أدناه.",errors:{duplicateName:"اسم العقدة موجود بالفعل",updateFailed:"فشل تحديث العقدة",tryAgainLater:"يرجى المحاولة مرة أخرى لاحقًا"},success:{entityUpdated:"تم تحديث العقدة بنجاح",relationUpdated:"تم تحديث العلاقة بنجاح"},node:{title:"عقدة",id:"المعرف",labels:"التسميات",degree:"الدرجة",properties:"الخصائص",relationships:"العلاقات (داخل الرسم الفرعي)",expandNode:"توسيع العقدة",pruneNode:"تقليم العقدة",deleteAllNodesError:"رفض حذف جميع العقد في الرسم البياني",nodesRemoved:"تم إزالة {{count}} عقدة، بما في ذلك العقد اليتيمة",noNewNodes:"لم يتم العثور على عقد قابلة للتوسيع",propertyNames:{description:"الوصف",entity_id:"الاسم",entity_type:"النوع",source_id:"معرف المصدر",Neighbour:"الجار",file_path:"المصدر",keywords:"الكلمات الرئيسية",weight:"الوزن"}},edge:{title:"علاقة",id:"المعرف",type:"النوع",source:"المصدر",target:"الهدف",properties:"الخصائص"}},search:{placeholder:"ابحث في العقد...",message:"و {{count}} آخرون"},graphLabels:{selectTooltip:"حدد تسمية الاستعلام",noLabels:"لم يتم العثور على تسميات",label:"التسمية",placeholder:"ابحث في التسميات...",andOthers:"و {{count}} آخرون",refreshTooltip:"إعادة تحميل البيانات (بعد إضافة الملف)"},emptyGraph:"فارغ (حاول إعادة التحميل)"},py={chatMessage:{copyTooltip:"نسخ إلى الحافظة",copyError:"فشل نسخ النص إلى الحافظة"},retrieval:{startPrompt:"ابدأ الاسترجاع بكتابة استفسارك أدناه",clear:"مسح",send:"إرسال",placeholder:"اكتب استفسارك (بادئة وضع الاستعلام: /<Query Mode>)",error:"خطأ: فشل الحصول على الرد",queryModeError:"يُسمح فقط بأنماط الاستعلام التالية: {{modes}}",queryModePrefixInvalid:"بادئة وضع الاستعلام غير صالحة. استخدم: /<الوضع> [مسافة] استفسارك"},querySettings:{parametersTitle:"المعلمات",parametersDescription:"تكوين معلمات الاستعلام الخاص بك",queryMode:"وضع الاستعلام",queryModeTooltip:`حدد استراتيجية الاسترجاع:
• ساذج: بحث أساسي بدون تقنيات متقدمة
• محلي: استرجاع معلومات يعتمد على السياق
• عالمي: يستخدم قاعدة المعرفة العالمية
• مختلط: يجمع بين الاسترجاع المحلي والعالمي
• مزيج: يدمج شبكة المعرفة مع الاسترجاع المتجهي
• تجاوز: يمرر الاستعلام مباشرة إلى LLM بدون استرجاع`,queryModeOptions:{naive:"ساذج",local:"محلي",global:"عالمي",hybrid:"مختلط",mix:"مزيج",bypass:"تجاوز"},responseFormat:"تنسيق الرد",responseFormatTooltip:`يحدد تنسيق الرد. أمثلة:
• فقرات متعددة
• فقرة واحدة
• نقاط نقطية`,responseFormatOptions:{multipleParagraphs:"فقرات متعددة",singleParagraph:"فقرة واحدة",bulletPoints:"نقاط نقطية"},topK:"أعلى K نتائج",topKTooltip:"عدد العناصر العلوية للاسترجاع. يمثل الكيانات في وضع 'محلي' والعلاقات في وضع 'عالمي'",topKPlaceholder:"عدد النتائج",maxTokensTextUnit:"أقصى عدد من الرموز لوحدة النص",maxTokensTextUnitTooltip:"الحد الأقصى لعدد الرموز المسموح به لكل جزء نصي مسترجع",maxTokensGlobalContext:"أقصى عدد من الرموز للسياق العالمي",maxTokensGlobalContextTooltip:"الحد الأقصى لعدد الرموز المخصص لأوصاف العلاقات في الاسترجاع العالمي",maxTokensLocalContext:"أقصى عدد من الرموز للسياق المحلي",maxTokensLocalContextTooltip:"الحد الأقصى لعدد الرموز المخصص لأوصاف الكيانات في الاسترجاع المحلي",historyTurns:"دورات التاريخ",historyTurnsTooltip:"عدد الدورات الكاملة للمحادثة (أزواج المستخدم-المساعد) التي يجب مراعاتها في سياق الرد",historyTurnsPlaceholder:"عدد دورات التاريخ",hlKeywords:"الكلمات المفتاحية عالية المستوى",hlKeywordsTooltip:"قائمة الكلمات المفتاحية عالية المستوى لإعطائها الأولوية في الاسترجاع. افصل بينها بفواصل",hlkeywordsPlaceHolder:"أدخل الكلمات المفتاحية",llKeywords:"الكلمات المفتاحية منخفضة المستوى",llKeywordsTooltip:"قائمة الكلمات المفتاحية منخفضة المستوى لتحسين تركيز الاسترجاع. افصل بينها بفواصل",onlyNeedContext:"تحتاج فقط إلى السياق",onlyNeedContextTooltip:"إذا كان صحيحًا، يتم إرجاع السياق المسترجع فقط دون إنشاء رد",onlyNeedPrompt:"تحتاج فقط إلى المطالبة",onlyNeedPromptTooltip:"إذا كان صحيحًا، يتم إرجاع المطالبة المولدة فقط دون إنتاج رد",streamResponse:"تدفق الرد",streamResponseTooltip:"إذا كان صحيحًا، يتيح إخراج التدفق للردود في الوقت الفعلي"}},yy={loading:"جارٍ تحميل وثائق واجهة برمجة التطبيقات..."},vy={title:"مفتاح واجهة برمجة التطبيقات مطلوب",description:"الرجاء إدخال مفتاح واجهة برمجة التطبيقات للوصول إلى الخدمة",placeholder:"أدخل مفتاح واجهة برمجة التطبيقات",save:"حفظ"},by={settings:fy,header:ry,login:dy,common:my,documentPanel:hy,graphPanel:gy,retrievePanel:py,apiSite:yy,apiKeyAlert:vy},Sy={language:"語言",theme:"主題",light:"淺色",dark:"深色",system:"系統"},Ty={documents:"文件",knowledgeGraph:"知識圖譜",retrieval:"檢索",api:"API",projectRepository:"專案庫",logout:"登出",themeToggle:{switchToLight:"切換至淺色主題",switchToDark:"切換至深色主題"}},xy={description:"請輸入您的帳號和密碼登入系統",username:"帳號",usernamePlaceholder:"請輸入帳號",password:"密碼",passwordPlaceholder:"請輸入密碼",loginButton:"登入",loggingIn:"登入中...",successMessage:"登入成功",errorEmptyFields:"請輸入您的帳號和密碼",errorInvalidCredentials:"登入失敗，請檢查帳號和密碼",authDisabled:"認證已停用，使用免登入模式",guestMode:"免登入"},Ay={cancel:"取消",save:"儲存"},Dy={clearDocuments:{button:"清空",tooltip:"清空文件",title:"清空文件",description:"此操作將從系統中移除所有文件",warning:"警告：此操作將永久刪除所有文件，無法復原！",confirm:"確定要清空所有文件嗎？",confirmPrompt:"請輸入 yes 確認操作",confirmPlaceholder:"輸入 yes 以確認",clearCache:"清空 LLM 快取",confirmButton:"確定",success:"文件清空成功",cacheCleared:"快取清空成功",cacheClearFailed:`清空快取失敗：
{{error}}`,failed:`清空文件失敗：
{{message}}`,error:`清空文件失敗：
{{error}}`},uploadDocuments:{button:"上傳",tooltip:"上傳文件",title:"上傳文件",description:"拖曳檔案至此處或點擊瀏覽",single:{uploading:"正在上傳 {{name}}：{{percent}}%",success:`上傳成功：
{{name}} 上傳完成`,failed:`上傳失敗：
{{name}}
{{message}}`,error:`上傳失敗：
{{name}}
{{error}}`},batch:{uploading:"正在上傳檔案...",success:"檔案上傳完成",error:"部分檔案上傳失敗"},generalError:`上傳失敗
{{error}}`,fileTypes:"支援的檔案類型：TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS",fileUploader:{singleFileLimit:"一次只能上傳一個檔案",maxFilesLimit:"最多只能上傳 {{count}} 個檔案",fileRejected:"檔案 {{name}} 被拒絕",unsupportedType:"不支援的檔案類型",fileTooLarge:"檔案過大，最大允許 {{maxSize}}",dropHere:"將檔案拖放至此處",dragAndDrop:"拖放檔案至此處，或點擊選擇檔案",removeFile:"移除檔案",uploadDescription:"您可以上傳{{isMultiple ? '多個' : count}}個檔案（每個檔案最大{{maxSize}}）",duplicateFile:"檔案名稱與伺服器上的快取重複"}},documentManager:{title:"文件管理",scanButton:"掃描",scanTooltip:"掃描輸入目錄中的文件",pipelineStatusButton:"pipeline 狀態",pipelineStatusTooltip:"查看pipeline 狀態",uploadedTitle:"已上傳文件",uploadedDescription:"已上傳文件清單及其狀態",emptyTitle:"無文件",emptyDescription:"尚未上傳任何文件",columns:{id:"ID",summary:"摘要",status:"狀態",length:"長度",chunks:"分塊",created:"建立時間",updated:"更新時間",metadata:"元資料"},status:{all:"全部",completed:"已完成",processing:"處理中",pending:"等待中",failed:"失敗"},errors:{loadFailed:`載入文件失敗
{{error}}`,scanFailed:`掃描文件失敗
{{error}}`,scanProgressFailed:`取得掃描進度失敗
{{error}}`},fileNameLabel:"檔案名稱",showButton:"顯示",hideButton:"隱藏",showFileNameTooltip:"顯示檔案名稱",hideFileNameTooltip:"隱藏檔案名稱"},pipelineStatus:{title:"pipeline 狀態",busy:"pipeline 忙碌中",requestPending:"待處理請求",jobName:"工作名稱",startTime:"開始時間",progress:"進度",unit:"梯次",latestMessage:"最新訊息",historyMessages:"歷史訊息",errors:{fetchFailed:`取得pipeline 狀態失敗
{{error}}`}}},Ny={dataIsTruncated:"圖資料已截斷至最大回傳節點數",statusDialog:{title:"LightRAG 伺服器設定",description:"查看目前系統狀態和連線資訊"},legend:"圖例",nodeTypes:{person:"人物角色",category:"分類",geo:"地理名稱",location:"位置",organization:"組織機構",event:"事件",equipment:"設備",weapon:"武器",animal:"動物",unknown:"未知",object:"物品",group:"群組",technology:"技術"},sideBar:{settings:{settings:"設定",healthCheck:"健康檢查",showPropertyPanel:"顯示屬性面板",showSearchBar:"顯示搜尋列",showNodeLabel:"顯示節點標籤",nodeDraggable:"節點可拖曳",showEdgeLabel:"顯示 Edge 標籤",hideUnselectedEdges:"隱藏未選取的 Edge",edgeEvents:"Edge 事件",maxQueryDepth:"最大查詢深度",maxNodes:"最大回傳節點數",maxLayoutIterations:"最大版面配置迭代次數",resetToDefault:"重設為預設值",edgeSizeRange:"Edge 粗細範圍",depth:"深度",max:"最大值",degree:"鄰邊",apiKey:"API key",enterYourAPIkey:"輸入您的 API key",save:"儲存",refreshLayout:"重新整理版面配置"},zoomControl:{zoomIn:"放大",zoomOut:"縮小",resetZoom:"重設縮放",rotateCamera:"順時針旋轉圖形",rotateCameraCounterClockwise:"逆時針旋轉圖形"},layoutsControl:{startAnimation:"繼續版面配置動畫",stopAnimation:"停止版面配置動畫",layoutGraph:"圖形版面配置",layouts:{Circular:"環形",Circlepack:"圓形打包",Random:"隨機",Noverlaps:"無重疊","Force Directed":"力導向","Force Atlas":"力圖"}},fullScreenControl:{fullScreen:"全螢幕",windowed:"視窗"},legendControl:{toggleLegend:"切換圖例顯示"}},statusIndicator:{connected:"已連線",disconnected:"未連線"},statusCard:{unavailable:"狀態資訊不可用",storageInfo:"儲存資訊",workingDirectory:"工作目錄",inputDirectory:"輸入目錄",llmConfig:"LLM 設定",llmBinding:"LLM 綁定",llmBindingHost:"LLM 綁定主機",llmModel:"LLM 模型",maxTokens:"最大權杖數",embeddingConfig:"嵌入設定",embeddingBinding:"嵌入綁定",embeddingBindingHost:"嵌入綁定主機",embeddingModel:"嵌入模型",storageConfig:"儲存設定",kvStorage:"KV 儲存",docStatusStorage:"文件狀態儲存",graphStorage:"圖形儲存",vectorStorage:"向量儲存"},propertiesView:{editProperty:"編輯{{property}}",editPropertyDescription:"在下方文字區域編輯屬性值。",errors:{duplicateName:"節點名稱已存在",updateFailed:"更新節點失敗",tryAgainLater:"請稍後重試"},success:{entityUpdated:"節點更新成功",relationUpdated:"關係更新成功"},node:{title:"節點",id:"ID",labels:"標籤",degree:"度數",properties:"屬性",relationships:"關係(子圖內)",expandNode:"展開節點",pruneNode:"修剪節點",deleteAllNodesError:"拒絕刪除圖中的所有節點",nodesRemoved:"已刪除 {{count}} 個節點，包括孤立節點",noNewNodes:"沒有發現可以展開的節點",propertyNames:{description:"描述",entity_id:"名稱",entity_type:"類型",source_id:"來源ID",Neighbour:"鄰接",file_path:"來源",keywords:"Keys",weight:"權重"}},edge:{title:"關係",id:"ID",type:"類型",source:"來源節點",target:"目標節點",properties:"屬性"}},search:{placeholder:"搜尋節點...",message:"還有 {count} 個"},graphLabels:{selectTooltip:"選擇查詢標籤",noLabels:"未找到標籤",label:"標籤",placeholder:"搜尋標籤...",andOthers:"還有 {count} 個",refreshTooltip:"重載圖形數據(新增檔案後需重載)"},emptyGraph:"無數據(請重載圖形數據)"},Ey={chatMessage:{copyTooltip:"複製到剪貼簿",copyError:"複製文字到剪貼簿失敗"},retrieval:{startPrompt:"輸入查詢開始檢索",clear:"清空",send:"送出",placeholder:"輸入查詢內容 (支援模式前綴：/<Query Mode>)",error:"錯誤：取得回應失敗",queryModeError:"僅支援以下查詢模式：{{modes}}",queryModePrefixInvalid:"無效的查詢模式前綴。請使用：/<模式> [空格] 查詢內容"},querySettings:{parametersTitle:"參數",parametersDescription:"設定查詢參數",queryMode:"查詢模式",queryModeTooltip:`選擇檢索策略：
• Naive：基礎搜尋，無進階技術
• Local：上下文相關資訊檢索
• Global：利用全域知識庫
• Hybrid：結合本地和全域檢索
• Mix：整合知識圖譜和向量檢索
• Bypass：直接傳遞查詢到LLM，不進行檢索`,queryModeOptions:{naive:"Naive",local:"Local",global:"Global",hybrid:"Hybrid",mix:"Mix",bypass:"Bypass"},responseFormat:"回應格式",responseFormatTooltip:`定義回應格式。例如：
• 多段落
• 單段落
• 重點`,responseFormatOptions:{multipleParagraphs:"多段落",singleParagraph:"單段落",bulletPoints:"重點"},topK:"Top K結果",topKTooltip:"檢索的前幾項結果數。在'local'模式下表示實體，在'global'模式下表示關係",topKPlaceholder:"結果數量",maxTokensTextUnit:"文字單元最大權杖數",maxTokensTextUnitTooltip:"每個檢索文字區塊允許的最大權杖數",maxTokensGlobalContext:"全域上下文最大權杖數",maxTokensGlobalContextTooltip:"全域檢索中關係描述的最大權杖數",maxTokensLocalContext:"本地上下文最大權杖數",maxTokensLocalContextTooltip:"本地檢索中實體描述的最大權杖數",historyTurns:"歷史輪次",historyTurnsTooltip:"回應上下文中考慮的完整對話輪次（使用者-助手對）數量",historyTurnsPlaceholder:"歷史輪次數",hlKeywords:"進階關鍵字",hlKeywordsTooltip:"檢索中優先考慮的進階關鍵字清單。用逗號分隔",hlkeywordsPlaceHolder:"輸入關鍵字",llKeywords:"基礎關鍵字",llKeywordsTooltip:"用於細化檢索重點的基礎關鍵字清單。用逗號分隔",onlyNeedContext:"僅需上下文",onlyNeedContextTooltip:"如果為True，僅回傳檢索到的上下文而不產生回應",onlyNeedPrompt:"僅需提示",onlyNeedPromptTooltip:"如果為True，僅回傳產生的提示而不產生回應",streamResponse:"串流回應",streamResponseTooltip:"如果為True，啟用即時串流輸出回應"}},My={loading:"正在載入 API 文件..."},zy={title:"需要 API key",description:"請輸入您的 API key 以存取服務",placeholder:"請輸入 API key",save:"儲存"},Cy={settings:Sy,header:Ty,login:xy,common:Ay,documentPanel:Dy,graphPanel:Ny,retrievePanel:Ey,apiSite:My,apiKeyAlert:zy},Oy=()=>{var g;try{const b=localStorage.getItem("settings-storage");if(b)return((g=JSON.parse(b).state)==null?void 0:g.language)||"en"}catch(b){console.error("Failed to get stored language:",b)}return"en"};cs.use($g).init({resources:{en:{translation:Qp},zh:{translation:Ip},fr:{translation:oy},ar:{translation:by},zh_TW:{translation:Cy}},lng:Oy(),fallbackLng:"en",interpolation:{escapeValue:!1},returnEmptyString:!1,returnNull:!1});ll.subscribe(g=>{const b=g.language;cs.language!==b&&cs.changeLanguage(b)});ap.createRoot(document.getElementById("root")).render(o.jsx(E.StrictMode,{children:o.jsx(Up,{})}));
